"""
Convenience functions for latex output.

Felix Metzner 2021
"""

import os
import re
import copy

from typing import Optional, Dict

from rdstar.utilities import PathType


__all__ = [
    "check_for_special_chars",
    "escape_special_chars",
    "get_latex_cmd_to_attributes_mapping",
    "create_latex_command_file",
]


def check_for_special_chars(s: str) -> None:
    regex_check = bool(
        re.match(
            pattern="^[a-zA-Z]*$",
            string=s,
        )
    )
    if not regex_check:
        raise ValueError(f"String {s} contains special characters")


def replace_digits(s: str) -> str:
    the_string = copy.copy(s)
    digits_dict = {
        "0": "ZERO",
        "1": "ONE",
        "2": "TWO",
        "3": "THREE",
        "4": "FOUR",
        "5": "FIVE",
        "6": "SIX",
        "7": "SEVEN",
        "8": "EIGHT",
        "9": "NINE",
    }
    for d, ds in digits_dict.items():
        the_string = the_string.replace(d, ds)
    return the_string


def escape_special_chars(s: str) -> str:
    _s = replace_digits(s=s)
    return "".join(word.title() if i else word for i, word in enumerate(_s.split("_")))


def get_latex_cmd_to_attributes_mapping(attributes_mapping: Dict[str, str]) -> Dict[str, str]:
    return {escape_special_chars(s=k): v for k, v in attributes_mapping.items()}


def create_latex_command_file(
    output_path: PathType,
    command_mapping_dict: Dict[str, str],
    command_prefix: Optional[str] = None,
    overwrite: bool = False,
) -> None:
    if os.path.exists(output_path) and not overwrite:
        raise FileExistsError(f"overwrite is deactivated and {output_path} exists!")

    _command_prefix = ""  # type: str
    if command_prefix is not None:
        _command_prefix = command_prefix

    with open(file=output_path, mode="w") as latex_file:
        for command_name, command_value in command_mapping_dict.items():
            cmd_name = _command_prefix + command_name  # type: str
            check_for_special_chars(s=cmd_name)
            latex_file.write(r"\newcommand{" + "\\" + cmd_name + "}{" + command_value + "}\n")
