import numpy as np
import matplotlib.pyplot as plt
import copy

from rdstar.utilities import PathType

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export

from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetup
from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict
from rdstar.offline_analysis.dss_fitting.bf_calculator import BFCalculator

__all__ = [
    "BFPlotter",
]


class BFPlotter:
    def __init__(
        self,
        fitting_setup: BFFittingSetup,
        bf_calculator: BFCalculator,
    ) -> None:
        self._fitting_setup = fitting_setup
        self._bf_calculator = bf_calculator

        self._yield_parameters = self._fitting_setup.yield_parameters
        self._fit_case = BFFitCaseDict.get_fit_case(fit_case_name=self._fitting_setup.fit_case_name)

    def plot_branching_fractions(
        self,
        file_name: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        y_pos = np.arange(len(self._yield_parameters))
        fig, ax = plt.subplots(figsize=(6.5, 5))
        bfs = []
        lower_errors = []
        upper_errors = []
        for yield_parameter in self._yield_parameters:
            bf, upper_error, lower_error = self._bf_calculator.get_bf_for_parameter(parameter=yield_parameter)
            bfs.append(bf)
            upper_errors.append(upper_error)
            lower_errors.append(lower_error)

        if self._fitting_setup.nominal_bfs:
            bf_dict = {
                key: (value.nominal_value, value.std_dev) for key, value in self._fitting_setup.nominal_bfs.items()
            }
            plot_colors = self._fit_case.plot_colors
            assert isinstance(plot_colors, dict)
            ax.errorbar(
                x=np.array([bf_dict[yield_param][0] for yield_param in self._yield_parameters]) * 100,
                y=y_pos,
                yerr=None,
                xerr=np.array([bf_dict[yield_param][1] for yield_param in self._yield_parameters]) * 100,
                ls="",
                elinewidth=40,
                marker="",
                color=[plot_colors[yield_param] for yield_param in self._yield_parameters],
                capsize=0,
                alpha=0.6,
            )

        ax.errorbar(
            x=np.array(bfs) * 100,
            y=y_pos,
            yerr=0,
            xerr=(np.array(lower_errors) * 100, np.array(upper_errors) * 100),
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self._fit_case.plot_info_text,
        )

        left, right = plt.xlim()
        ax.set_xlim(0, right * 1.5)

        bottom, top = plt.ylim()
        ax.set_ylim(bottom - 0.5, top + 0.5)

        for i in range(len(y_pos)):
            bf_value_str = "{:0.3f}".format(bfs[i] * 100)
            upper_error = upper_errors[i] * 100
            lower_error = lower_errors[i] * 100
            error = "{:0.3f}".format(max(upper_error, lower_error))
            bf_str = fr"${bf_value_str}\pm {error}$"
            ax.text(right * 1.1, y_pos[i], bf_str, color=plot_style.KITColors.dark_grey, va="center")

        plt.legend(loc="best")
        ax.set_xlabel(r"$\mathcal{B}(B\rightarrow D^{(*,**)} \ell \nu)$ in %", plot_style.xlabel_pos)
        ax.set_yticks(y_pos)
        plot_labels = self._fit_case.plot_labels
        assert isinstance(plot_labels, dict)
        ax.set_yticklabels([plot_labels[yield_param] for yield_param in self._yield_parameters])

        export(fig=fig, filename=file_name, target_dir=output_dir)

    def plot_yields(
        self,
        file_name: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        x_pos = np.arange(len(self._yield_parameters) + 1)
        fig, ax = plt.subplots(figsize=(6.5, 5))
        yields = []
        yield_errors = []
        assert isinstance(self._yield_parameters, list), type(self._yield_parameters)
        all_yield_parameters = copy.copy(self._yield_parameters)
        all_yield_parameters.append(self._fitting_setup.norm_parameter)
        assert isinstance(all_yield_parameters, list), type(all_yield_parameters)
        for yield_parameter in all_yield_parameters:
            yield_dict_for_param = self._bf_calculator.get_yield_dict()[yield_parameter]
            assert isinstance(yield_dict_for_param["yield"], tuple)
            yield_value, yield_error = yield_dict_for_param["yield"]
            yields.append(yield_value)
            yield_errors.append(yield_error)

        plot_colors = self._fit_case.plot_colors
        assert isinstance(plot_colors, dict)
        ax.errorbar(
            x=x_pos,
            y=np.array(yields),
            yerr=np.array(yield_errors),
            xerr=0,
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self._fit_case.plot_info_text,
        )

        plt.legend(loc="best")
        ax.set_ylabel(r"Yield", plot_style.xlabel_pos)
        ax.set_xticks(x_pos)
        plot_labels = self._fit_case.plot_labels
        assert isinstance(plot_labels, dict)
        ax.set_xticklabels([plot_labels[yield_param] for yield_param in all_yield_parameters], rotation=70)

        export(fig=fig, filename=file_name, target_dir=output_dir)

    def plot_efficiencies(
        self,
        file_name: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        x_pos = np.arange(len(self._yield_parameters) + 1)
        fig, ax = plt.subplots(figsize=(6.5, 5))
        efficiencies = []
        upper_errors = []
        lower_errors = []
        assert isinstance(self._yield_parameters, list), type(self._yield_parameters)
        all_yield_parameters = copy.copy(self._yield_parameters)
        all_yield_parameters.append(self._fitting_setup.norm_parameter)
        assert isinstance(all_yield_parameters, list), type(all_yield_parameters)
        for yield_parameter in all_yield_parameters:
            eff, upper_eff, lower_eff = self._bf_calculator.get_efficiency_from_dict(parameter=yield_parameter)
            efficiencies.append(eff)
            upper_errors.append(upper_eff)
            lower_errors.append(lower_eff)

        plot_colors = self._fit_case.plot_colors
        assert isinstance(plot_colors, dict)
        ax.errorbar(
            x=x_pos,
            y=np.array(efficiencies),
            yerr=(np.array(lower_errors), np.array(upper_errors)),
            xerr=0,
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self._fit_case.plot_info_text,
        )

        plt.legend(loc="best")
        ax.set_ylabel(r"Efficiency", plot_style.xlabel_pos)
        ax.set_xticks(x_pos)
        plot_labels = self._fit_case.plot_labels
        assert isinstance(plot_labels, dict)
        ax.set_xticklabels([plot_labels[yield_param] for yield_param in all_yield_parameters], rotation=70)

        export(fig=fig, filename=file_name, target_dir=output_dir)
