import pandas as pd

from uncertainties import ufloat, UFloat
from typing import Dict, Tuple, Union, List

from templatefitter.plotter.plot_style import KITColors


from rdstar.branching_fractions import b_meson_bf_calculation as our_b_bfs
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses as ldp

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo, MCSigIDNameToQueryMappingType
from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import (
    SigIDAndGenIDCombiner,
    SigIDAndGenIDCombinerInputType,
)

from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.dss_fitting.utilities import utilities as dss_utils


def update_colors_and_labels(
    initial_color_dict: Dict[str, str],
    initial_labels_dict: Dict[str, str],
    template_setups: Dict[str, TemplateSetup],
) -> Tuple[Dict[str, str], Dict[str, str]]:
    new_colors: Dict[str, str] = dict()
    new_labels: Dict[str, str] = dict()
    for mode, template_setup in template_setups.items():
        for component in template_setup.template_setup:
            if component not in initial_color_dict.keys() and component not in new_colors.keys():
                for start, color in initial_color_dict.items():
                    if component.startswith(f"{start}_"):
                        new_colors.update({component: color})
            if component not in initial_labels_dict.keys() and component not in new_labels.keys():
                for start, color in initial_labels_dict.items():
                    if component.startswith(f"{start}_"):
                        new_labels.update({component: color})
    assert len(new_colors) == len(new_labels), (len(new_colors), len(new_labels))
    if len(new_colors) > 0:
        initial_color_dict.update(new_colors)
        initial_labels_dict.update(new_labels)

    assert all(
        comp in initial_color_dict.keys()
        for template_setup in template_setups.values()
        for comp in template_setup.template_setup
    )
    assert all(
        comp in initial_labels_dict.keys()
        for template_setup in template_setups.values()
        for comp in template_setup.template_setup
    )

    return initial_color_dict, initial_labels_dict


dss_sig_id_column: str = "dss_sig_id"

dss_sig_queries: MCSigIDNameToQueryMappingType = {
    # fmt: off
    "ChargedBDzeroStarPiZ": MatchQueryInfo(id_info=201, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(201)),
    "ChargedBDonePrimePiZ": MatchQueryInfo(id_info=202, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(202)),
    "ChargedBDonePiZ": MatchQueryInfo(id_info=203, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(203)),
    "ChargedBDtwoStarPiZ": MatchQueryInfo(id_info=204, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(204)),
    "ChargedBDzeroStarPiP": MatchQueryInfo(id_info=205, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(205)),
    "ChargedBDonePrimePiP": MatchQueryInfo(id_info=206, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(206)),
    "ChargedBDonePiP": MatchQueryInfo(id_info=207, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(207)),
    "ChargedBDtwoStarPiP": MatchQueryInfo(id_info=208, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(208)),
    "ChargedBGapDone1": MatchQueryInfo(id_info=209, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(209)),
    "ChargedBGapDone2": MatchQueryInfo(id_info=210, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(210)),
    "ChargedBDl": MatchQueryInfo(id_info=211, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(211)),
    "ChargedBDSl": MatchQueryInfo(id_info=212, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(212)),
    "ChargedBBbarBKG": MatchQueryInfo(id_info=213, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(213)),
    "ChargedBDtau": MatchQueryInfo(id_info=214, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(214)),
    "ChargedBDStau": MatchQueryInfo(id_info=215, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(215)),
    "ChargedHadronicBKG": MatchQueryInfo(id_info=216, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(216)),
    "NeutralBDzeroStarPiZ": MatchQueryInfo(id_info=301, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(301)),
    "NeutralBDonePrimePiZ": MatchQueryInfo(id_info=302, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(302)),
    "NeutralBDonePiZ": MatchQueryInfo(id_info=303, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(303)),
    "NeutralBDtwoStarPiZ": MatchQueryInfo(id_info=304, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(304)),
    "NeutralBDzeroStarPiP": MatchQueryInfo(id_info=305, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(305)),
    "NeutralBDonePrimePiP": MatchQueryInfo(id_info=306, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(306)),
    "NeutralBDonePiP": MatchQueryInfo(id_info=307, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(307)),
    "NeutralBDtwoStarPiP": MatchQueryInfo(id_info=308, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(308)),
    "NeutralBGapDone1": MatchQueryInfo(id_info=309, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(309)),
    "NeutralBGapDone2": MatchQueryInfo(id_info=310, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(310)),
    "NeutralBDl": MatchQueryInfo(id_info=311, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(311)),
    "NeutralBDSl": MatchQueryInfo(id_info=312, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(312)),
    "NeutralBBbarBKG": MatchQueryInfo(id_info=313, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(313)),
    "NeutralBDtau": MatchQueryInfo(id_info=314, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(314)),
    "NeutralBDStau": MatchQueryInfo(id_info=315, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(315)),
    "NeutralHadronicBKG": MatchQueryInfo(id_info=316, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(316)),
    "Continuum": MatchQueryInfo(id_info=401, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(401)),
    # fmt: on
}


def get_id(component: str) -> Tuple[int, ...]:
    return (dss_sig_queries[component].id,)


def get_ids_without(components: Union[Tuple[str, ...], List]) -> Tuple[int, ...]:
    return tuple([mq_info.id for comp, mq_info in dss_sig_queries.items() if comp not in components])


dss_colors: Dict[str, str] = {
    # Background Components
    "BKG": KITColors.white,
    # Normalisation Components
    "NormBzeroDs": KITColors.kit_cyan,
    "NormBzeroD": KITColors.kit_blue,
    "NormBchargedDs": KITColors.kit_cyan,
    "NormBchargedD": KITColors.kit_blue,
    # Signal Components
    "ChargedBDzeroStarPiP": KITColors.kit_purple,
    "ChargedBDonePrimePiP": KITColors.kit_orange,
    "ChargedBDonePiP": KITColors.kit_green,
    "ChargedBDtwoStarPiP": KITColors.kit_maygreen,
    "NeutralBDzeroStarPiP": KITColors.kit_purple,
    "NeutralBDonePrimePiP": KITColors.kit_orange,
    "NeutralBDonePiP": KITColors.kit_green,
    "NeutralBDtwoStarPiP": KITColors.kit_maygreen,
    "ChargedBDzeroStarPiZ": KITColors.kit_purple,
    "ChargedBDonePrimePiZ": KITColors.kit_orange,
    "ChargedBDonePiZ": KITColors.kit_green,
    "ChargedBDtwoStarPiZ": KITColors.kit_maygreen,
    "NeutralBDzeroStarPiZ": KITColors.kit_purple,
    "NeutralBDonePrimePiZ": KITColors.kit_orange,
    "NeutralBDonePiZ": KITColors.kit_green,
    "NeutralBDtwoStarPiZ": KITColors.kit_maygreen,
    # Other Background
    "DstarstarPiZ": KITColors.light_grey,
    "DstarstarPiP": KITColors.light_grey,
    "Continuum": KITColors.light_grey,
    "ChargedHadronicBKG": KITColors.dark_grey,
    "ChargedBBbarBKG": KITColors.kit_brown,
    "NeutralHadronicBKG": KITColors.dark_grey,
    "NeutralBBbarBKG": KITColors.kit_brown,
}

dss_labels: Dict[str, str] = {
    # Signal Components
    "ChargedBDzeroStarPiP": ldp.bpm_to_d0st_l_nu_with_pi_pm,
    "ChargedBDonePrimePiP": ldp.bpm_to_dprime1_l_nu_with_pi_pm,
    "ChargedBDonePiP": ldp.bpm_to_d_1_l_nu_with_pi_pm,
    "ChargedBDtwoStarPiP": ldp.bpm_to_dst_2_l_nu_with_pi_pm,
    "NeutralBDzeroStarPiP": ldp.bz_to_d0st_l_nu_with_pi_pm,
    "NeutralBDonePrimePiP": ldp.bz_to_dprime1_l_nu_with_pi_pm,
    "NeutralBDonePiP": ldp.bz_to_d_1_l_nu_with_pi_pm,
    "NeutralBDtwoStarPiP": ldp.bz_to_dst_2_l_nu_with_pi_pm,
    "ChargedBDzeroStarPiZ": ldp.bpm_to_d0st_l_nu_with_pi_zero,
    "ChargedBDonePrimePiZ": ldp.bpm_to_dprime1_l_nu_with_pi_zero,
    "ChargedBDonePiZ": ldp.bpm_to_d_1_l_nu_with_pi_zero,
    "ChargedBDtwoStarPiZ": ldp.bpm_to_dst_2_l_nu_with_pi_zero,
    "NeutralBDzeroStarPiZ": ldp.bz_to_d0st_l_nu_with_pi_zero,
    "NeutralBDonePrimePiZ": ldp.bz_to_dprime1_l_nu_with_pi_zero,
    "NeutralBDonePiZ": ldp.bz_to_d_1_l_nu_with_pi_zero,
    "NeutralBDtwoStarPiZ": ldp.bz_to_dst_2_l_nu_with_pi_zero,
    # Normalisation Components
    "NormBzeroDs": ldp.bz_to_d_st_l_nu,
    "NormBchargedDs": ldp.bpm_to_d_st_l_nu,
    "NormBzeroD": ldp.bz_to_d_l_nu,
    "NormBchargedD": ldp.bpm_to_d_l_nu,
    # Background Components
    "BKG": "BKG",
    # Other Background
    "DstarstarPiZ": r"$B \rightarrow D^{**}(\rightarrow D^{(\ast)} \pi^0) \ell \nu$ BKG",
    "DstarstarPiP": r"$B \rightarrow D^{**}(\rightarrow D^{(\ast)} \pi^\pm) \ell \nu$ BKG",
    "Continuum": "Continuum",
    "ChargedHadronicBKG": "Hadronic BKG",
    "ChargedBBbarBKG": r"$\mathrm{B\bar{B}}$ BKG",
    "NeutralHadronicBKG": "Hadronic BKG",
    "NeutralBBbarBKG": r"$\mathrm{B\bar{B}}$ BKG",
}

__dss_pi_plus_factor: float = 0.6667
__dss_pi_zero_factor: float = 0.3333

sig_dss_nominal_bfs: Dict[str, UFloat] = {
    "ChargedBDzeroStarPiP": our_b_bfs.bf_bp_to_d0st_l_nu_our_cal * __dss_pi_plus_factor,
    "ChargedBDonePrimePiP": our_b_bfs.bf_bp_to_dprime1_l_nu_our_cal * __dss_pi_plus_factor,
    "ChargedBDonePiP": our_b_bfs.bf_bp_to_d_1_l_nu_our_cal * __dss_pi_plus_factor,
    "ChargedBDtwoStarPiP": our_b_bfs.bf_bp_to_dst_2_l_nu_our_cal * __dss_pi_plus_factor,
    "NeutralBDzeroStarPiP": our_b_bfs.bf_b0_to_d0st_l_nu_our_cal * __dss_pi_plus_factor,
    "NeutralBDonePrimePiP": our_b_bfs.bf_b0_to_dprime1_l_nu_our_cal * __dss_pi_plus_factor,
    "NeutralBDonePiP": our_b_bfs.bf_b0_to_d_1_l_nu_our_cal * __dss_pi_plus_factor,
    "NeutralBDtwoStarPiP": our_b_bfs.bf_b0_to_dst_2_l_nu_our_cal * __dss_pi_plus_factor,
    "ChargedBDzeroStarPiZ": our_b_bfs.bf_bp_to_d0st_l_nu_our_cal * __dss_pi_zero_factor,
    "ChargedBDonePrimePiZ": our_b_bfs.bf_bp_to_dprime1_l_nu_our_cal * __dss_pi_zero_factor,
    "ChargedBDonePiZ": our_b_bfs.bf_bp_to_d_1_l_nu_our_cal * __dss_pi_zero_factor,
    "ChargedBDtwoStarPiZ": our_b_bfs.bf_bp_to_dst_2_l_nu_our_cal * __dss_pi_zero_factor,
    "NeutralBDzeroStarPiZ": our_b_bfs.bf_b0_to_d0st_l_nu_our_cal * __dss_pi_zero_factor,
    "NeutralBDonePrimePiZ": our_b_bfs.bf_b0_to_dprime1_l_nu_our_cal * __dss_pi_zero_factor,
    "NeutralBDonePiZ": our_b_bfs.bf_b0_to_d_1_l_nu_our_cal * __dss_pi_zero_factor,
    "NeutralBDtwoStarPiZ": our_b_bfs.bf_b0_to_dst_2_l_nu_our_cal * __dss_pi_zero_factor,
    "NormBchargedD": our_b_bfs.bf_bp_to_d_l_nu_avg,
    "NormBzeroD": our_b_bfs.bf_b0_to_d_l_nu_avg,
}


_current_nominal_bfs: Dict[str, UFloat] = {
    "NormBchargedD": our_b_bfs.bf_bp_to_d_l_nu_avg,
    "NormBchargedDs": our_b_bfs.bf_bp_to_d_st_l_nu_avg,
    "NormBzeroD": our_b_bfs.bf_b0_to_d_l_nu_avg,
    "NormBzeroDs": our_b_bfs.bf_b0_to_d_st_l_nu_avg,
    "NormBzeroDs_DM": our_b_bfs.bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DM": our_b_bfs.bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DSM": our_b_bfs.bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DSM": our_b_bfs.bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DZ": our_b_bfs.bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DZ": our_b_bfs.bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DSZ": our_b_bfs.bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DSZ": our_b_bfs.bf_b0_to_d_l_nu_avg,
    "NormBchargedDs_DM": our_b_bfs.bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DM": our_b_bfs.bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DSM": our_b_bfs.bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DSM": our_b_bfs.bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DZ": our_b_bfs.bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DZ": our_b_bfs.bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DSZ": our_b_bfs.bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DSZ": our_b_bfs.bf_bp_to_d_l_nu_avg,
}

pdg_bf_b0_to_d_l_nu_avg = ufloat(0.0231, 0.001)
pdg_bf_bp_to_d_l_nu_avg = ufloat(0.0235, 0.0009)
pdg_bf_b0_to_d_st_l_nu_avg = ufloat(0.0506, 0.0012)
pdg_bf_bp_to_d_st_l_nu_avg = ufloat(0.0566, 0.0022)

_current_pdg_bfs: Dict[str, UFloat] = {
    "NormBchargedD": pdg_bf_bp_to_d_l_nu_avg,
    "NormBchargedDs": pdg_bf_bp_to_d_st_l_nu_avg,
    "NormBzeroD": pdg_bf_b0_to_d_l_nu_avg,
    "NormBzeroDs": pdg_bf_b0_to_d_st_l_nu_avg,
    "NormBzeroDs_DM": pdg_bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DM": pdg_bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DSM": pdg_bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DSM": pdg_bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DZ": pdg_bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DZ": pdg_bf_b0_to_d_l_nu_avg,
    "NormBzeroDs_DSZ": pdg_bf_b0_to_d_st_l_nu_avg,
    "NormBzeroD_DSZ": pdg_bf_b0_to_d_l_nu_avg,
    "NormBchargedDs_DM": pdg_bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DM": pdg_bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DSM": pdg_bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DSM": pdg_bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DZ": pdg_bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DZ": pdg_bf_bp_to_d_l_nu_avg,
    "NormBchargedDs_DSZ": pdg_bf_bp_to_d_st_l_nu_avg,
    "NormBchargedD_DSZ": pdg_bf_bp_to_d_l_nu_avg,
}

current_pdg_bfs: Dict[str, UFloat] = _current_nominal_bfs

# Charged B charged pi setup
charged_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs": ["ChargedBDSl"],
            "NormBchargedD": ["ChargedBDl"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedD": ["ChargedBDl"],
            "NormBchargedDs": ["ChargedBDSl"],
        }
    ),
    "DPPiM": TemplateSetup(
        template_setup={
            "BKG_DPPiM": ["BKG_DPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
    "DSPPiM": TemplateSetup(
        template_setup={
            "BKG_DSPPiM": ["BKG_DSPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=charged_B_charged_pi_template_setup,
)

charged_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 1): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("BKG_DSZ", 2): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("BKG_DPPiM", 3): {
        "DPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
    ("BKG_DSPPiM", 4): {
        "DSPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
}

charged_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=charged_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Alternative Setup
alternative_charged_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "Continuum_DZ": ["Continuum_DZ"],
            "ChargedHadronicBKG_DZ": ["ChargedHadronicBKG_DZ"],
            "ChargedBBbarBKG_DZ": ["ChargedBBbarBKG_DZ"],
            "NormBchargedDs": ["ChargedBDSl"],
            "NormBchargedD": ["ChargedBDl"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "Continuum_DSZ": ["Continuum_DSZ"],
            "ChargedHadronicBKG_DSZ": ["ChargedHadronicBKG_DSZ"],
            "ChargedBBbarBKG_DSZ": ["ChargedBBbarBKG_DSZ"],
            "NormBchargedD": ["ChargedBDl"],
            "NormBchargedDs": ["ChargedBDSl"],
        }
    ),
    "DPPiM": TemplateSetup(
        template_setup={
            "BKG_DPPiM": ["BKG_DPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
    "DSPPiM": TemplateSetup(
        template_setup={
            "BKG_DSPPiM": ["BKG_DSPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=alternative_charged_B_charged_pi_template_setup,
)

alternative_charged_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 1): {
        "DZ": get_ids_without(
            components=("ChargedBDSl", "ChargedBDl", "Continuum", "ChargedHadronicBKG", "ChargedBBbarBKG")
        )
    },
    ("Continuum_DZ", 2): {"DZ": get_id(component="Continuum")},
    ("ChargedHadronicBKG_DZ", 3): {"DZ": get_id(component="ChargedHadronicBKG")},
    ("ChargedBBbarBKG_DZ", 9): {"DZ": get_id(component="ChargedBBbarBKG")},
    ("BKG_DSZ", 4): {
        "DSZ": get_ids_without(
            components=("ChargedBDSl", "ChargedBDl", "Continuum", "ChargedHadronicBKG", "ChargedBBbarBKG")
        )
    },
    ("Continuum_DSZ", 5): {"DSZ": get_id(component="Continuum")},
    ("ChargedHadronicBKG_DSZ", 6): {"DSZ": get_id(component="ChargedHadronicBKG")},
    ("ChargedBBbarBKG_DSZ", 10): {"DSZ": get_id(component="ChargedBBbarBKG")},
    ("BKG_DPPiM", 7): {
        "DPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
    ("BKG_DSPPiM", 8): {
        "DSPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
}

alternative_charged_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=alternative_charged_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)


alternative_decoupled_charged_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "Continuum_DZ": ["Continuum_DZ"],
            "ChargedHadronicBKG_DZ": ["ChargedHadronicBKG_DZ"],
            "ChargedBBbarBKG_DZ": ["ChargedBBbarBKG_DZ"],
            "NormBchargedDs_DZ": ["NormBchargedDs_DZ"],
            "NormBchargedD_DZ": ["NormBchargedD_DZ"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "Continuum_DSZ": ["Continuum_DSZ"],
            "ChargedHadronicBKG_DSZ": ["ChargedHadronicBKG_DSZ"],
            "ChargedBBbarBKG_DSZ": ["ChargedBBbarBKG_DSZ"],
            "NormBchargedD_DSZ": ["NormBchargedD_DSZ"],
            "NormBchargedDs_DSZ": ["NormBchargedDs_DSZ"],
        }
    ),
    "DPPiM": TemplateSetup(
        template_setup={
            "BKG_DPPiM": ["BKG_DPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
    "DSPPiM": TemplateSetup(
        template_setup={
            "BKG_DSPPiM": ["BKG_DSPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=alternative_decoupled_charged_B_charged_pi_template_setup,
)

alternative_decoupled_charged_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 1): {
        "DZ": get_ids_without(
            components=("ChargedBDSl", "ChargedBDl", "Continuum", "ChargedHadronicBKG", "ChargedBBbarBKG")
        )
    },
    ("Continuum_DZ", 2): {"DZ": get_id(component="Continuum")},
    ("ChargedHadronicBKG_DZ", 3): {"DZ": get_id(component="ChargedHadronicBKG")},
    ("ChargedBBbarBKG_DZ", 9): {"DZ": get_id(component="ChargedBBbarBKG")},
    ("BKG_DSZ", 4): {
        "DSZ": get_ids_without(
            components=("ChargedBDSl", "ChargedBDl", "Continuum", "ChargedHadronicBKG", "ChargedBBbarBKG")
        )
    },
    ("Continuum_DSZ", 5): {"DSZ": get_id(component="Continuum")},
    ("ChargedHadronicBKG_DSZ", 6): {"DSZ": get_id(component="ChargedHadronicBKG")},
    ("ChargedBBbarBKG_DSZ", 10): {"DSZ": get_id(component="ChargedBBbarBKG")},
    ("BKG_DPPiM", 7): {
        "DPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
    ("BKG_DSPPiM", 8): {
        "DSPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
    ("NormBchargedDs_DZ", 11): {"DZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DZ", 12): {"DZ": get_id(component="ChargedBDl")},
    ("NormBchargedDs_DSZ", 21): {"DSZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DSZ", 22): {"DSZ": get_id(component="ChargedBDl")},
}

alternative_decoupled_charged_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=alternative_decoupled_charged_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Charged B neutral pi setup
charged_B_neutral_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs": ["ChargedBDSl"],
            "NormBchargedD": ["ChargedBDl"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedD": ["ChargedBDl"],
            "NormBchargedDs": ["ChargedBDSl"],
        }
    ),
    "DZPiZ": TemplateSetup(
        template_setup={
            "BKG_DZPiZ": ["BKG_DZPiZ"],
            "ChargedBDzeroStarPiZ": ["ChargedBDzeroStarPiZ"],
            "ChargedBDonePrimePiZ": ["ChargedBDonePrimePiZ"],
            "ChargedBDtwoStarPiZ": ["ChargedBDtwoStarPiZ"],
            "ChargedBDonePiZ": ["ChargedBDonePiZ"],
        }
    ),
    "DSZPiZ": TemplateSetup(
        template_setup={
            "BKG_DSZPiZ": ["BKG_DSZPiZ"],
            "ChargedBDzeroStarPiZ": ["ChargedBDzeroStarPiZ"],
            "ChargedBDonePrimePiZ": ["ChargedBDonePrimePiZ"],
            "ChargedBDtwoStarPiZ": ["ChargedBDtwoStarPiZ"],
            "ChargedBDonePiZ": ["ChargedBDonePiZ"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=charged_B_neutral_pi_template_setup,
)

charged_B_neutral_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 1): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("BKG_DSZ", 2): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("BKG_DZPiZ", 3): {
        "DZPiZ": get_ids_without(
            components=("ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDtwoStarPiZ", "ChargedBDonePiZ")
        )
    },
    ("BKG_DSZPiZ", 4): {
        "DSZPiZ": get_ids_without(
            components=("ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDtwoStarPiZ", "ChargedBDonePiZ")
        )
    },
}

charged_B_neutral_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=charged_B_neutral_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Neutral B charged pi setup
neutral_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs": ["NeutralBDSl"],
            "NormBzeroD": ["NeutralBDl"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroD": ["NeutralBDl"],
            "NormBzeroDs": ["NeutralBDSl"],
        }
    ),
    "DZPiM": TemplateSetup(
        template_setup={
            "BKG_DZPiM": ["BKG_DZPiM"],
            "NeutralBDzeroStarPiP": ["NeutralBDzeroStarPiP"],
            "NeutralBDonePrimePiP": ["NeutralBDonePrimePiP"],
            "NeutralBDtwoStarPiP": ["NeutralBDtwoStarPiP"],
            "NeutralBDonePiP": ["NeutralBDonePiP"],
        }
    ),
    "DSZPiM": TemplateSetup(
        template_setup={
            "BKG_DSZPiM": ["BKG_DSZPiM"],
            "NeutralBDzeroStarPiP": ["NeutralBDzeroStarPiP"],
            "NeutralBDonePrimePiP": ["NeutralBDonePrimePiP"],
            "NeutralBDtwoStarPiP": ["NeutralBDtwoStarPiP"],
            "NeutralBDonePiP": ["NeutralBDonePiP"],
        }
    ),
}

neutral_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 1): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("BKG_DSM", 2): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("BKG_DZPiM", 3): {
        "DZPiM": get_ids_without(
            components=("NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDtwoStarPiP", "NeutralBDonePiP")
        )
    },
    ("BKG_DSZPiM", 4): {
        "DSZPiM": get_ids_without(
            components=("NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDtwoStarPiP", "NeutralBDonePiP")
        )
    },
}

neutral_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=neutral_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Neutral B neutral pi setup
neutral_B_neutral_pi_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs": ["NeutralBDSl"],
            "NormBzeroD": ["NeutralBDl"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroD": ["NeutralBDl"],
            "NormBzeroDs": ["NeutralBDSl"],
        }
    ),
    "DMPiZ": TemplateSetup(
        template_setup={
            "BKG_DMPiZ": ["BKG_DMPiZ"],
            "NeutralBDzeroStarPiZ": ["NeutralBDzeroStarPiZ"],
            "NeutralBDonePrimePiZ": ["NeutralBDonePrimePiZ"],
            "NeutralBDtwoStarPiZ": ["NeutralBDtwoStarPiZ"],
            "NeutralBDonePiZ": ["NeutralBDonePiZ"],
        }
    ),
    "DSMPiZ": TemplateSetup(
        template_setup={
            "BKG_DSMPiZ": ["BKG_DSMPiZ"],
            "NeutralBDzeroStarPiZ": ["NeutralBDzeroStarPiZ"],
            "NeutralBDonePrimePiZ": ["NeutralBDonePrimePiZ"],
            "NeutralBDtwoStarPiZ": ["NeutralBDtwoStarPiZ"],
            "NeutralBDonePiZ": ["NeutralBDonePiZ"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=neutral_B_neutral_pi_template_setup,
)

neutral_B_neutral_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 1): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("BKG_DSM", 2): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("BKG_DMPiZ", 3): {
        "DMPiZ": get_ids_without(
            components=("NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDtwoStarPiZ", "NeutralBDonePiZ")
        )
    },
    ("BKG_DSMPiZ", 4): {
        "DSMPiZ": get_ids_without(
            components=("NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDtwoStarPiZ", "NeutralBDonePiZ")
        )
    },
}

neutral_B_neutral_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=neutral_B_neutral_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Decoupled Setups
# Charged B charged pi setup
decoupled_charged_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs_DZ": ["NormBchargedDs_DZ"],
            "NormBchargedD_DZ": ["NormBchargedD_DZ"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedDs_DSZ": ["NormBchargedDs_DSZ"],
            "NormBchargedD_DSZ": ["NormBchargedD_DSZ"],
        }
    ),
    "DPPiM": TemplateSetup(
        template_setup={
            "BKG_DPPiM": ["BKG_DPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
    "DSPPiM": TemplateSetup(
        template_setup={
            "BKG_DSPPiM": ["BKG_DSPPiM"],
            "ChargedBDzeroStarPiP": ["ChargedBDzeroStarPiP"],
            "ChargedBDonePrimePiP": ["ChargedBDonePrimePiP"],
            "ChargedBDtwoStarPiP": ["ChargedBDtwoStarPiP"],
            "ChargedBDonePiP": ["ChargedBDonePiP"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=decoupled_charged_B_charged_pi_template_setup,
)

decoupled_charged_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 10): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DZ", 11): {"DZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DZ", 12): {"DZ": get_id(component="ChargedBDl")},
    ("BKG_DSZ", 20): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DSZ", 21): {"DSZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DSZ", 22): {"DSZ": get_id(component="ChargedBDl")},
    ("BKG_DPPiM", 3): {
        "DPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
    ("BKG_DSPPiM", 4): {
        "DSPPiM": get_ids_without(
            components=("ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDtwoStarPiP", "ChargedBDonePiP")
        )
    },
}

decoupled_charged_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_charged_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Charged B neutral pi setup
decoupled_charged_B_neutral_pi_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs_DZ": ["NormBchargedDs_DZ"],
            "NormBchargedD_DZ": ["NormBchargedD_DZ"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedDs_DSZ": ["NormBchargedDs_DSZ"],
            "NormBchargedD_DSZ": ["NormBchargedD_DSZ"],
        }
    ),
    "DZPiZ": TemplateSetup(
        template_setup={
            "BKG_DZPiZ": ["BKG_DZPiZ"],
            "ChargedBDzeroStarPiZ": ["ChargedBDzeroStarPiZ"],
            "ChargedBDonePrimePiZ": ["ChargedBDonePrimePiZ"],
            "ChargedBDtwoStarPiZ": ["ChargedBDtwoStarPiZ"],
            "ChargedBDonePiZ": ["ChargedBDonePiZ"],
        }
    ),
    "DSZPiZ": TemplateSetup(
        template_setup={
            "BKG_DSZPiZ": ["BKG_DSZPiZ"],
            "ChargedBDzeroStarPiZ": ["ChargedBDzeroStarPiZ"],
            "ChargedBDonePrimePiZ": ["ChargedBDonePrimePiZ"],
            "ChargedBDtwoStarPiZ": ["ChargedBDtwoStarPiZ"],
            "ChargedBDonePiZ": ["ChargedBDonePiZ"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=decoupled_charged_B_neutral_pi_template_setup,
)

decoupled_charged_B_neutral_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 10): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DZ", 11): {"DZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DZ", 12): {"DZ": get_id(component="ChargedBDl")},
    ("BKG_DSZ", 20): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DSZ", 21): {"DSZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DSZ", 22): {"DSZ": get_id(component="ChargedBDl")},
    ("BKG_DZPiZ", 3): {
        "DZPiZ": get_ids_without(
            components=("ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDtwoStarPiZ", "ChargedBDonePiZ")
        )
    },
    ("BKG_DSZPiZ", 4): {
        "DSZPiZ": get_ids_without(
            components=("ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDtwoStarPiZ", "ChargedBDonePiZ")
        )
    },
}

decoupled_charged_B_neutral_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_charged_B_neutral_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Neutral B charged pi setup
decoupled_neutral_B_charged_pi_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs_DM": ["NormBzeroDs_DM"],
            "NormBzeroD_DM": ["NormBzeroD_DM"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroDs_DSM": ["NormBzeroDs_DSM"],
            "NormBzeroD_DSM": ["NormBzeroD_DSM"],
        }
    ),
    "DZPiM": TemplateSetup(
        template_setup={
            "BKG_DZPiM": ["BKG_DZPiM"],
            "NeutralBDzeroStarPiP": ["NeutralBDzeroStarPiP"],
            "NeutralBDonePrimePiP": ["NeutralBDonePrimePiP"],
            "NeutralBDtwoStarPiP": ["NeutralBDtwoStarPiP"],
            "NeutralBDonePiP": ["NeutralBDonePiP"],
        }
    ),
    "DSZPiM": TemplateSetup(
        template_setup={
            "BKG_DSZPiM": ["BKG_DSZPiM"],
            "NeutralBDzeroStarPiP": ["NeutralBDzeroStarPiP"],
            "NeutralBDonePrimePiP": ["NeutralBDonePrimePiP"],
            "NeutralBDtwoStarPiP": ["NeutralBDtwoStarPiP"],
            "NeutralBDonePiP": ["NeutralBDonePiP"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=decoupled_neutral_B_charged_pi_template_setup,
)

decoupled_neutral_B_charged_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 10): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DM", 11): {"DM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DM", 12): {"DM": get_id(component="NeutralBDl")},
    ("BKG_DSM", 20): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DSM", 21): {"DSM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DSM", 22): {"DSM": get_id(component="NeutralBDl")},
    ("BKG_DZPiM", 3): {
        "DZPiM": get_ids_without(
            components=("NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDtwoStarPiP", "NeutralBDonePiP")
        )
    },
    ("BKG_DSZPiM", 4): {
        "DSZPiM": get_ids_without(
            components=("NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDtwoStarPiP", "NeutralBDonePiP")
        )
    },
}

decoupled_neutral_B_charged_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_neutral_B_charged_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)

# Neutral B neutral pi setup
decoupled_neutral_B_neutral_pi_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs_DM": ["NormBzeroDs_DM"],
            "NormBzeroD_DM": ["NormBzeroD_DM"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroD_DSM": ["NormBzeroD_DSM"],
            "NormBzeroDs_DSM": ["NormBzeroDs_DSM"],
        }
    ),
    "DMPiZ": TemplateSetup(
        template_setup={
            "BKG_DMPiZ": ["BKG_DMPiZ"],
            "NeutralBDzeroStarPiZ": ["NeutralBDzeroStarPiZ"],
            "NeutralBDonePrimePiZ": ["NeutralBDonePrimePiZ"],
            "NeutralBDtwoStarPiZ": ["NeutralBDtwoStarPiZ"],
            "NeutralBDonePiZ": ["NeutralBDonePiZ"],
        }
    ),
    "DSMPiZ": TemplateSetup(
        template_setup={
            "BKG_DSMPiZ": ["BKG_DSMPiZ"],
            "NeutralBDzeroStarPiZ": ["NeutralBDzeroStarPiZ"],
            "NeutralBDonePrimePiZ": ["NeutralBDonePrimePiZ"],
            "NeutralBDtwoStarPiZ": ["NeutralBDtwoStarPiZ"],
            "NeutralBDonePiZ": ["NeutralBDonePiZ"],
        }
    ),
}
dss_colors, dss_labels = update_colors_and_labels(
    initial_color_dict=dss_colors,
    initial_labels_dict=dss_labels,
    template_setups=decoupled_neutral_B_neutral_pi_template_setup,
)

decoupled_neutral_B_neutral_pi_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 10): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DM", 11): {"DM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DM", 12): {"DM": get_id(component="NeutralBDl")},
    ("BKG_DSM", 20): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DSM", 21): {"DSM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DSM", 22): {"DSM": get_id(component="NeutralBDl")},
    ("BKG_DMPiZ", 3): {
        "DMPiZ": get_ids_without(
            components=("NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDtwoStarPiZ", "NeutralBDonePiZ")
        )
    },
    ("BKG_DSMPiZ", 4): {
        "DSMPiZ": get_ids_without(
            components=("NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDtwoStarPiZ", "NeutralBDonePiZ")
        )
    },
}

decoupled_neutral_B_neutral_pi_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_neutral_B_neutral_pi_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)


# Coupled Neutral B Normalization Fit on data
coupled_neutral_B_normalization_fit_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs": ["NeutralBDSl"],
            "NormBzeroD": ["NeutralBDl"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroD": ["NeutralBDl"],
            "NormBzeroDs": ["NeutralBDSl"],
        }
    ),
}

coupled_neutral_B_normalization_fit_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 1): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("BKG_DSM", 2): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
}

coupled_neutral_B_normalization_fit_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=coupled_neutral_B_normalization_fit_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)


# Coupled Charged B Normalization Fit on data
coupled_charged_B_normalization_fit_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs": ["ChargedBDSl"],
            "NormBchargedD": ["ChargedBDl"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedD": ["ChargedBDl"],
            "NormBchargedDs": ["ChargedBDSl"],
        }
    ),
}

coupled_charged_B_normalization_fit_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 1): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("BKG_DSZ", 2): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
}

coupled_charged_B_normalization_fit_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=coupled_charged_B_normalization_fit_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)


# Decoupled Neutral B Normalization Fit on data
decoupled_neutral_B_normalization_fit_template_setup: Dict[str, TemplateSetup] = {
    "DM": TemplateSetup(
        template_setup={
            "BKG_DM": ["BKG_DM"],
            "NormBzeroDs_DM": ["NormBzeroDs_DM"],
            "NormBzeroD_DM": ["NormBzeroD_DM"],
        }
    ),
    "DSM": TemplateSetup(
        template_setup={
            "BKG_DSM": ["BKG_DSM"],
            "NormBzeroD_DSM": ["NormBzeroD_DSM"],
            "NormBzeroDs_DSM": ["NormBzeroDs_DSM"],
        }
    ),
}

decoupled_neutral_B_normalization_fit_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DM", 10): {"DM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DM", 11): {"DM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DM", 12): {"DM": get_id(component="NeutralBDl")},
    ("BKG_DSM", 20): {"DSM": get_ids_without(components=("NeutralBDSl", "NeutralBDl"))},
    ("NormBzeroDs_DSM", 21): {"DSM": get_id(component="NeutralBDSl")},
    ("NormBzeroD_DSM", 22): {"DSM": get_id(component="NeutralBDl")},
}

decoupled_neutral_B_normalization_fit_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_neutral_B_normalization_fit_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)


# Decoupled Charged B Normalization Fit on data
decoupled_charged_B_normalization_fit_template_setup: Dict[str, TemplateSetup] = {
    "DZ": TemplateSetup(
        template_setup={
            "BKG_DZ": ["BKG_DZ"],
            "NormBchargedDs_DZ": ["NormBchargedDs_DZ"],
            "NormBchargedD_DZ": ["NormBchargedD_DZ"],
        }
    ),
    "DSZ": TemplateSetup(
        template_setup={
            "BKG_DSZ": ["BKG_DSZ"],
            "NormBchargedDs_DSZ": ["NormBchargedDs_DSZ"],
            "NormBchargedD_DSZ": ["NormBchargedD_DSZ"],
        }
    ),
}

decoupled_charged_B_normalization_fit_sigID_customizing_dict: SigIDAndGenIDCombinerInputType = {
    ("BKG_DZ", 10): {"DZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DZ", 11): {"DZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DZ", 12): {"DZ": get_id(component="ChargedBDl")},
    ("BKG_DSZ", 20): {"DSZ": get_ids_without(components=("ChargedBDSl", "ChargedBDl"))},
    ("NormBchargedDs_DSZ", 21): {"DSZ": get_id(component="ChargedBDSl")},
    ("NormBchargedD_DSZ", 22): {"DSZ": get_id(component="ChargedBDl")},
}

decoupled_charged_B_normalization_fit_sig_queries = SigIDAndGenIDCombiner(
    customizing_dict=decoupled_charged_B_normalization_fit_sigID_customizing_dict,
    sig_id_column_name=dss_sig_id_column,
    decay_mode_dict=dss_utils.all_decay_modes,
).get_updated_component_dict(initial_sig_query_dict=dss_sig_queries)
