"""
Fitter class for extraction of B -> D** l nu BF and/or efficiency correction factors.

Felix Metzner 2022
"""

import numpy as np
import pandas as pd

from typing import Optional, Tuple

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.dss_cal_fitters.dss_cal_fitter_base import (
    DssBFCalibrationMappingType,
    DssBFCorrFitterBase,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "DssBFCorrFitter",
]


logger = log_to_default(name=__name__)


class DssBFCorrFitter(DssBFCorrFitterBase):

    this_fitter_setup: FitSetupTriplet = FitSetupCollection.dss_c_first_approach__fs

    ordered_parameter_names: Tuple[str, ...] = (
        # D** pi0 Components
        "cB_D_npi_l_yield",
        "cB_Ds_npi_l_yield",
        "nB_D_npi_l_yield",
        "nB_Ds_npi_l_yield",
        # D** pi+/- Components
        "cB_D_cpi_l_yield",
        "cB_Ds_cpi_l_yield",
        "nB_D_cpi_l_yield",
        "nB_Ds_cpi_l_yield",
        # D(*) Components
        "B_to_Dwwos_in_D_cpi_yield",
        "B_to_Dwwos_in_Ds_cpi_yield",
        "B_to_Dwwos_in_D_npi_yield",
        "B_to_Dwwos_in_Ds_npi_yield",
        # Bkg Components
        "bkg_in_d_cpi_yield",
        "bkg_in_ds_cpi_yield",
        "bkg_in_d_npi_yield",
        "bkg_in_ds_npi_yield",
    )

    cB_D_npi_l_yield_index: int = 0
    cB_Ds_npi_l_yield_index: int = 1
    nB_D_npi_l_yield_index: int = 2
    nB_Ds_npi_l_yield_index: int = 3

    cB_D_cpi_l_yield_index: int = 4
    cB_Ds_cpi_l_yield_index: int = 5
    nB_D_cpi_l_yield_index: int = 6
    nB_Ds_cpi_l_yield_index: int = 7

    B_to_Dwwos_in_D_cpi_yield_index: int = 8
    B_to_Dwwos_in_Ds_cpi_yield_index: int = 9
    B_to_Dwwos_in_D_npi_yield_index: int = 10
    B_to_Dwwos_in_Ds_npi_yield_index: int = 11

    bkg_in_d_cpi_yield_index: int = 12
    bkg_in_ds_cpi_yield_index: int = 13
    bkg_in_d_npi_yield_index: int = 14
    bkg_in_ds_npi_yield_index: int = 15

    end_yield_params_index: int = bkg_in_ds_npi_yield_index

    main_floating_params_slice = np.s_[cB_D_npi_l_yield_index : end_yield_params_index + 1 : 1]

    number_of_hardcoded_floating_params: int = end_yield_params_index + 1

    fixed_params_indices: Tuple[int, ...] = tuple([])
    manual_nuisances_params_indices: Tuple[int, ...] = tuple([])

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=None,
            **kwargs,
        )

        assert len(self.fit_setup_info.components) == self.end_yield_params_index + 1, (
            [(c.comp_index, c.name) for c in self.fit_setup_info.components],
            len(self.fit_setup_info.components),
            self.end_yield_params_index + 1,
        )

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    @property
    def _parameter_limits(self) -> Tuple[Tuple[int, Tuple[Optional[float], Optional[float]]], ...]:
        return tuple([(i, (0.0, None)) for i in range(self.number_of_hardcoded_floating_params)])

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        _chi2 = np.sum(
            (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
            / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
            axis=None,
        )
        assert isinstance(_chi2, float)
        return _chi2

    def gaussian_constrains(self, params: np.ndarray) -> float:
        # TODO!!!!!!
        return 0.0

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return params[self.main_floating_params_slice]

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = 0  # TODO!
        return self.number_of_hardcoded_floating_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        # Nuisances:
        automated_nuisances_initial_values: Tuple[float, ...] = tuple([])

        _initial_params = np.array(
            [
                _normalization_factors[self.cB_D_npi_l_yield_index],
                _normalization_factors[self.cB_Ds_npi_l_yield_index],
                _normalization_factors[self.nB_D_npi_l_yield_index],
                _normalization_factors[self.nB_Ds_npi_l_yield_index],
                #
                _normalization_factors[self.cB_D_cpi_l_yield_index],
                _normalization_factors[self.cB_Ds_cpi_l_yield_index],
                _normalization_factors[self.nB_D_cpi_l_yield_index],
                _normalization_factors[self.nB_Ds_cpi_l_yield_index],
                #
                _normalization_factors[self.B_to_Dwwos_in_D_cpi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_Ds_cpi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_D_npi_yield_index],
                _normalization_factors[self.B_to_Dwwos_in_Ds_npi_yield_index],
                #
                _normalization_factors[self.bkg_in_d_cpi_yield_index],
                _normalization_factors[self.bkg_in_ds_cpi_yield_index],
                _normalization_factors[self.bkg_in_d_npi_yield_index],
                _normalization_factors[self.bkg_in_ds_npi_yield_index],
                #
                *automated_nuisances_initial_values,
            ]
        )

        assert len(_initial_params) == self.end_of_floating_params, (
            len(_initial_params),
            self.end_of_floating_params,
            _initial_params,
        )

        return _initial_params

    # endregion

    # region Correction application info
    @classmethod
    def get_correction_mapping(cls) -> DssBFCalibrationMappingType:
        _possible_result: DssBFCalibrationMappingType = {
            (RecoModeInfos.charged_b_to_d_modes.ids, (1,)): cls.cB_D_npi_l_yield_index,
            (RecoModeInfos.neutral_b_to_d_modes.ids, (1,)): cls.cB_Ds_npi_l_yield_index,
            (RecoModeInfos.charged_b_to_ds_modes.ids, (1,)): cls.nB_D_npi_l_yield_index,
            (RecoModeInfos.neutral_b_to_ds_modes.ids, (1,)): cls.nB_Ds_npi_l_yield_index,
            #
            (RecoModeInfos.charged_b_to_d_modes.ids, (2,)): cls.cB_D_cpi_l_yield_index,
            (RecoModeInfos.neutral_b_to_d_modes.ids, (2,)): cls.cB_Ds_cpi_l_yield_index,
            (RecoModeInfos.charged_b_to_ds_modes.ids, (2,)): cls.nB_D_cpi_l_yield_index,
            (RecoModeInfos.neutral_b_to_ds_modes.ids, (2,)): cls.nB_Ds_cpi_l_yield_index,
        }
        raise NotImplementedError("The 'get_correction_mapping' is WIP!")
        return _possible_result

    # endregion
