"""
Tools for rdstar fit plots results.

Felix Metzner 2023
"""


import os
import numpy as np
import pandas as pd

from iminuit import Minuit
from typing import Tuple, Optional

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_units import LatexUnits

from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.comparison_tools.data_mc_comparer import DataMCComparer

from rdstar.offline_analysis.systematics import full_systematics_cols
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollections

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_rdstar import RDStarFitter
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.rdstar_fit_obs_setups import FitObsSetupMain


__all__ = [
    "RDStarFitPlots",
]


logger = log_to_default(name=__name__)


# region Sideband plot infos
data_component_label: str = "Asimov Data"
rdstar_fit_plot_title: str = r"Asimov $-$ Post Fit"
rdstar_pre_fit_plot_title: str = r"Asimov $-$ Pre Fit"

rdstar_fit_plot_text: str = ""

additional_selection_mms_gt_1p5gev: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{RDStarHistVars.missing_mass_res_corr.df_label} > 1.5",
    cut_latex_label=rf"{RDStarHistVars.missing_mass_res_corr.variable_name[:-1]} > 1.5\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)
additional_mms_gt_2gev_selection_fit_plot_text: str = additional_selection_mms_gt_1p5gev.cut_latex_label


plot_reco_mode_info_collection = BSigRecoModes.main_modes
plot_mc_matching_collection = MCMatchingCollections.extended_sig_id__basic__collection

fit_plot_fig_size: Tuple[float, float] = (6.04, 5.0)
fit_plot_height_ratio: Tuple[float, float] = (3.5, 1.0)


rdstar_plot_infos: Tuple[SidebandInfo, ...] = (
    SidebandInfo(
        key="",
        hist_var=RDStarHistVars.missing_mass_res_corr_for_main_fit,
        title=rdstar_fit_plot_title,
        info_text=rdstar_fit_plot_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        fig_size=fit_plot_fig_size,
        height_ratio=fit_plot_height_ratio,
        y_scale=(1.3, 1.6),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=data_component_label,
        additional_info={"special_binning": FitObsSetupMain.binning_for_reco_modes_obs_fit_mms},
    ),
    SidebandInfo(
        key="",
        hist_var=RDStarHistVars.pl_star_signal_fit,
        title=rdstar_fit_plot_title,
        info_text=rdstar_fit_plot_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.03),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        fig_size=fit_plot_fig_size,
        height_ratio=fit_plot_height_ratio,
        y_scale=(1.3, 1.6),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=data_component_label,
    ),
    SidebandInfo(
        key="",
        hist_var=RDStarHistVars.pl_star_signal_fit,
        title=rdstar_fit_plot_title,
        info_text=additional_mms_gt_2gev_selection_fit_plot_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        fig_size=fit_plot_fig_size,
        height_ratio=fit_plot_height_ratio,
        y_scale=(1.3, 1.6),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=data_component_label,
        file_name_appendix="with_mms_gt_2gev",
        additional_selection_info=additional_selection_mms_gt_1p5gev,
    ),
    SidebandInfo(
        key="",
        hist_var=RDStarHistVars.neextra_for_main_fit,
        title=rdstar_fit_plot_title,
        info_text=rdstar_fit_plot_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        fig_size=fit_plot_fig_size,
        height_ratio=fit_plot_height_ratio,
        y_scale=(1.3, 1.6),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=data_component_label,
        additional_info={"special_binning": FitObsSetupMain.binning_for_reco_modes_obs_fit_neextra},
    ),
    SidebandInfo(
        key="",
        hist_var=RDStarHistVars.neextra_for_main_fit,
        title=rdstar_fit_plot_title,
        info_text=additional_mms_gt_2gev_selection_fit_plot_text,
        reco_mode_info_collection=plot_reco_mode_info_collection,
        mc_matching_collection=plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        fig_size=fit_plot_fig_size,
        height_ratio=fit_plot_height_ratio,
        y_scale=(1.3, 1.6),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=data_component_label,
        file_name_appendix="with_mms_gt_2gev",
        additional_selection_info=additional_selection_mms_gt_1p5gev,
        additional_info={"special_binning": FitObsSetupMain.binning_for_reco_modes_obs_fit_neextra},
    ),
)

# endregion


# region RDStar fit plotter class
class RDStarFitPlots:

    plot_infos: Tuple[SidebandInfo, ...] = rdstar_plot_infos

    def __init__(
        self,
        evaluation_name: str,
        base_output_path: PathType,
    ) -> None:

        self.evaluation_name: str = evaluation_name
        self.base_output_path: PathType = base_output_path

        self.output_path: PathType = os.path.join(self.base_output_path, self.evaluation_name)

    def produce_fit_result_plots(
        self,
        fitter: RDStarFitter,
        minimizer: Minuit,
        mc_sample: pd.DataFrame,
        data_sample: Optional[pd.DataFrame],
        add_required_sig_id_col: bool,
        produce_pre_fit_plot: bool = False,
        use_systematics: bool = False,
        ratio_type: str = "normal",
        gof_check_method: Optional[str] = None,
    ) -> None:
        if not fitter.fit_setup_info.data_setup_info.recorded_data:
            assert data_sample is None
            used_data_sample: pd.DataFrame = mc_sample
            data_weight_col: Optional[str] = fitter.mc_weight_col_name
        else:
            used_data_sample = data_sample
            data_weight_col = None

        os.makedirs(self.output_path, exist_ok=True)

        if not produce_pre_fit_plot:
            params: np.ndarray = np.array(minimizer.values)
        else:
            params = fitter.initial_parameters

        rescaled_mc_sample = fitter.get_rescaled_mc_sample_with(
            sample=mc_sample,
            weight_col_name=fitter.mc_weight_col_name,
            parameters=params,
            add_required_sig_id_col=add_required_sig_id_col,
        )

        for base_plot_info in self.plot_infos:

            if not produce_pre_fit_plot:
                plot_info: SidebandInfo = base_plot_info
                plot_name: str = f"post_fit_{plot_info.hist_var.df_label}"
            else:
                plot_name = f"pre_fit_{base_plot_info.hist_var.df_label}"

                appendix: str = "pre_fit"
                if base_plot_info.file_name_appendix is not None:
                    appendix += f"_{base_plot_info.file_name_appendix}"

                plot_info = SidebandInfo.create_modified_comparer_info(
                    comparer_info=base_plot_info,
                    title=rdstar_pre_fit_plot_title,
                    file_name_appendix=appendix,
                )

            if plot_info.file_name_appendix is not None:
                plot_name += f"_{plot_info.file_name_appendix}"
            plot_output_path: PathType = os.path.join(self.output_path, plot_name)

            os.makedirs(plot_output_path, exist_ok=True)

            if use_systematics:
                used_plot_info: SidebandInfo = SidebandInfo.create_comparer_infos_with_systematics(
                    comparer_info_list=[plot_info],
                    systematics_collection=full_systematics_cols,
                    sys_name_tag="with_sys",
                )[0]
            else:
                used_plot_info = plot_info

            data_mc_comparer = DataMCComparer(comparer_info=used_plot_info)

            _mc_df = data_mc_comparer.apply_additional_selection(df=rescaled_mc_sample)
            _data_df = data_mc_comparer.apply_additional_selection(df=used_data_sample)

            data_mc_comparer.plot_data_mc_comparison(
                mc_df=_mc_df,
                data_df=_data_df,
                dry_run=False,
                export_path=plot_output_path,
                weight_col=fitter.mc_weight_col_name,
                data_weight_col=data_weight_col,
                style="stacked",
                normalize_mc_to_data=False,
                ratio_type=ratio_type,
                gof_check_method=gof_check_method,
                adaptive_binning=False,
                show_plots=False,
            )


# endregion
