"""
Base fitter class for extraction of tagging efficiency correction factors.

Felix Metzner 2023
"""

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Dict

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_base import FitterBase
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "TagCalCorrIDType",
    "TagCorrFitterBase",
]


logger = log_to_default(name=__name__)


TagCalCorrIDType = Tuple[Tuple[int, ...], Tuple[int, ...]]


class TagCorrFitterBase(FitterBase, ABC):

    this_fitter_setup: Optional[FitSetupTriplet] = None

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        file_dict_dir_path: Optional[PathType] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=file_dict_dir_path,
            **kwargs,
        )

    # region Cost function
    @abstractmethod
    def cost_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'cost_fcn' must be implemented in the derived class.")

    @abstractmethod
    def chi2_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'chi2_fcn' must be implemented in the derived class.")

    @abstractmethod
    def gaussian_constrains(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'gaussian_constrains_fcn' must be implemented in the derived class.")

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    @abstractmethod
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        raise NotImplementedError("The 'get_yields' must be implemented in the derived class.")

    # endregion

    # region Correction application info
    @classmethod
    @abstractmethod
    def get_correction_mapping(cls) -> Dict[TagCalCorrIDType, int]:
        raise NotImplementedError("The 'get_correction_mapping' must be implemented in the derived class.")

    @classmethod
    @abstractmethod
    def get_full_correction_mapping(cls) -> Dict[TagCalCorrIDType, int]:
        raise NotImplementedError("The 'get_full_correction_mapping' must be implemented in the derived class.")

    @classmethod
    @abstractmethod
    def get_correction_mapping_for_bkg(cls) -> Dict[TagCalCorrIDType, int]:
        raise NotImplementedError("The 'get_correction_mapping_for_bkg' must be implemented in the derived class.")

    @classmethod
    @abstractmethod
    def get_simple_correction_mapping(cls) -> Dict[Tuple[int, ...], int]:
        raise NotImplementedError("The 'get_simple_correction_mapping' must be implemented in the derived class.")

    @classmethod
    @abstractmethod
    def _get_tag_cal_corr_eta_values(cls) -> Dict[str, int]:
        raise NotImplementedError("The '_get_tag_cal_corr_eta_values' must be implemented in the derived class.")

    @classmethod
    def get_background_indices(cls) -> Tuple[int, ...]:
        background_indices = getattr(cls, "background_indices", None)
        if background_indices is None:
            raise RuntimeError(f"Tag Cal fitter class {cls.__name__} does not define 'background_indices'")
        else:
            assert isinstance(background_indices, tuple), (background_indices, type(background_indices))
            assert all(isinstance(v, int) for v in background_indices), (
                background_indices,
                [type(v) for v in background_indices],
            )
            return background_indices

    @classmethod
    def get_tag_cal_corr_eta_values(cls) -> Dict[str, int]:
        required_keys: Tuple[str, ...] = (
            "Bz_to_Dm",
            "Bm_to_Dz",
            "Bz_to_Dsm",
            "Bm_to_Dsz",
        )

        di = cls._get_tag_cal_corr_eta_values()
        assert set(required_keys) == set(di.keys()), (required_keys, di.keys())
        return di

    # endregion
