"""
Class providing information on cuts to be applied for fits.

Felix Metzner 2021
"""

import copy
import itertools
import pandas as pd

from typing import Optional, Tuple, List, Dict, NamedTuple, Any

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos


__all__ = [
    "MVACutRecoModeNames",
    "MVACutInfo",
    "MVACutInfoEntry",
    "SubCutInfoTupleType",
    "CutInfoTupleType",
]


FitCaseRecoModeIDMapType = Dict[str, Tuple[int, ...]]
SubCutInfoTupleType = Tuple[str, float]
CutInfoTupleType = Tuple[SubCutInfoTupleType, ...]


# Reco Mode Names:
class MVACutRecoModeNames:
    reco_mode_DM: str = RecoModeInfos.neutral_b_to_d_modes.name
    reco_mode_DZ: str = RecoModeInfos.charged_b_to_d_modes.name
    reco_mode_DSM: str = RecoModeInfos.neutral_b_to_ds_modes.name
    reco_mode_DSZ: str = RecoModeInfos.charged_b_to_ds_modes.name


# Mapping to old and simpler reco mode keys:
fit_case_reco_mode_str_map: Dict[str, str] = {
    MVACutRecoModeNames.reco_mode_DM: "DM",
    MVACutRecoModeNames.reco_mode_DZ: "DZ",
    MVACutRecoModeNames.reco_mode_DSM: "DSM",
    MVACutRecoModeNames.reco_mode_DSZ: "DSZ",
}
inverted_fit_case_reco_mode_str_map: Dict[str, str] = {v: k for k, v in fit_case_reco_mode_str_map.items()}


class MVACutInfoEntry(NamedTuple):
    # Name of reco mode for which this cut is valid. None means valid for all (other) modes.
    reco_modes: Optional[Tuple[str, ...]]
    cut_variable: str
    cut_value: float  # Cut will be applied as df[cut_variable] >= cut_value under consideration of the reco mode.
    probe_values: Optional[Tuple[float, ...]]

    def get_with_new_cut_value(self, cut_value) -> "MVACutInfoEntry":
        return MVACutInfoEntry(
            reco_modes=self.reco_modes,
            cut_variable=self.cut_variable,
            cut_value=cut_value,
            probe_values=self.probe_values,
        )

    def cuts_on_same_as(self, other: "MVACutInfoEntry") -> bool:
        if self.reco_modes is None and other.reco_modes is None:
            return True
        assert isinstance(self.reco_modes, tuple), type(self.reco_modes).__name__
        assert isinstance(other.reco_modes, tuple), type(other.reco_modes).__name__
        return set(self.reco_modes) == set(other.reco_modes) and self.cut_variable == other.cut_variable

    @staticmethod
    def get_reco_channels_separator() -> str:
        reco_channels_separator: str = "_AND_"
        return copy.copy(reco_channels_separator)

    @property
    def reco_modes_str(self) -> str:
        if self.reco_modes is None:
            return ""
        reco_mode_strings: List[str] = [fit_case_reco_mode_str_map[k] for k in self.reco_modes]
        return self.get_reco_channels_separator().join(reco_mode_strings)

    @staticmethod
    def get_reco_modes_from_str(reco_modes_str: str) -> Optional[Tuple[str, ...]]:
        if reco_modes_str == "":
            return None
        encoded_reco_mode_strings: List[str] = reco_modes_str.split(MVACutInfoEntry.get_reco_channels_separator())
        return tuple(inverted_fit_case_reco_mode_str_map[k] for k in encoded_reco_mode_strings)

    def as_dict(self) -> Dict[str, Any]:
        return self._asdict()

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "MVACutInfoEntry":
        return MVACutInfoEntry(**input_dict)


class MVACutInfo:
    all_other_key: str = "all_or_other_reco_modes"

    fit_case_setting_key: str = "mva_cut_info"
    probing_fit_case_setting_key: str = "mva_cut_info_for_probing"

    __reco_ch_var_separator: str = "__var__"

    reco_mode_column_name: str = Cols.decay_mode_column_name

    def __init__(self, cut_infos: Tuple[MVACutInfoEntry, ...]) -> None:
        assert sum([ci.reco_modes is None for ci in cut_infos]) <= 1, [ci.reco_modes for ci in cut_infos]
        not_none_reco_modes: List[Tuple[str, ...]] = [ci.reco_modes for ci in cut_infos if ci.reco_modes is not None]
        assert len(set(not_none_reco_modes)) == len(not_none_reco_modes), [ci.reco_modes for ci in cut_infos]

        assert all(len(set(ci.reco_modes)) == len(ci.reco_modes) for ci in cut_infos if ci.reco_modes is not None), [
            ci.reco_modes for ci in cut_infos
        ]

        all_reco_modes: List[str] = [v for ci in cut_infos if ci.reco_modes is not None for v in ci.reco_modes]
        assert len(all_reco_modes) == len(set(all_reco_modes)), all_reco_modes

        self._cut_info_dict: Dict[str, MVACutInfoEntry] = {
            ci.reco_modes_str if ci.reco_modes is not None else self.all_other_key: ci for ci in cut_infos
        }
        assert len(self._cut_info_dict.keys()) == len(cut_infos), list(self._cut_info_dict.keys())

        self._probe_combinations: Optional[Tuple["MVACutInfo", ...]] = None

    def check_columns(self, df: pd.DataFrame) -> None:
        required_columns: List[str] = [ci.cut_variable for ci in self._cut_info_dict.values()]
        required_columns.append(self.reco_mode_column_name)
        assert all(c in df.columns for c in required_columns), [c for c in required_columns if c not in df.columns]

    def check_reco_modes(self, fit_case_reco_mode_id_map: FitCaseRecoModeIDMapType) -> None:
        reco_mode_combo_names: List[str] = [k for k in self._cut_info_dict.keys() if k != self.all_other_key]
        reco_mode_combo_tuples: List[Optional[Tuple[str, ...]]] = [
            MVACutInfoEntry.get_reco_modes_from_str(reco_modes_str=rmc) for rmc in reco_mode_combo_names
        ]
        reco_mode_names: List[str] = [rmn for rmt in reco_mode_combo_tuples if rmt is not None for rmn in rmt]

        fit_case_reco_modes: List[str] = list(fit_case_reco_mode_id_map.keys())
        assert all(rmn in fit_case_reco_modes for rmn in reco_mode_names), (
            [n for n in reco_mode_names if n not in fit_case_reco_modes],
            fit_case_reco_modes,
        )

        all_covered_ids = self.get_all_covered_reco_ids(fit_case_reco_mode_id_map=fit_case_reco_mode_id_map)
        assert len(all_covered_ids) == len(set(all_covered_ids)), (
            all_covered_ids,
            len(all_covered_ids),
            len(set(all_covered_ids)),
        )

    def get_cut_query_str(self, fit_case_reco_mode_id_map: FitCaseRecoModeIDMapType) -> str:
        reco_mode_col: str = self.reco_mode_column_name
        reco_mode_id_map: FitCaseRecoModeIDMapType = fit_case_reco_mode_id_map

        string_queries: List[str] = [
            f"(({ci.cut_variable} >= {ci.cut_value}) & ({reco_mode_col} in {[rm_id for r_mode in ci.reco_modes for rm_id in reco_mode_id_map[r_mode]]}))"
            for r_mode_key, ci in self._cut_info_dict.items()
            if r_mode_key != self.all_other_key and ci.reco_modes is not None
        ]

        # Adding case for all (other) reco modes:
        all_ids: Tuple[int, ...] = tuple([i for ids in reco_mode_id_map.values() for i in ids])
        all_covered_ids: Tuple[int, ...] = self.get_all_covered_reco_ids(
            fit_case_reco_mode_id_map=fit_case_reco_mode_id_map
        )
        all_other_ids: Tuple[int, ...] = tuple([i for i in all_ids if i not in all_covered_ids])

        if self.all_other_key in self._cut_info_dict.keys():
            var_col: str = self._cut_info_dict[self.all_other_key].cut_variable
            cut_val: float = self._cut_info_dict[self.all_other_key].cut_value
            others_query_str: str = f"(({var_col} >= {cut_val}) & ({reco_mode_col} in {all_other_ids}))"
        else:
            others_query_str = f"({reco_mode_col} in {all_other_ids})"

        string_queries.append(others_query_str)

        query_str: str = "(" + " | ".join(string_queries) + ")"
        return query_str

    def get_all_covered_reco_ids(self, fit_case_reco_mode_id_map: FitCaseRecoModeIDMapType) -> Tuple[int, ...]:
        all_covered_id_tuples: List[Tuple[int, ...]] = [
            fit_case_reco_mode_id_map[reco_mode]
            for k, ci in self._cut_info_dict.items()
            if k != self.all_other_key and ci.reco_modes is not None
            for reco_mode in ci.reco_modes
        ]
        all_covered_ids: List[int] = [i for ids in all_covered_id_tuples for i in ids]
        return tuple(all_covered_ids)

    @property
    def applies_cuts(self) -> bool:
        return len(self._cut_info_dict) > 0

    @property
    def is_simple_cut(self) -> bool:
        return len(self._cut_info_dict.keys()) == 1 and self.all_other_key in self._cut_info_dict.keys()

    @property
    def cut_variables(self) -> Tuple[str, ...]:
        cut_vars: List[str] = [c.cut_variable for c in self._cut_info_dict.values()]
        return tuple(set(cut_vars))

    def get_cut_tag(self) -> str:
        if self.applies_cuts:
            if self.is_simple_cut:
                return str(self._cut_info_dict[self.all_other_key].cut_value).replace(".", "p")

            return "_".join([f"{k}_{str(v.cut_value).replace('.', 'p')}" for k, v in self._cut_info_dict.items()])
        else:
            return "no_cut"

    @property
    def signature(self) -> str:
        if not self.applies_cuts:
            return "no_cut"

        if self.is_simple_cut:
            return f"{self._cut_info_dict[self.all_other_key].cut_variable}_" + str(
                self._cut_info_dict[self.all_other_key].cut_value
            ).replace(".", "p")

        return "_".join(
            [f"{k}_{v.cut_variable}_{str(v.cut_value).replace('.', 'p')}" for k, v in self._cut_info_dict.items()]
        )

    def get_cut_var_val_pair_from_signature(self, signature: str) -> CutInfoTupleType:
        if signature == "no_cut":
            return tuple()

        assert all(c in signature for c in self.cut_variables), (signature, self.cut_variables)

        if self.is_simple_cut:
            assert len(self.cut_variables) == 1, self.cut_variables
            cvv, cvs = signature.rsplit("_", 1)  # type: str, str
            assert cvv == self._cut_info_dict[self.all_other_key].cut_variable, (
                cvv,
                self._cut_info_dict[self.all_other_key].cut_variable,
            )
            var_val_pairs: List[SubCutInfoTupleType] = [(cvv, float(cvs.replace("p", ".")))]
        else:
            _sig: str = signature + "_"
            var_val_pairs = [
                (
                    f"{k}{self.__reco_ch_var_separator}{v.cut_variable}",
                    float(_sig.split(f"{k}_{v.cut_variable}_", 1)[1].split("_", 1)[0].replace("p", ".")),
                )
                for k, v in self._cut_info_dict.items()
            ]
            assert all(k.split(self.__reco_ch_var_separator, 1)[1] in self.cut_variables for k, _ in var_val_pairs), (
                var_val_pairs,
                self.cut_variables,
            )
            assert len(var_val_pairs) == len(self._cut_info_dict), (var_val_pairs, self._cut_info_dict.keys())

        return tuple(var_val_pairs)

    def split_cut_reco_ch_var_str(self, reco_ch_var_pair: str) -> Tuple[Optional[Tuple[str, ...]], str]:
        if self.is_simple_cut:
            return None, reco_ch_var_pair
        else:
            split_str: List[str] = reco_ch_var_pair.split(self.__reco_ch_var_separator, 1)
            assert len(split_str) == 2, (split_str, reco_ch_var_pair)

            reco_mode_str: str = split_str[0]
            assert reco_mode_str in [ci.reco_modes_str for ci in self._cut_info_dict.values()], (
                reco_mode_str,
                [ci.reco_modes_str for ci in self._cut_info_dict.values()],
            )
            reco_modes_tuple: Tuple[str, ...] = tuple(reco_mode_str.split(MVACutInfoEntry.get_reco_channels_separator()))
            assert set(reco_modes_tuple) in [
                set(ci.reco_modes) for ci in self._cut_info_dict.values() if ci.reco_modes is not None
            ], (
                reco_modes_tuple,
                [ci.reco_modes for ci in self._cut_info_dict.values()],
            )

            cut_variable: str = split_str[1]
            assert cut_variable in self.cut_variables, (cut_variable, self.cut_variables)

            return reco_modes_tuple, cut_variable

    def create_with_updated_entry(self, entry: MVACutInfoEntry) -> "MVACutInfo":
        assert sum([ci.cuts_on_same_as(entry) for ci in self._cut_info_dict.values()]) == 1, entry.as_dict()
        return MVACutInfo(
            cut_infos=tuple([entry if ci.cuts_on_same_as(entry) else ci for ci in self._cut_info_dict.values()])
        )

    @property
    def probe_combinations(self) -> Tuple["MVACutInfo", ...]:
        if self._probe_combinations is None:
            self._probe_combinations = self.get_all_combinations_to_probe()
        return self._probe_combinations

    def get_probe_combinations_dict(self) -> Dict[str, "MVACutInfo"]:
        probe_combinations = self.probe_combinations
        probe_signatures: List[str] = [mci.signature for mci in self.probe_combinations]
        assert len(probe_signatures) == len(set(probe_signatures)), (len(probe_signatures), len(set(probe_signatures)))
        return {mci.signature: mci for mci in probe_combinations}

    def get_all_combinations_to_probe(self) -> Tuple["MVACutInfo", ...]:
        cut_info_keys: Tuple[str, ...] = tuple(self._cut_info_dict.keys())
        probe_options = {k: v.probe_values for k, v in self._cut_info_dict.items() if v.probe_values is not None}

        assert all(len(pv) == len(set(pv)) for pv in probe_options.values()), [
            pk for pk, pv in probe_options.items() if len(pv) != len(set(pv))
        ]

        probe_combinations: List[Tuple[float, ...]] = list(itertools.product(*probe_options.values()))
        probe_mva_cut_infos = tuple(
            [
                MVACutInfo(
                    cut_infos=tuple(
                        [self._cut_info_dict[k].get_with_new_cut_value(cut_value=cv) for k, cv in zip(cut_info_keys, pc)]
                    )
                )
                for pc in probe_combinations
            ]
        )
        return probe_mva_cut_infos

    def as_dict(self) -> Dict[str, Any]:
        return {k: v.as_dict() for k, v in self._cut_info_dict.items()}

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "MVACutInfo":
        new_instance = MVACutInfo(cut_infos=tuple([MVACutInfoEntry.from_dict(v) for v in input_dict.values()]))
        assert new_instance.as_dict() == input_dict, (new_instance.as_dict(), input_dict)
        return new_instance
