"""
Provides variables for fit setups

Felix Metzner 2019
"""

from typing import Optional, Dict, Tuple

from templatefitter.binned_distributions.binning import BinsInputType

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.fitting.fit_case import FitVariable, FitVarDictType
from rdstar.offline_analysis.reco_modes.reco_mode_collections import AllBMesonRecoModes


__all__ = [
    "RecoModeNames",
    "RDStarFitVariables",
]


class RecoModeNames:
    DM: str = "neutral_b_to_d_modes"
    DZ: str = "charged_b_to_d_modes"
    DSM: str = "neutral_b_to_ds_modes"
    DSZ: str = "charged_b_to_ds_modes"
    DMPi: str = "neutral_b_to_dss_l_via_d_pi0_modes"
    DZPi: str = "charged_b_to_dss_l_via_d_pi0_modes"
    DSMPi: str = "neutral_b_to_dss_l_via_ds_pi0_modes"
    DSZPi: str = "charged_b_to_dss_l_via_ds_pi0_modes"
    DMPiPlus: str = "charged_b_to_dss_l_via_d_pip_modes"
    DZPiPlus: str = "neutral_b_to_dss_l_via_d_pip_modes"
    DSMPiPlus: str = "charged_b_to_dss_l_via_ds_pip_modes"
    DSZPiPlus: str = "neutral_b_to_dss_l_via_ds_pip_modes"

    @classmethod
    def get_all_names(cls) -> Tuple[str, ...]:
        return tuple(
            val for attr_name, val in cls.__dict__.items() if not (attr_name.endswith("__") or attr_name.startswith("_"))
        )


_all_reco_mode_names: Tuple[str, ...] = tuple(
    v.name for v in AllBMesonRecoModes.b_to_all_d_l_modes_by_b_charge_d_excitation_and_pi_charge.reco_mode_infos
)
assert all(n in _all_reco_mode_names for n in RecoModeNames.get_all_names()), (
    [n for n in RecoModeNames.get_all_names() if n not in _all_reco_mode_names],
)


class RDStarFitVariables:

    # Simple Signal
    _simple_signal_adapted_mms_bin_edges = (-2.0, -1.4, -0.8, -0.2, 0.4, 1.0, 1.6, 2.2, 3.0, 4.0, 5.0, 6.0, 8.0, 10.0)
    _simple_signal_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: _simple_signal_adapted_mms_bin_edges,
        RecoModeNames.DM: _simple_signal_adapted_mms_bin_edges,
        RecoModeNames.DSZ: _simple_signal_adapted_mms_bin_edges,
        RecoModeNames.DSM: _simple_signal_adapted_mms_bin_edges,
    }
    simple_signal_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_for_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_signal_fit,
                alternative_binning=7,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_signal_missing_mass_binning_dict.items()
    }

    # Simple Signal Wide
    _simple_signal_adapted_wide_mms_bin_edges = (-2.0, -1.25, -0.5, 0.5, 1.25, 2.0, 2.75, 3.5, 4.25, 5.0, 6.0, 8.0, 10.0)
    _simple_signal_wide_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: _simple_signal_adapted_wide_mms_bin_edges,
        RecoModeNames.DM: _simple_signal_adapted_wide_mms_bin_edges,
        RecoModeNames.DSZ: _simple_signal_adapted_wide_mms_bin_edges,
        RecoModeNames.DSM: _simple_signal_adapted_wide_mms_bin_edges,
    }
    simple_signal_wide_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_wide,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_signal_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_signal_wide_missing_mass_binning_dict.items()
    }

    # Complex Signal
    __dss_bins: int = 8
    _complex_signal_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
        RecoModeNames.DMPi: __dss_bins,
        RecoModeNames.DSMPi: __dss_bins,
        RecoModeNames.DZPi: __dss_bins,
        RecoModeNames.DSZPi: __dss_bins,
        RecoModeNames.DZPiPlus: __dss_bins,
        RecoModeNames.DSZPiPlus: __dss_bins,
        RecoModeNames.DMPiPlus: __dss_bins,
        RecoModeNames.DSMPiPlus: __dss_bins,
    }
    complex_signal_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_signal_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _complex_signal_missing_mass_binning_dict.items()
    }

    #
    #
    # Bin Adjustment Setup for Signal
    _bin_adjust_signal_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }
    bin_adjust_signal_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_signal_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _bin_adjust_signal_missing_mass_binning_dict.items()
    }
    #
    #

    # 3D Signal
    _missing_mass_wide_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }
    three_d_signal_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_signal_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.mva_output_for_fit_var,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _missing_mass_wide_binning_dict.items()
    }

    # Simple Q2SP
    _simple_q2sb_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }
    simple_q2sb_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_sideband_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_q2sb_missing_mass_binning_dict.items()
    }

    # Simple Q2SP Wide
    _simple_q2sb_adapted_wide_mms_bin_edges = (-2.0, -1.25, -0.5, 0.5, 1.25, 2.0, 3.0, 4.0)
    _simple_q2sb_missing_mass_wide_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: _simple_q2sb_adapted_wide_mms_bin_edges,
        RecoModeNames.DM: _simple_q2sb_adapted_wide_mms_bin_edges,
        RecoModeNames.DSZ: _simple_q2sb_adapted_wide_mms_bin_edges,
        RecoModeNames.DSM: _simple_q2sb_adapted_wide_mms_bin_edges,
    }
    simple_q2sb_wide_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_wide,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_sideband_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_q2sb_missing_mass_wide_binning_dict.items()
    }

    # Five by five bin setup for test of fit with nuisances in q2-sideband
    _five_bin_q2sb_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 5,
        RecoModeNames.DM: 5,
        RecoModeNames.DSZ: 5,
        RecoModeNames.DSM: 5,
    }
    five_bin_q2sb_fit_with_stat_nuisances_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_wide,
                alternative_binning=mms_binning,
                alternative_scope=(-2.0, 3.0),
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_sideband_fit,
                alternative_binning=5,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _five_bin_q2sb_missing_mass_binning_dict.items()
    }

    # Q2SP 1D MMS
    one_d_mms_q2sb_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_q2sb_missing_mass_binning_dict.items()
    }

    # E_miss - P_miss Binning Dictionaries
    e_miss_minus_p_miss_binning_dict_signal_corr: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 22,
        RecoModeNames.DM: 20,
        RecoModeNames.DSZ: 22,
        RecoModeNames.DSM: 17,
    }
    e_miss_minus_p_miss_binning_dict_q2sb_corr: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 22,
        RecoModeNames.DM: 20,
        RecoModeNames.DSZ: 22,
        RecoModeNames.DSM: 17,
    }
    e_miss_minus_p_miss_wide_binning_dict_q2sb_corr: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }

    # Q2SP 1D E_miss - P_miss
    one_d_e_miss_minus_p_miss_q2sb_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.e_miss_minus_p_miss_res_corr_q2sb_var,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in e_miss_minus_p_miss_wide_binning_dict_q2sb_corr.items()
    }

    # Simple Mbc SP
    _simple_mbcsb_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }
    simple_mbcsb_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_sideband_fit,
                alternative_binning=None,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_mbcsb_missing_mass_binning_dict.items()
    }

    # Complex Mbc SP
    _complex_mbcsb_missing_mass_binning_dict: Dict[str, Optional[BinsInputType]] = {
        RecoModeNames.DZ: 11,
        RecoModeNames.DM: 10,
        RecoModeNames.DSZ: 11,
        RecoModeNames.DSM: 9,
    }
    complex_mbcsb_fit_var_dict: FitVarDictType = {
        reco_mode_name: (
            FitVariable(
                variable=RDStarHistVars.missing_mass_res_corr_q2sb_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
            FitVariable(
                variable=RDStarHistVars.pl_star_sideband_fit,
                alternative_binning=mms_binning,
                alternative_scope=None,
            ),
        )
        for reco_mode_name, mms_binning in _simple_mbcsb_missing_mass_binning_dict.items()
    }
