"""
Queries for old / original mc matching information.

Felix Metzner 2023
"""

import numpy as np
import pandas as pd

from typing import Tuple

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.pdg_code_collections import PDGCodeCollection as PcC


__all__ = [
    "OldMCSigQueries",
]


class OldMCSigQueries:

    required_columns: Tuple[str, ...] = (
        Cols.is_continuum_bkg_col,
        Cols.is_not_continuum_evt,
        #
        Cols.reco_lepton_truth_matched_pdg,
        Cols.reco_lepton_truth_matched_b_meson_pdg,
        #
        Cols.reco_lepton_truth_matched_fourth_daughter_pdg_match,
        Cols.reco_lepton_truth_matched_fourth_daughter_pdg_via_tau,
        #
        Cols.reco_lepton_truth_matched_b_meson_pdg,
        #
        Cols.reco_lepton_truth_matched_first_b_daugther_pdg,
        Cols.reco_lepton_truth_matched_2nd_daugther_of_1st_daugther_pdg,
        Cols.reco_lepton_truth_matched_3rd_daugther_of_1st_daugther_pdg,
        #
        Cols.reco_lepton_truth_matched_first_b_daughter_pdg_via_tau,
    )

    @staticmethod
    def is_lepton_signal(sample: pd.DataFrame) -> pd.Series:
        is_lepton = abs(sample[Cols.reco_lepton_truth_matched_pdg]).isin(PcC.light_lepton_pdg_ids)
        is_from_b = abs(sample[Cols.reco_lepton_truth_matched_b_meson_pdg]).isin(PcC.b_meson_pdg_ids)
        has_no_forth_daughter = sample[Cols.reco_lepton_truth_matched_fourth_daughter_pdg_match].isin(
            PcC.valid_forth_daughter
        )

        return is_lepton & is_from_b & has_no_forth_daughter

    @staticmethod
    def is_tau_to_lepton_signal(sample: pd.DataFrame) -> pd.Series:
        is_lepton = abs(sample[Cols.reco_lepton_truth_matched_pdg]).isin(PcC.light_lepton_pdg_ids)
        is_from_tau = abs(sample[Cols.reco_lepton_truth_matched_b_meson_pdg]) == PcC.tau_pdg_id
        is_from_b = abs(sample[Cols.reco_lepton_truth_matched_b_meson_pdg_via_tau]).isin(PcC.b_meson_pdg_ids)
        has_no_forth_daughter = sample[Cols.reco_lepton_truth_matched_fourth_daughter_pdg_via_tau].isin(
            PcC.valid_forth_daughter
        )

        return is_lepton & is_from_tau & is_from_b & has_no_forth_daughter

    @staticmethod
    def is_tau_to_hadron_bkg(sample: pd.DataFrame) -> pd.Series:
        is_not_lepton = ~abs(sample[Cols.reco_lepton_truth_matched_pdg]).isin(PcC.light_lepton_pdg_ids)
        is_from_tau = abs(sample[Cols.reco_lepton_truth_matched_b_meson_pdg]) == PcC.tau_pdg_id
        is_from_b = abs(sample[Cols.reco_lepton_truth_matched_b_meson_pdg_via_tau]).isin(PcC.b_meson_pdg_ids)
        has_no_forth_daughter = sample[Cols.reco_lepton_truth_matched_fourth_daughter_pdg_via_tau].isin(
            PcC.valid_forth_daughter
        )

        return is_not_lepton & is_from_tau & is_from_b & has_no_forth_daughter

    @staticmethod
    def b_to_dstst_l_signal(sample: pd.DataFrame) -> pd.Series:
        is_lepton = OldMCSigQueries.is_lepton_signal(sample=sample)

        is_b_to_dstst = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daugther_pdg]).isin(PcC.dstst_pdg_ids)
        is_dstst_with_pi_zero = (
            np.abs(sample[Cols.reco_lepton_truth_matched_2nd_daugther_of_1st_daugther_pdg]) == PcC.pi_zero_pdg_id
        )
        dstst_has_no_third_daughter = sample[Cols.reco_lepton_truth_matched_3rd_daugther_of_1st_daugther_pdg].isin(
            PcC.valid_forth_daughter
        )

        is_not_continuum = sample[Cols.is_not_continuum_evt] == 1.0

        return is_lepton & is_b_to_dstst & is_dstst_with_pi_zero & dstst_has_no_third_daughter & is_not_continuum

    @staticmethod
    def b_to_dstst_l_bkg(sample: pd.DataFrame) -> pd.Series:
        is_lepton = OldMCSigQueries.is_lepton_signal(sample=sample)
        is_b_to_dstst = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daugther_pdg]).isin(PcC.dstst_pdg_ids)

        is_dstst_with_pi_charged = (
            np.abs(sample[Cols.reco_lepton_truth_matched_2nd_daugther_of_1st_daugther_pdg]) == PcC.pi_charged_pdg_id
        )
        dstst_has_no_third_daughter = sample[Cols.reco_lepton_truth_matched_3rd_daugther_of_1st_daugther_pdg].isin(
            PcC.valid_forth_daughter
        )

        is_not_continuum = sample[Cols.is_not_continuum_evt] == 1.0

        return is_lepton & is_b_to_dstst & is_dstst_with_pi_charged & dstst_has_no_third_daughter & is_not_continuum

    @staticmethod
    def b_to_d_l_signal(sample: pd.DataFrame) -> pd.Series:

        is_lepton = OldMCSigQueries.is_lepton_signal(sample=sample)
        is_b_to_d = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daugther_pdg]).isin(PcC.d_pdg_ids)
        is_not_continuum = sample[Cols.is_not_continuum_evt] == 1.0

        return is_lepton & is_b_to_d & is_not_continuum

    @staticmethod
    def b_to_dst_l_signal(sample: pd.DataFrame) -> pd.Series:
        is_lepton = OldMCSigQueries.is_lepton_signal(sample=sample)
        is_b_to_dst = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daugther_pdg]).isin(PcC.dst_pdg_ids)
        is_not_continuum: pd.Series = sample[Cols.is_not_continuum_evt] == 1.0

        return is_lepton & is_b_to_dst & is_not_continuum

    @staticmethod
    def b_to_d_tau_signal(sample: pd.DataFrame) -> pd.Series:
        is_tau_to_lepton = OldMCSigQueries.is_tau_to_lepton_signal(sample=sample)
        is_b_to_d = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daughter_pdg_via_tau]).isin(PcC.d_pdg_ids)
        is_not_continuum: pd.Series = sample[Cols.is_not_continuum_evt] == 1.0

        return is_tau_to_lepton & is_b_to_d & is_not_continuum

    @staticmethod
    def b_to_dst_tau_signal(sample: pd.DataFrame) -> pd.Series:
        is_tau_to_lepton = OldMCSigQueries.is_tau_to_lepton_signal(sample=sample)
        is_b_to_dst = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daughter_pdg_via_tau]).isin(PcC.dst_pdg_ids)
        is_not_continuum: pd.Series = sample[Cols.is_not_continuum_evt] == 1.0

        return is_tau_to_lepton & is_b_to_dst & is_not_continuum

    @staticmethod
    def b_to_pi_l_nu_signal(sample: pd.DataFrame) -> pd.Series:
        is_lepton = OldMCSigQueries.is_lepton_signal(sample=sample)
        is_b_to_pi = np.abs(sample[Cols.reco_lepton_truth_matched_first_b_daugther_pdg]).isin(PcC.pion_pdg_ids)
        is_not_continuum: pd.Series = sample[Cols.is_not_continuum_evt] == 1.0

        return is_lepton & is_b_to_pi & is_not_continuum

    @staticmethod
    def hadronic_bkg(sample: pd.DataFrame) -> pd.Series:
        fake_lepton = ~np.abs(sample[Cols.reco_lepton_truth_matched_pdg]).isin(PcC.light_lepton_pdg_ids)
        is_not_continuum: pd.Series = sample[Cols.is_not_continuum_evt] == 1.0

        return fake_lepton & is_not_continuum

    @staticmethod
    def continuum_bkg(sample: pd.DataFrame) -> pd.Series:
        return sample[Cols.is_continuum_evt] == 1.0

    @staticmethod
    def bbar_bkg(sample: pd.DataFrame) -> pd.Series:
        not_dl: pd.Series = ~OldMCSigQueries.b_to_d_l_signal(sample=sample)
        not_dstl: pd.Series = ~OldMCSigQueries.b_to_dst_l_signal(sample=sample)
        not_dststl: pd.Series = ~OldMCSigQueries.b_to_dstst_l_signal(sample=sample)
        not_dststl_bkg: pd.Series = ~OldMCSigQueries.b_to_dstst_l_bkg(sample=sample)
        not_dtau: pd.Series = ~OldMCSigQueries.b_to_d_tau_signal(sample=sample)
        not_dsttau: pd.Series = ~OldMCSigQueries.b_to_dst_tau_signal(sample=sample)
        not_pulnu: pd.Series = ~OldMCSigQueries.b_to_pi_l_nu_signal(sample=sample)
        not_hadronic_bkg: pd.Series = ~OldMCSigQueries.hadronic_bkg(sample=sample)
        not_continuum: pd.Series = ~OldMCSigQueries.continuum_bkg(sample=sample)
        return (
            not_dl
            & not_dstl
            & not_dststl
            & not_dststl_bkg
            & not_dtau
            & not_dsttau
            & not_pulnu
            & not_hadronic_bkg
            & not_continuum
        )
