"""
Definition of observables describing the observed decay process

Felix Metzner 2021
"""

import numpy as np
import pandas as pd

from typing import Union, Tuple, List, Dict, Callable, Type

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.observables.offline_observable_definition_base import BinnedOfflineObservableBase

from rdstar.online_analysis.decay_channels import Dstar_decay_label_to_decay_id, D_decay_label_to_decay_id
from rdstar.online_analysis.decay_channels import B_Dlpi0_decay_label_to_decay_id, B_Dlpim_decay_label_to_decay_id


__all__ = [
    "BSigNeutralCountVariable",
    "BSigTrackCountVariable",
    "BTagNeutralCountVariable",
    "BTagTrackCountVariable",
    "DecayProcessVariables",
]


class BSigNeutralCountVariable(BinnedOfflineObservableBase):
    new_column_name = "sig_neutral_particle_count"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=5,  # Can be overwritten by binning instance below.
        scope=(-0.5, 4.5),  # Resulting from calculation method and values of neutrals_mapping dict.
        var_name=LatexObsNames.decay_info_number_of_neutrals_in_b_sig,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    binning = Binning(
        bins=(-0.5, 0.5, 1.5, 4.5),
        dimensions=1,
        scope=hist_var.scope,
        log_scale=False,
    )

    b_sig_reco_mode_id = "b_sig_reco_mode_id"  # type: str
    d_or_d_st_meson_reco_mode_id = "d_or_d_st_meson_reco_mode_id"  # type: str
    d_from_d_st_meson_reco_mode_id = "d_from_d_st_meson_reco_mode_id"  # type: str

    required_variables = {
        b_sig_reco_mode_id: Cols.decay_mode_column_name,
        d_or_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc",
        d_from_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc__bc",
    }  # type: Dict[str, str]

    neutrals_mapping = {
        21: 1,  # r"\mathrm{D}$^{*0} \rightarrow \mathrm{D}^{0} \pi^{0}$"
        22: 1,  # r"\mathrm{D}$^{*0} \rightarrow \mathrm{D}^{0} \gamma$"
        23: 0,  # r"\mathrm{D}$^{*+} \rightarrow \mathrm{D}^{0} \pi^{+}$"
        24: 0,  # r"\mathrm{D}$^{*+} \rightarrow \mathrm{D}^{+} \pi^{0}$"
        31: 0,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+}$"
        32: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{0}$"
        33: 0,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{+} \pi^{-}$"
        34: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+}$"
        35: 2,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{0}$"
        36: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{+} \pi^{-}$"
        37: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \mathrm{K}^{+}$"
        38: 0,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{+} \mathrm{K}^{-} \pi^{+}$"
        41: 0,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+}$"
        42: 1,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{0}$"
        43: 0,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{-}$"
        44: 1,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{-} \pi^{0}$"
        45: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{0}$"
        46: 1,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{-}$"
        47: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{-} \pi^{0}$"
        48: 0,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \mathrm{K}^{+}$"
        -999: -999,  # Must be added to not cause problems during calculation with np.vectorize!
    }  # type: Dict[int, int]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_st_reco_ids = tuple(Dstar_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        b_sig_with_neutral_pion_reco_ids = tuple(B_Dlpi0_decay_label_to_decay_id.values())  # type: Tuple[int, ...]

        is_d = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_reco_ids)
        is_from_dst = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_st_reco_ids)
        is_d_from_dst = is_from_dst & (df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].isin(d_reco_ids))

        assert np.sum(is_d & is_d_from_dst) == 0, np.sum(is_d & is_d_from_dst)
        assert np.sum(is_d | is_d_from_dst) == len(np.unique(df.index)), (
            np.sum(is_d | is_d_from_dst),
            len(np.unique(df.index)),
        )

        neutrals_count_array = np.select(
            condlist=[is_d.values, is_d_from_dst.values],
            choicelist=[
                np.vectorize(cls.neutrals_mapping.get)(
                    df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values
                ),
                np.vectorize(cls.neutrals_mapping.get)(
                    df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values
                )
                + np.vectorize(cls.neutrals_mapping.get)(
                    df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].values
                ),
            ],
            default=-1,
        )

        # Adding pion from B_sig -> D** (-> D(*) pi0) l nu Decay
        is_b_w_neutral_pi = df[cls.required_variables[cls.b_sig_reco_mode_id]].isin(b_sig_with_neutral_pion_reco_ids)
        neutrals_count_array += is_b_w_neutral_pi

        assert np.sum(neutrals_count_array == -1) == 0, np.sum(neutrals_count_array == -1)

        if check_range:
            assert np.all(neutrals_count_array > cls.hist_var.scope[0]), (
                np.unique(neutrals_count_array),
                cls.hist_var.scope,
            )
            assert np.all(neutrals_count_array < cls.hist_var.scope[1]), (
                np.unique(neutrals_count_array),
                cls.hist_var.scope,
            )

        return neutrals_count_array


class BSigTrackCountVariable(BinnedOfflineObservableBase):
    new_column_name = "sig_track_count"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=8,  # Can be overwritten by binning instance below.
        scope=(0.5, 8.5),  # 0 Zero tracks is not possible, because of the signal B side lepton and max #tracks is 8.
        var_name=LatexObsNames.decay_info_number_of_tracks_in_b_sig,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    binning = Binning(
        bins=(0.5, 3.5, 4.5, 5.5, 8.5),
        dimensions=1,
        scope=hist_var.scope,
        log_scale=False,
    )

    b_sig_reco_mode_id = "b_sig_reco_mode_id"  # type: str
    d_or_d_st_meson_reco_mode_id = "d_or_d_st_meson_reco_mode_id"  # type: str
    d_from_d_st_meson_reco_mode_id = "d_from_d_st_meson_reco_mode_id"  # type: str

    required_variables = {
        b_sig_reco_mode_id: Cols.decay_mode_column_name,
        d_or_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc",
        d_from_d_st_meson_reco_mode_id: "daughter__bo1__cm__spdaughter__bo0__cm__spdaughter__bo0__cm__spextraInfo__bodecayModeID__bc__bc__bc__bc",
    }  # type: Dict[str, str]

    tracks_mapping = {
        21: 0,  # r"\mathrm{D}$^{*0} \rightarrow \mathrm{D}^{0} \pi^{0}$"
        22: 0,  # r"\mathrm{D}$^{*0} \rightarrow \mathrm{D}^{0} \gamma$"
        23: 1,  # r"\mathrm{D}$^{*+} \rightarrow \mathrm{D}^{0} \pi^{+}$"
        24: 0,  # r"\mathrm{D}$^{*+} \rightarrow \mathrm{D}^{+} \pi^{0}$"
        31: 3,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+}$"
        32: 3,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{0}$"
        33: 5,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{+} \pi^{-}$"
        34: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+}$"
        35: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{0}$"
        36: 3,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{+} \pi^{-}$"
        37: 1,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}_{\mathrm{S}0} \mathrm{K}^{+}$"
        38: 3,  # r"\mathrm{D}$^{+} \rightarrow \mathrm{K}^{+} \mathrm{K}^{-} \pi^{+}$"
        41: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+}$"
        42: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{0}$"
        43: 4,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{-}$"
        44: 4,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \pi^{+} \pi^{+} \pi^{-} \pi^{0}$"
        45: 0,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{0}$"
        46: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{-}$"
        47: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}_{\mathrm{S}0} \pi^{+} \pi^{-} \pi^{0}$"
        48: 2,  # r"\mathrm{D}$^{0} \rightarrow \mathrm{K}^{-} \mathrm{K}^{+}$"
        -999: -999,  # Must be added to not cause problems during calculation with np.vectorize!
    }  # type: Dict[int, int]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        d_reco_ids = tuple(D_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        d_st_reco_ids = tuple(Dstar_decay_label_to_decay_id.values())  # type: Tuple[int, ...]
        b_sig_with_charged_pion_reco_ids = tuple(B_Dlpim_decay_label_to_decay_id.values())  # type: Tuple[int, ...]

        is_d = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_reco_ids)
        is_from_dst = df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].isin(d_st_reco_ids)
        is_d_from_dst = is_from_dst & (df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].isin(d_reco_ids))

        assert np.sum(is_d & is_d_from_dst) == 0, np.sum(is_d & is_d_from_dst)
        assert np.sum(is_d | is_d_from_dst) == len(np.unique(df.index)), (
            np.sum(is_d | is_d_from_dst),
            len(np.unique(df.index)),
        )

        tracks_count_array = np.select(
            condlist=[is_d.values, is_d_from_dst.values],
            choicelist=[
                np.vectorize(cls.tracks_mapping.get)(df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values),
                np.vectorize(cls.tracks_mapping.get)(df[cls.required_variables[cls.d_or_d_st_meson_reco_mode_id]].values)
                + np.vectorize(cls.tracks_mapping.get)(
                    df[cls.required_variables[cls.d_from_d_st_meson_reco_mode_id]].values
                ),
            ],
            default=-1,
        )

        # Increasing track count by one to account for lepton track, which is always present on the signal side.
        tracks_count_array += 1

        # Adding pion from B_sig -> D** (-> D(*) pi+) l nu Decay
        is_b_w_charged_pi = df[cls.required_variables[cls.b_sig_reco_mode_id]].isin(b_sig_with_charged_pion_reco_ids)
        tracks_count_array += is_b_w_charged_pi

        assert np.sum(tracks_count_array == -1) == 0, np.sum(tracks_count_array == -1)
        if check_range:
            assert np.all(tracks_count_array > cls.hist_var.scope[0]), (np.unique(tracks_count_array), cls.hist_var.scope)
            assert np.all(tracks_count_array < cls.hist_var.scope[1]), (np.unique(tracks_count_array), cls.hist_var.scope)

        return tracks_count_array


class BTagNeutralCountVariable(BinnedOfflineObservableBase):
    new_column_name = "tag_neutral_particle_count"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=27,  # Can be overwritten by binning instance below.
        scope=(-0.5, 26.5),
        var_name=LatexObsNames.decay_info_number_of_neutrals_in_b_tag,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    binning = Binning(
        bins=(-0.5, 2.5, 4.5, 5.5, 6.5, 8.5, 26.5),
        dimensions=1,
        scope=hist_var.scope,
        log_scale=False,
    )

    event_neutral_count = "event_neutral_count"  # type: str

    required_variables = {
        event_neutral_count: "nCleanedECLClusters__bogoodBelleGamma__sp__eq__eq__sp1__spand__spclusterTrackMatch__sp__eq__eq__sp0__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        event_neutral_count_array = df[cls.required_variables[cls.event_neutral_count]].values  # type: np.ndarray
        b_sig_neutral_count_array = BSigNeutralCountVariable.calculate_new_column_values(df=df)  # type: np.ndarray

        assert len(b_sig_neutral_count_array) == len(event_neutral_count_array), (
            len(b_sig_neutral_count_array),
            len(event_neutral_count_array),
        )

        neutral_count_array = event_neutral_count_array - b_sig_neutral_count_array  # type: np.ndarray
        neutral_count_array[neutral_count_array < 0] = 0  # Resetting values below 0, caused by B sig counting method

        if check_range:
            assert np.all(neutral_count_array > cls.hist_var.scope[0]), (
                np.unique(neutral_count_array),
                cls.hist_var.scope,
            )
            assert np.all(neutral_count_array < cls.hist_var.scope[1]), (
                np.unique(neutral_count_array),
                cls.hist_var.scope,
            )

        return neutral_count_array


class BTagTrackCountVariable(BinnedOfflineObservableBase):
    new_column_name = "tag_track_count"  # type: str

    hist_var = HistVariable(
        df_label=new_column_name,
        n_bins=20,  # Can be overwritten by binning instance below.
        scope=(-0.5, 19.5),
        var_name=LatexObsNames.decay_info_number_of_tracks_in_b_tag,
        unit=LatexUnits.no_unit,
        use_log_scale=False,
    )  # type: HistVariable

    binning = Binning(
        bins=(-0.5, 3.5, 4.5, 5.5, 6.5, 19.5),
        dimensions=1,
        scope=hist_var.scope,
        log_scale=False,
    )

    event_track_count = "event_track_count"  # type: str

    required_variables = {
        event_track_count: "nCleanedTracks__bodr__sp__st__sp2__spand__spabs__bodz__bc__sp__st__sp4__bc",
    }  # type: Dict[str, str]

    @classmethod
    def calculate_new_column_values(cls, df: pd.DataFrame, check_range: bool = True) -> Union[pd.Series, np.array]:
        event_track_count_array = df[cls.required_variables[cls.event_track_count]].values  # type: np.ndarray
        b_sig_track_count_array = BSigTrackCountVariable.calculate_new_column_values(df=df)  # type: np.ndarray

        assert len(b_sig_track_count_array) == len(event_track_count_array), (
            len(b_sig_track_count_array),
            len(event_track_count_array),
        )

        tracks_count_array = event_track_count_array - b_sig_track_count_array  # type: np.ndarray

        assert np.all(tracks_count_array >= 0), (
            np.unique(tracks_count_array),
            event_track_count_array[tracks_count_array < 0],
            b_sig_track_count_array[tracks_count_array < 0],
            [
                (v, np.unique(df[v].values[tracks_count_array < 0]))
                for v in BSigTrackCountVariable.required_variables.values()
            ],
        )

        if check_range:
            assert np.all(tracks_count_array > cls.hist_var.scope[0]), (np.unique(tracks_count_array), cls.hist_var.scope)
            assert np.all(tracks_count_array < cls.hist_var.scope[1]), (np.unique(tracks_count_array), cls.hist_var.scope)

        return tracks_count_array


class DecayProcessVariables:

    definitions = {
        BSigNeutralCountVariable.get_new_column_name(): BSigNeutralCountVariable,
        BSigTrackCountVariable.get_new_column_name(): BSigTrackCountVariable,
        BTagNeutralCountVariable.get_new_column_name(): BTagNeutralCountVariable,
        BTagTrackCountVariable.get_new_column_name(): BTagTrackCountVariable,
    }  # type: Dict[str, Type[BinnedOfflineObservableBase]]

    required_variables = tuple(
        [c for d in definitions.values() for c in d.get_required_columns()]
    )  # type: Tuple[str, ...]

    @staticmethod
    def add_decay_process_information_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in DecayProcessVariables.get_required_columns_list()), [
            c for c in DecayProcessVariables.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in DecayProcessVariables.get_new_columns_list()), [
            c for c in DecayProcessVariables.get_new_columns_list() if c in df.columns
        ]

        for new_col, definition_class in DecayProcessVariables.definitions.items():
            new_values = definition_class.calculate_new_column_values(df=df, check_range=check_range)
            assert len(new_values) == len(df), (len(new_values), len(df), new_col)
            df.loc[:, new_col] = new_values

            assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_definition_function(
        name: str,
        fail_when_exists: bool = True,
    ) -> Callable:
        def definition_function(
            df: pd.DataFrame,
            check_range: bool = True,
            accept_existing_column: bool = False,
        ) -> None:
            definition = DecayProcessVariables.definitions[name]  # type: Type[BinnedOfflineObservableBase]
            new_col_name = definition.get_new_column_name()

            if fail_when_exists and not accept_existing_column:
                assert new_col_name not in df.columns, new_col_name
            else:
                if new_col_name in df.columns:
                    return

            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df, check_range=check_range)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(DecayProcessVariables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(DecayProcessVariables.definitions.keys())

    @staticmethod
    def get_decay_info_definition(name: str) -> Type[BinnedOfflineObservableBase]:
        return DecayProcessVariables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(DecayProcessVariables.definitions.keys())
