"""
Generator level MC matching observables.

Felix Metzner 2023
"""

import numpy as np
import pandas as pd

from abc import ABC
from typing import Union, Tuple, List, Dict, Callable, Type, Optional

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.offline_observable_definition_base import OfflineObservableBase

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCMatchingInfo
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_utilities import get_gen_b_id_col_name


__all__ = [
    "MainB1GenIDsObservable",
    "MainB2GenIDsObservable",
    "GenIDObservables",
]


logger = log_to_default(name=__name__)


class GenIDsObservableBaseClass(OfflineObservableBase, ABC):
    b_index: int = 0
    _gen_ids_info_class: Optional[Type[GenMCMatchingInfo]] = None

    @classmethod
    def get_gen_ids_info_class(cls) -> Type[GenMCMatchingInfo]:
        assert cls._gen_ids_info_class is not None
        return cls._gen_ids_info_class

    @classmethod
    def get_new_column_name(cls) -> str:
        col_name: str = get_gen_b_id_col_name(b_index=cls.b_index)
        assert col_name in (
            cls.get_gen_ids_info_class().gen_b1_id_col_name,
            cls.get_gen_ids_info_class().gen_b2_id_col_name,
        ), col_name
        return col_name

    @classmethod
    def get_required_variables(cls) -> Tuple[str, ...]:
        return cls.get_gen_ids_info_class().required_columns

    @classmethod
    def get_gen_id_mapping(cls) -> Dict[str, MatchQueryInfo]:
        return cls.get_gen_ids_info_class().get_gen_id_mapping()

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:
        assert cls.b_index in (1, 2), (cls.__name__, cls.b_index)
        assert f"B{cls.b_index}" in cls.__name__, (cls.__name__, cls.b_index)

        new_gen_id_values: np.array = np.full(shape=len(df), fill_value=np.nan)

        n_rows: int = len(df.index)
        n_matched_total: int = 0
        logger.info(f"Number of candidates: {n_rows}")

        gen_info_cls = cls.get_gen_ids_info_class()

        for key, gen_id_info in cls.get_gen_id_mapping().items():
            logger.info(f"Processing GenID {gen_id_info.id} query {key} for {cls.__name__} ({gen_info_cls.__name__})...")

            query_fcn: Callable = gen_id_info.query
            gen_id_mask: np.ndarray = query_fcn(df=df, b_index=cls.b_index - 1)

            n_matching: int = int(np.sum(gen_id_mask))
            n_matched_total += n_matching
            n_remaining: int = n_rows - n_matched_total
            logger.info(
                f"Number of matching candidates: {n_matching} "
                f"(covered {n_matched_total} of {n_rows} = {n_matched_total / n_rows:1.2f}), "
                f"{n_remaining} rows remaining."
            )

            if not np.all(np.isnan(new_gen_id_values[gen_id_mask])):
                _duplicate_ids: List[int] = list(
                    np.unique(new_gen_id_values[gen_id_mask & (~np.isnan(new_gen_id_values))])
                )
                raise RuntimeError(
                    f"Overwriting Gen ID values of events with the GenID(s) {_duplicate_ids} "
                    f"with the current GenID {gen_id_info.id} for {key} events!"
                )

            new_gen_id_values[gen_id_mask] = gen_id_info.id

        check = np.sum(np.isnan(new_gen_id_values))
        if check != 0:
            raise RuntimeError(f"Gen ID queries do not cover all events. Number of Rows w/o ID: {check}")

        _all_gen_ids: Tuple[int, ...] = cls.get_gen_ids_info_class().get_all_gen_ids()
        assert all(v in _all_gen_ids for v in np.unique(new_gen_id_values)), (
            [v for v in np.unique(new_gen_id_values) if v not in _all_gen_ids],
        )

        return new_gen_id_values

    @classmethod
    def add_gen_id_to_df(
        cls,
        df: pd.DataFrame,
    ) -> None:
        _gen_id_column_name: str = cls.get_new_column_name()

        if _gen_id_column_name in df.columns:
            raise RuntimeError(
                f"New Gen ID column '{_gen_id_column_name}' for definition class '{cls.__name__}' "
                f"is already in the data frame!"
            )

        new_values = cls.calculate_new_column_values(df=df)
        assert len(new_values) == len(df), (len(new_values), len(df), cls.__name__, _gen_id_column_name)
        df.loc[:, _gen_id_column_name] = new_values

        assert not pd.isnull(df[_gen_id_column_name]).any(), (cls.__name__, _gen_id_column_name)


class MainB1GenIDsObservable(GenIDsObservableBaseClass):
    b_index: int = 1
    _gen_ids_info_class: Optional[Type[GenMCMatchingInfo]] = GenMCMatchingInfo


class MainB2GenIDsObservable(GenIDsObservableBaseClass):
    b_index: int = 2
    _gen_ids_info_class: Optional[Type[GenMCMatchingInfo]] = GenMCMatchingInfo


class GenIDObservables:

    observable_clss: Tuple[Type[GenIDsObservableBaseClass], ...] = (
        # Main Gen ID Definition
        MainB1GenIDsObservable,
        MainB2GenIDsObservable,
    )

    obs_cls_dict: Dict[str, Type[GenIDsObservableBaseClass]] = {c.__name__: c for c in observable_clss}

    definitions: Dict[str, Type[GenIDsObservableBaseClass]] = {c.get_new_column_name(): c for c in observable_clss}

    required_variables: Tuple[str, ...] = tuple(set([c for d in definitions.values() for c in d.get_required_columns()]))

    @staticmethod
    def add_gen_id_observables_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in GenIDObservables.get_required_columns_list()), [
            c for c in GenIDObservables.get_required_columns_list() if c not in df.columns
        ]

        new_cols: List[str] = GenIDObservables.get_new_columns_list()
        assert all(c not in df.columns for c in new_cols), [c for c in new_cols if c in df.columns]

        for new_col, definition_class in GenIDObservables.definitions.items():
            definition_class.add_gen_id_to_df(df=df)

    @staticmethod
    def get_definition_function(
        name: str,
        fail_when_exists: bool = True,
    ) -> Callable:
        def definition_function(
            df: pd.DataFrame,
            check_range: bool = True,
            accept_existing_column: bool = False,
        ) -> None:
            definition: Type[OfflineObservableBase] = GenIDObservables.definitions[name]
            new_col_name: str = definition.get_new_column_name()

            if fail_when_exists and not accept_existing_column:
                assert new_col_name not in df.columns, new_col_name
            else:
                if new_col_name in df.columns:
                    return

            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(GenIDObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(GenIDObservables.definitions.keys())

    @staticmethod
    def get_gen_id_obs_definition(name: str) -> Type[GenIDsObservableBaseClass]:
        return GenIDObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(GenIDObservables.definitions.keys())

    @staticmethod
    def run_checks() -> None:
        assert all(any(b_i in c.__name__ for b_i in ("B1", "B2")) for c in GenIDObservables.observable_clss), [
            c.__name__ for c in GenIDObservables.observable_clss if not any(b_i in c.__name__ for b_i in ("B1", "B2"))
        ]
        assert all(c.b_index == (1 if "B1" in c.__name__ else 2) for c in GenIDObservables.observable_clss), {
            c.__name__: c.b_index for c in GenIDObservables.observable_clss
        }

        assert all(c.b_index in (1, 2) for c in GenIDObservables.observable_clss), [
            (c.__name__, c.b_index) for c in GenIDObservables.observable_clss if c.b_index not in (1, 2)
        ]
        assert all(
            c.get_gen_ids_info_class()
            == GenIDObservables.obs_cls_dict[c.__name__.replace("B1", "B2")].get_gen_ids_info_class()
            for c in GenIDObservables.observable_clss
            if c.b_index == 1
        ), [
            (c.__name__, GenIDObservables.obs_cls_dict[c.__name__.replace("B1", "B2")].__name__)
            for c in GenIDObservables.observable_clss
            if (
                c.get_gen_ids_info_class()
                != GenIDObservables.obs_cls_dict[c.__name__.replace("B1", "B2")].get_gen_ids_info_class()
            )
        ]


if __name__ == "__main__":
    print(GenIDObservables.obs_cls_dict)
