import numpy as np

from uncertainties import ufloat, UFloat, correlated_values
from typing import Union, Optional, Tuple, Dict, NamedTuple, Any

from templatefitter.minimizer import MinimizeResult
from templatefitter.fit_model.parameter_handler import ParameterHandler

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import (
    EfficiencyInfo,
    EfficiencyNumbersContainer,
)
from rdstar.offline_analysis.efficiency_calculation.efficiency_systematics_calculator import EfficiencySysInfoContainer


__all__ = [
    "UpDownUValue",
    "RatioCalculator",
]


class FitYieldInfo(NamedTuple):
    name: str
    id: int
    result: UFloat


class RatioInfoContainer(NamedTuple):
    numerator: FitYieldInfo
    denominator: FitYieldInfo


class EffSysInfoContainer(NamedTuple):
    numerator_eff_with_all_sys: UFloat
    denominator_eff_with_all_sys: UFloat
    numerator_eff_sys_dict: Dict[str, UFloat]
    denominator_eff_sys_dict: Dict[str, UFloat]
    sys_cov_matrix_dict: Dict[str, np.ndarray]


class YieldsInfoContainer(NamedTuple):
    nominal_values: np.ndarray
    errors: np.ndarray
    covariance_matrix: np.ndarray

    @staticmethod
    def from_minimize_result(minimize_result: MinimizeResult) -> "YieldsInfoContainer":
        expected_shape = len([i for i in minimize_result.params.fixed_params if not i])  # type: int
        assert expected_shape == minimize_result.params.num_params_not_fixed, (
            expected_shape,
            minimize_result.params.num_params_not_fixed,
        )

        cov_matrix = minimize_result.params.covariance  # type: np.ndarray

        assert len(cov_matrix.shape) == 2, cov_matrix.shape
        assert cov_matrix.shape[0] == cov_matrix.shape[1], cov_matrix.shape
        assert cov_matrix.shape[0] == expected_shape, (cov_matrix.shape, expected_shape)

        unfixed_yield_param_indices = tuple(
            [
                index
                for index, (pf, pt) in enumerate(
                    zip(minimize_result.params.fixed_params, minimize_result.params.parameter_types)
                )
                if not pf and pt == ParameterHandler.yield_parameter_type
            ]
        )  # type: Tuple[int, ...]
        num_unfixed_yield_params = len(unfixed_yield_param_indices)  # type: int
        assert num_unfixed_yield_params <= expected_shape, (num_unfixed_yield_params, expected_shape)

        nom_values = minimize_result.params.values[list(unfixed_yield_param_indices)]  # type: np.ndarray
        error_values = minimize_result.params.errors[list(unfixed_yield_param_indices)]  # type: np.ndarray
        sub_cov_matrix = minimize_result.params.covariance[
            np.ix_(list(unfixed_yield_param_indices), list(unfixed_yield_param_indices))
        ]  # type: np.ndarray

        assert len(nom_values.shape) == len(error_values.shape) == 1, (nom_values.shape, error_values.shape)
        assert nom_values.shape[0] == error_values.shape[0], (nom_values.shape, error_values.shape)
        assert nom_values.shape[0] == sub_cov_matrix.shape[0], (nom_values.shape, sub_cov_matrix.shape)

        assert len(sub_cov_matrix.shape) == 2, sub_cov_matrix.shape
        assert sub_cov_matrix.shape[0] == sub_cov_matrix.shape[1], sub_cov_matrix.shape
        assert sub_cov_matrix.shape[0] == num_unfixed_yield_params, (sub_cov_matrix.shape, num_unfixed_yield_params)

        _instance = YieldsInfoContainer(
            nominal_values=nom_values,
            errors=error_values,
            covariance_matrix=sub_cov_matrix,
        )
        return _instance


class UpDownUValue(NamedTuple):
    nominal_value: float
    std_dev_up: float
    std_dev_down: float
    std_dev_all_sys: float
    systematic_uncertainties: Dict[str, float]

    @property
    def ufloat_std_dev_up(self) -> UFloat:
        return ufloat(nominal_value=self.nominal_value, std_dev=self.std_dev_up)

    @property
    def ufloat_std_dev_down(self) -> UFloat:
        return ufloat(nominal_value=self.nominal_value, std_dev=self.std_dev_down)

    def get_sys_uncert_for(self, sys_name: str) -> float:
        return self.systematic_uncertainties[sys_name]

    @property
    def as_dict(self) -> Dict[str, Union[float, Dict[str, float]]]:
        return {
            "nominal_value": self.nominal_value,
            "std_dev_up": self.std_dev_up,
            "std_dev_down": self.std_dev_down,
            "std_dev_all_sys": self.std_dev_all_sys,
            "systematic_uncertainties": self.systematic_uncertainties,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Union[float, Dict[str, float]]]) -> "UpDownUValue":
        assert len(input_dict) == 5, input_dict
        assert all(isinstance(k, str) for k in input_dict.keys()), input_dict
        assert all(
            isinstance(v, dict) if k == "systematic_uncertainties" else isinstance(v, float)
            for k, v in input_dict.items()
        ), input_dict
        _input_dict = input_dict  # type: Dict[str, Any]
        return UpDownUValue(**_input_dict)


class RatioCalculator:
    def __init__(
        self,
        fit_case: FitCase,
        fit_result: MinimizeResult,
        numerator_component_key: str,
        denominator_component_key: str,
        efficiency_dict: EfficiencyNumbersContainer,
        efficiencies_sys_container: EfficiencySysInfoContainer,
        factor: Optional[float] = None,
    ) -> None:

        self._fit_case = fit_case  # type: FitCase
        self._fit_result = fit_result  # type: MinimizeResult
        self._efficiency_dict = efficiency_dict  # type: EfficiencyNumbersContainer
        self._efficiencies_sys_container = efficiencies_sys_container  # type: EfficiencySysInfoContainer

        self._ratio_factor = 1.0 if factor is None else factor  # type: float

        assert numerator_component_key in self.fit_case.all_components, (
            numerator_component_key,
            self.fit_case.all_components,
        )
        assert denominator_component_key in self.fit_case.all_components, (
            denominator_component_key,
            self.fit_case.all_components,
        )

        self._numerator_component_key = numerator_component_key  # type: str
        self._denominator_component_key = denominator_component_key  # type: str

        self.numerator_yield_parameter_name = FitModelCreator.get_yield_parameter_name(
            component_name=self._numerator_component_key
        )  # type: str
        self.denominator_yield_parameter_name = FitModelCreator.get_yield_parameter_name(
            component_name=self._denominator_component_key
        )  # type: str

        self._ratio_info_container = self._get_ratio_info_container()  # type: RatioInfoContainer
        self._all_yields = YieldsInfoContainer.from_minimize_result(
            minimize_result=self.fit_result
        )  # type: YieldsInfoContainer

        self._eff_dict = dict()  # type: Dict[str, UpDownUValue]
        self.all_mode_ids = tuple(
            [d_id for _, d_ids in self.fit_case.reco_mode_infos for d_id in d_ids]
        )  # type: Tuple[int, ...]

        self._yield_ratios = None  # type: Optional[Tuple[np.ndarray, np.ndarray]]
        self._efficiency_corrected_ratio = None  # type: Optional[UpDownUValue]

    def calculate_efficiency_corrected_ratio(self) -> UpDownUValue:
        numerator_id = self._ratio_info_container.numerator.id  # type: int

        bare_ratios, bare_ratio_errors = self.yield_ratios  # type: Tuple[np.ndarray, np.ndarray]
        bare_ratio = bare_ratios[numerator_id]  # type: float
        bare_ratio_error = bare_ratio_errors[numerator_id]  # type: float
        bare_ratio_ufloat = ufloat(nominal_value=bare_ratio, std_dev=bare_ratio_error)  # type: UFloat

        numerator_efficiency_info = self.get_efficiency_info(
            parameter=self._ratio_info_container.numerator.name
        )  # type: UpDownUValue
        denominator_efficiency_info = self.get_efficiency_info(
            parameter=self._ratio_info_container.denominator.name
        )  # type: UpDownUValue

        ratio_ufloat_stat = (
            self._ratio_factor
            * bare_ratio_ufloat
            * (denominator_efficiency_info.nominal_value / numerator_efficiency_info.nominal_value)
        )  # type: UFloat

        mc_stat_sys_cov_mat = np.array(
            [
                [numerator_efficiency_info.std_dev_up ** 2, 0.0],
                [0.0, denominator_efficiency_info.std_dev_up ** 2],
            ]
        )

        numerator_eff, denominator_eff = correlated_values(
            nom_values=[numerator_efficiency_info.nominal_value, denominator_efficiency_info.nominal_value],
            covariance_mat=mc_stat_sys_cov_mat,
        )

        ratio_ufloat_eff_sys = (
            self._ratio_factor * bare_ratio_ufloat.nominal_value * (denominator_eff / numerator_eff)
        )  # type: UFloat
        assert ratio_ufloat_stat.nominal_value == ratio_ufloat_eff_sys.nominal_value, (
            ratio_ufloat_stat,
            ratio_ufloat_eff_sys,
        )

        ratio_ufloat_up = (  # Not used anymore...
            self._ratio_factor
            * bare_ratio_ufloat
            * (denominator_efficiency_info.ufloat_std_dev_up / numerator_efficiency_info.ufloat_std_dev_up)
        )  # type: UFloat
        ratio_ufloat_down = (  # Not used anymore...
            self._ratio_factor
            * bare_ratio_ufloat
            * (denominator_efficiency_info.ufloat_std_dev_down / numerator_efficiency_info.ufloat_std_dev_down)
        )  # type: UFloat

        assert ratio_ufloat_stat.nominal_value == ratio_ufloat_up.nominal_value, (ratio_ufloat_stat, ratio_ufloat_up)
        assert ratio_ufloat_up.nominal_value == ratio_ufloat_down.nominal_value, (ratio_ufloat_up, ratio_ufloat_down)

        eff_with_sys_container = self.get_correlated_efficiency_values_for(
            numerator_parameter=self._ratio_info_container.numerator.name,
            denominator_parameter=self._ratio_info_container.denominator.name,
            additional_sys_covariance_mat_dict=None,
        )  # type: EffSysInfoContainer

        systematic_uncertainties_dict = dict()  # type: Dict[str, float]
        for sys_name in eff_with_sys_container.numerator_eff_sys_dict.keys():
            assert sys_name not in systematic_uncertainties_dict, (sys_name, systematic_uncertainties_dict.keys())
            numerator_efficiency = eff_with_sys_container.numerator_eff_sys_dict[sys_name]  # type: UFloat
            denominator_efficiency = eff_with_sys_container.denominator_eff_sys_dict[sys_name]  # type: UFloat

            ratio_ufloat_sys = (
                self._ratio_factor * bare_ratio_ufloat.nominal_value * (denominator_efficiency / numerator_efficiency)
            )  # type: UFloat
            assert ratio_ufloat_sys.nominal_value == ratio_ufloat_stat.nominal_value, (
                sys_name,
                ratio_ufloat_sys.nominal_value,
                ratio_ufloat_stat.nominal_value,
                systematic_uncertainties_dict.keys(),
            )
            systematic_uncertainties_dict.update({sys_name: ratio_ufloat_sys.std_dev})

        assert "mc_stat_eff_sys" not in systematic_uncertainties_dict, systematic_uncertainties_dict.keys()
        systematic_uncertainties_dict.update({"mc_stat_eff_sys": ratio_ufloat_eff_sys.std_dev})

        all_sys_key = EfficiencySysInfoContainer.all_sys_combined_key  # type: str
        systematic_uncertainties_dict.update({"all_but_mc_stat_sys": systematic_uncertainties_dict[all_sys_key]})
        all_sys_w_mc_stat = np.sqrt(systematic_uncertainties_dict[all_sys_key] ** 2 + ratio_ufloat_eff_sys.std_dev ** 2)
        systematic_uncertainties_dict.update({all_sys_key: all_sys_w_mc_stat})

        return UpDownUValue(
            nominal_value=ratio_ufloat_stat.nominal_value,
            std_dev_up=ratio_ufloat_stat.std_dev,
            std_dev_down=ratio_ufloat_stat.std_dev,
            std_dev_all_sys=systematic_uncertainties_dict[self._efficiencies_sys_container.all_sys_combined_key],
            systematic_uncertainties=systematic_uncertainties_dict,
        )

    @property
    def fit_case(self) -> FitCase:
        return self._fit_case

    @property
    def fit_case_name(self) -> str:
        return self.fit_case.name

    @property
    def fit_result(self) -> MinimizeResult:
        return self._fit_result

    @property
    def efficiency_dict(self) -> EfficiencyNumbersContainer:
        return self._efficiency_dict

    @property
    def yield_ratios(self) -> Tuple[np.ndarray, np.ndarray]:
        if self._yield_ratios is None:
            self._yield_ratios = self._calculate_yield_ratios()
        return self._yield_ratios

    @property
    def efficiency_corrected_ratio(self) -> UpDownUValue:
        if self._efficiency_corrected_ratio is None:
            self._efficiency_corrected_ratio = self.calculate_efficiency_corrected_ratio()
        return self._efficiency_corrected_ratio

    def get_fitted_yield_info(self, parameter_name: str) -> FitYieldInfo:
        _id = self.fit_result.params.param_id_to_index(param_id=parameter_name)  # type: int
        _result = ufloat(
            nominal_value=self.fit_result.params.get_param_value(param_id=parameter_name),
            std_dev=self.fit_result.params.get_param_error(param_id=parameter_name),
        )  # type: UFloat
        return FitYieldInfo(
            name=parameter_name,
            id=_id,
            result=_result,
        )

    def get_efficiency_info(self, parameter: str) -> UpDownUValue:
        if parameter in self._eff_dict.keys():
            return self._eff_dict[parameter]
        else:
            return self._calculate_efficiency_for_parameter(parameter=parameter)

    def _get_ratio_info_container(self) -> RatioInfoContainer:
        numerator_yield_info = self.get_fitted_yield_info(parameter_name=self.numerator_yield_parameter_name)
        denominator_yield_info = self.get_fitted_yield_info(parameter_name=self.denominator_yield_parameter_name)
        return RatioInfoContainer(
            numerator=numerator_yield_info,
            denominator=denominator_yield_info,
        )

    def _calculate_efficiency_for_parameter(
        self,
        parameter: str,
    ) -> UpDownUValue:
        this_component_name = self.get_component_name_from_yield_parameter_name(parameter=parameter)  # type: str
        sig_ids = self.fit_case.get_sig_id_for_component(component_name=this_component_name)
        eff_info = self.efficiency_dict.get_efficiency(
            sig_ids=sig_ids,
            decay_mode_ids=self.all_mode_ids,
        )  # type: EfficiencyInfo

        if parameter in self._eff_dict.keys():
            raise KeyError(f"You are trying to overwrite an efficiency entry for the parameter key {parameter}!")

        systematics_dict = {
            sys_name: eff_sys_info.systematic_uncert_value
            for sys_name, eff_sys_info in self._efficiencies_sys_container.get_all_systematics_for(
                sig_ids=sig_ids
            ).items()
        }  # type: Dict[str, float]

        new_eff_dict_entry = UpDownUValue(
            nominal_value=eff_info.nominal_value,
            std_dev_up=eff_info.std_dev_up,
            std_dev_down=eff_info.std_dev_down,
            std_dev_all_sys=0.0,  # This is wrong, obviously, and the info in systematic_uncertainties below, too!
            systematic_uncertainties=systematics_dict,  # These uncertainties are handled correctly now with the
            # get_correlated_efficiency_values_for method!
        )

        self._eff_dict[parameter] = new_eff_dict_entry
        return new_eff_dict_entry

    def get_correlated_efficiency_values_for(
        self,
        numerator_parameter: str,
        denominator_parameter: str,
        additional_sys_covariance_mat_dict: Optional[Dict[str, np.ndarray]] = None,
    ) -> EffSysInfoContainer:
        numerator_component_name = self.get_component_name_from_yield_parameter_name(
            parameter=numerator_parameter
        )  # type: str
        numerator_sig_ids = self.fit_case.get_sig_id_for_component(
            component_name=numerator_component_name
        )  # type: Tuple[int, ...]
        denominator_component_name = self.get_component_name_from_yield_parameter_name(
            parameter=denominator_parameter
        )  # type: str
        denominator_sig_ids = self.fit_case.get_sig_id_for_component(
            component_name=denominator_component_name
        )  # type: Tuple[int, ...]

        sys_cov_matrix_dict = self._efficiencies_sys_container.get_sys_covariance_matrix_for(
            sig_ids_a=numerator_sig_ids,
            sig_ids_b=denominator_sig_ids,
            additional_sys_covariance_mat_dict=additional_sys_covariance_mat_dict,
        )  # type: Dict[str, np.ndarray]
        assert self._efficiencies_sys_container.all_sys_combined_key in sys_cov_matrix_dict, (
            self._efficiencies_sys_container.all_sys_combined_key,
            sys_cov_matrix_dict.keys(),
        )

        numerator_eff_sys_dict = dict()  # type: Dict[str, UFloat]
        denominator_eff_sys_dict = dict()  # type: Dict[str, UFloat]
        for sys_k, sys_cov_matrix in sys_cov_matrix_dict.items():
            numerator_eff_with_sys, denominator_eff_with_sys = self.efficiency_dict.get_efficiencies_with_systematics(
                sig_ids_a=numerator_sig_ids,
                sig_ids_b=denominator_sig_ids,
                decay_mode_ids=self.all_mode_ids,
                systematics_cov_matrix=sys_cov_matrix,
            )  # type: UFloat, UFloat
            numerator_eff_sys_dict.update({sys_k: numerator_eff_with_sys})
            denominator_eff_sys_dict.update({sys_k: denominator_eff_with_sys})

        return EffSysInfoContainer(
            numerator_eff_with_all_sys=numerator_eff_sys_dict[self._efficiencies_sys_container.all_sys_combined_key],
            denominator_eff_with_all_sys=denominator_eff_sys_dict[self._efficiencies_sys_container.all_sys_combined_key],
            numerator_eff_sys_dict=numerator_eff_sys_dict,
            denominator_eff_sys_dict=denominator_eff_sys_dict,
            sys_cov_matrix_dict=sys_cov_matrix_dict,
        )

    def _generate_jacobi_matrix(self) -> np.ndarray:
        yields = self._all_yields.nominal_values  # type: np.ndarray
        dim = len(yields)  # type: int

        denominator_param_id = self._ratio_info_container.denominator.id  # type: int

        jacobi = np.zeros((dim, dim))
        for i in range(0, dim):
            for j in range(0, dim):
                if i == j and i != denominator_param_id:
                    jacobi[i][j] = 1.0 / yields[denominator_param_id]
                elif i != j and j == denominator_param_id:
                    jacobi[i][j] = -yields[i] / (yields[denominator_param_id] ** 2)
                else:
                    jacobi[i][j] = 0
        return jacobi

    def _calculate_ratio_errors(self) -> np.ndarray:
        yields = self._all_yields.nominal_values  # type: np.ndarray
        cov_matrix = self._all_yields.covariance_matrix  # type: np.ndarray
        dim = len(yields)  # type: int

        assert cov_matrix.shape[0] == cov_matrix.shape[1] == dim, (dim, cov_matrix.shape)

        jacobi = self._generate_jacobi_matrix()  # type: np.ndarray
        dov = np.matmul(np.matmul(jacobi, cov_matrix), np.transpose(jacobi))  # type: np.ndarray
        assert dov.shape[0] == dov.shape[1] == dim, (dim, dov.shape)

        ratio_errors = np.sqrt(np.diagonal(dov))  # type: np.ndarray
        assert len(ratio_errors.shape) == 1, ratio_errors.shape
        assert ratio_errors.shape[0] == dim, (dim, ratio_errors.shape)

        return ratio_errors

    def _calculate_yield_ratios(self) -> Tuple[np.ndarray, np.ndarray]:
        yields = self._all_yields.nominal_values  # type: np.ndarray
        dim = len(yields)  # type: int

        denominator_yield = self._ratio_info_container.denominator.result.nominal_value  # type: float

        ratios = yields / denominator_yield  # type: np.ndarray
        assert len(ratios.shape) == 1, ratios.shape
        assert ratios.shape[0] == dim, (ratios.shape, dim)

        ratio_errors = self._calculate_ratio_errors()

        return ratios, ratio_errors

    @staticmethod
    def get_component_name_from_yield_parameter_name(parameter: str) -> str:
        yield_param_prefix = f"{ParameterHandler.yield_parameter_type}_"  # type: str
        assert parameter.startswith(yield_param_prefix), (parameter, yield_param_prefix)
        return parameter.split(sep=yield_param_prefix, maxsplit=1)[1]
