"""
Class handling the Event and Best Candidate Selection
optimizing the yield of signal events.

Felix Metzner 2020
"""

import os
import copy
import warnings
import numpy as np
import pandas as pd

from pandas.core.common import SettingWithCopyWarning
from typing import Union, Optional, List, Tuple, Dict, Any, Generator

from rdstar.utilities import PathType
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.selection.selection_evaluator import SelectionEvaluator
from rdstar.offline_analysis.selection.selection_evaluation_info_container import SelectionEvaluationContainer
from rdstar.offline_analysis.selection.selection_info_containers import (
    BCSVariableInfo,
    SelectionSpecifics,
    MVAInfo,
    SelectionInfo,
)

from rdstar.mva_selection.mva_selector import MVASelector


__all__ = [
    "Selection",
    "SelectionSetupInterpreter",
]


class SelectionSetupInterpreter:
    def __init__(
        self,
        selection_name: str,
        cs_tag: str,
        output_directory_path: PathType,
        output_file_name: str,
        evaluation_output_prefix: str,
    ) -> None:
        assert selection_name in SelectionSetups.available_selections, selection_name
        self.selection_name: str = selection_name
        self.cs_tag: str = cs_tag
        self.output_directory_path: PathType = output_directory_path
        self.output_file_name = output_file_name
        self.evaluation_output_prefix: str = evaluation_output_prefix

    def get_target_paths(self, selection_tag: str) -> PathType:
        target_path = os.path.join(self.output_directory_path, f"{self.output_file_name}_{selection_tag}.h5")
        return target_path

    def get_evaluation_paths(self, selection_tag: str) -> Tuple[str, PathType]:
        target_path = self.get_target_paths(selection_tag=selection_tag)
        dir_name, base_name = os.path.split(os.path.splitext(target_path)[0])
        path = os.path.join(dir_name, f"{self.evaluation_output_prefix}_{base_name}.json")
        key = f"{self.output_file_name}_{self.evaluation_output_prefix}_{selection_tag}"
        return key, path

    def get_selection_infos(self) -> Generator[SelectionInfo, None, None]:
        for selection_tag, selection in SelectionSetups.get_selection_infos(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        ).items():
            info_output_key, info_target_path = self.get_evaluation_paths(selection_tag=selection_tag)
            yield SelectionInfo(
                selection_tag=selection_tag,
                selection=selection,
                target_path=self.get_target_paths(selection_tag=selection_tag),
                output_key=f"{self.output_file_name}_{selection_tag}",
                info_target_path=info_target_path,
                info_output_key=info_output_key,
            )

    @staticmethod
    def reset_selection(
        selection_specifics: SelectionSpecifics,
        new_cuts: Optional[List[CutInfo]] = None,
        new_post_bcs_cuts: Optional[List[CutInfo]] = None,
        reset_bcs: Union[Tuple, bool] = False,
    ) -> SelectionSpecifics:
        if reset_bcs is True:
            prim_bcs_var = BCSVariableInfo(None)
            sec_bcs_var = BCSVariableInfo(None)
        elif reset_bcs is False:
            prim_bcs_var = selection_specifics.primary_bcs_var
            sec_bcs_var = selection_specifics.secondary_bcs_var
        elif isinstance(reset_bcs, tuple) and all(isinstance(bcs, BCSVariableInfo) for bcs in reset_bcs):
            prim_bcs_var = reset_bcs[0]
            sec_bcs_var = reset_bcs[1]
        else:
            raise RuntimeError(
                f"reset_bcs should either be bool or a tuple of two BCSVariableInfo objects.\n"
                f"Provided was an object of type {type(reset_bcs)}"
            )

        n_p_bcs_cuts = tuple(new_post_bcs_cuts) if new_post_bcs_cuts is not None else selection_specifics.post_bcs_cuts

        return SelectionSpecifics(
            name=selection_specifics.name,
            cuts=tuple(new_cuts) if new_cuts is not None else selection_specifics.cuts,
            post_bcs_cuts=n_p_bcs_cuts,
            primary_bcs_var=prim_bcs_var,
            secondary_bcs_var=sec_bcs_var,
            check_cuts=tuple([]),
            check_bcs=False,
            apply_post_mva_selection_cut=selection_specifics.apply_post_mva_selection_cut,
            run_evaluation=False,
            mva_setup_name=selection_specifics.mva_setup_name,
            post_mva_cut_info=selection_specifics.post_mva_cut_info,
        )

    @staticmethod
    def get_bcs_check_extension(random: bool = False) -> str:
        return "_random_bcs" if random else "_no_bcs"

    @staticmethod
    def get_cut_check_extension(
        cut_number: int,
        no_bcs: bool = False,
        random_bcs: bool = False,
    ) -> str:
        assert isinstance(cut_number, int) and cut_number >= 0, cut_number
        # Check that not both no_bcs and random_bcs are True at the same time:
        if (no_bcs is random_bcs) and no_bcs:
            raise ValueError("The arguments no_bcs and random_bcs must not be both True at the same time!")
        if no_bcs:
            return f"_without_cut_{cut_number}{SelectionSetupInterpreter.get_bcs_check_extension(False)}"
        elif random_bcs:
            return f"_without_cut_{cut_number}{SelectionSetupInterpreter.get_bcs_check_extension(True)}"
        else:
            return f"_without_cut_{cut_number}"

    @staticmethod
    def get_possible_subsample_extensions(input_dict: Dict[str, Any]) -> List[str]:
        extensions = []
        checking = True
        index = 0
        while checking:
            checking = False
            if index == 0:
                ext = SelectionSetupInterpreter.get_bcs_check_extension(random=False)
                if any(ext in key for key in input_dict):
                    extensions.append(ext)
                ext = SelectionSetupInterpreter.get_bcs_check_extension(random=True)
                if any(ext in key for key in input_dict):
                    extensions.append(ext)
            ext = SelectionSetupInterpreter.get_cut_check_extension(cut_number=index)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            ext = SelectionSetupInterpreter.get_cut_check_extension(cut_number=index, no_bcs=True)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            ext = SelectionSetupInterpreter.get_cut_check_extension(cut_number=index, random_bcs=True)
            if any(ext in key for key in input_dict):
                extensions.append(ext)
                checking = True
            index += 1
        return extensions

    def get_subselection_info(self, selection_info: SelectionInfo) -> Generator[SelectionInfo, None, None]:
        if not selection_info.selection.check_cuts and not selection_info.selection.check_bcs:
            return

        if selection_info.selection.check_bcs:
            subselection_tag = f"{selection_info.selection_tag}{self.get_bcs_check_extension(random=False)}"
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, reset_bcs=True),
                target_path=self.get_target_paths(selection_tag=subselection_tag),
                output_key=f"{self.output_file_name}_{subselection_tag}",
            )
            subselection_tag = f"{selection_info.selection_tag}{self.get_bcs_check_extension(random=True)}"
            bcs_tuple_rand = (BCSVariableInfo("random"), BCSVariableInfo(None))
            yield SelectionInfo(
                selection_tag=subselection_tag,
                selection=self.reset_selection(selection_info.selection, reset_bcs=bcs_tuple_rand),
                target_path=self.get_target_paths(selection_tag=subselection_tag),
                output_key=f"{self.output_file_name}_{subselection_tag}",
            )

        if not selection_info.selection.check_cuts:
            return

        for cut_num, cut in enumerate(selection_info.selection.check_cuts):
            other_cuts = [c for c in selection_info.selection.cuts if c is not cut]
            other_post_bcs_cuts = [c for c in selection_info.selection.post_bcs_cuts if c is not cut]

            bcs_options: List[Tuple[Union[bool, Tuple[BCSVariableInfo, BCSVariableInfo]], Dict[str, bool]]] = [
                (False, {}),
                (True, {"no_bcs": True}),
                ((BCSVariableInfo("random"), BCSVariableInfo(None)), {"random_bcs": True}),
            ]

            for bcs_option, bcs_kwarg in bcs_options:
                subselection_tag = f"{selection_info.selection_tag}{self.get_cut_check_extension(cut_num, **bcs_kwarg)}"

                new_selection_for_check = self.reset_selection(
                    selection_info.selection,
                    new_cuts=other_cuts,
                    new_post_bcs_cuts=other_post_bcs_cuts,
                    reset_bcs=bcs_option,
                )

                yield SelectionInfo(
                    selection_tag=subselection_tag,
                    selection=new_selection_for_check,
                    target_path=self.get_target_paths(selection_tag=subselection_tag),
                    output_key=f"{self.output_file_name}_{subselection_tag}",
                )

    def get_selection_output_dict(self) -> Dict[str, PathType]:
        selection_infos = list(self.get_selection_infos())
        number_of_selections: int = len(selection_infos)
        output_dict = {info.output_key: info.target_path for info in selection_infos}
        assert len(output_dict) == number_of_selections, (len(output_dict), number_of_selections)

        # Adding selection evaluation output
        for info in selection_infos:
            if info.is_selection_with_evaluation:
                assert info.info_output_key is not None
                assert info.info_target_path is not None
                assert info.info_output_key not in output_dict.keys(), info.info_output_key
                assert info.info_target_path not in output_dict.values(), info.info_target_path
                output_dict.update({info.info_output_key: info.info_target_path})

        num_selection_w_infos = sum([int(info.is_selection_with_evaluation) for info in selection_infos])

        assert len(output_dict) == number_of_selections + num_selection_w_infos, (
            len(output_dict),
            number_of_selections + num_selection_w_infos,
            number_of_selections,
            num_selection_w_infos,
        )

        for selection_info in selection_infos:
            for sub_selection in self.get_subselection_info(selection_info=selection_info):
                assert sub_selection.output_key not in output_dict, sub_selection.output_key
                assert sub_selection.target_path not in output_dict.values(), sub_selection.target_path
                output_dict.update({sub_selection.output_key: sub_selection.target_path})

        return output_dict


class Selection:
    def __init__(
        self,
        selection_name: str,
        unique_event_id_column_name: str,
        bcs_random_seed: int = 1337,
    ) -> None:
        assert selection_name in SelectionSetups.available_selections, selection_name
        self._selection_name: str = selection_name
        self._u_evt_id_col_name: str = unique_event_id_column_name
        self._bcs_random_seed: int = bcs_random_seed

    @staticmethod
    def apply_cuts(df: pd.DataFrame, cut_list: Tuple[CutInfo, ...]) -> pd.DataFrame:
        cut_string = " and ".join([c.cut_str for c in cut_list])
        return df.query(cut_string)

    @staticmethod
    def apply_bcs(
        df: pd.DataFrame,
        primary_var: BCSVariableInfo,
        secondary_var: BCSVariableInfo,
        bcs_random_seed: int,
        unique_event_id_column_name: str,
    ) -> pd.DataFrame:
        assert isinstance(primary_var, BCSVariableInfo), (primary_var, type(primary_var))
        assert isinstance(secondary_var, BCSVariableInfo), (secondary_var, type(secondary_var))
        assert len(df.index) == len(np.unique(df.index)), (len(df.index), len(np.unique(df.index)))

        if len(df.index) == 0:
            return df

        if primary_var.bcs_variable is None:
            return df

        if primary_var.bcs_variable == "random":
            return (
                df.sample(
                    frac=1.0,
                    random_state=bcs_random_seed,
                )
                .groupby(unique_event_id_column_name)
                .head(1)
                .sort_index()
            )

        prim_var_name = primary_var.bcs_variable
        if primary_var.absolute_value:
            prim_var_name = Selection.add_absolute_value_of_column(df=df, column_name=primary_var.bcs_variable)

        if secondary_var.bcs_variable is not None:
            # Use secondary bcs var if duplicates for best primary bcs variable exist
            sec_var_name = secondary_var.bcs_variable
            if secondary_var.absolute_value:
                sec_var_name = Selection.add_absolute_value_of_column(df=df, column_name=secondary_var.bcs_variable)

            if primary_var.max:
                selected_ids: pd.Index = (
                    df.sample(
                        frac=1.0,
                        random_state=bcs_random_seed,
                    )
                    .sort_values(
                        by=[sec_var_name],
                        ascending=True if not secondary_var.max else False,
                    )
                    .groupby(unique_event_id_column_name)[prim_var_name]
                    .idxmax()
                )

                if secondary_var.absolute_value:
                    del df[sec_var_name]
            else:
                selected_ids = (
                    df.sample(
                        frac=1.0,
                        random_state=bcs_random_seed,
                    )
                    .sort_values(
                        by=[sec_var_name],
                        ascending=True if not secondary_var.max else False,
                    )
                    .groupby(unique_event_id_column_name)[prim_var_name]
                    .idxmin()
                )

                if secondary_var.absolute_value:
                    del df[sec_var_name]
        else:
            # Don't use a secondary bcs variable, instead choose candidate by random
            # if duplicates for best primary bcs variable exist
            if primary_var.max:
                selected_ids = (
                    df.sample(
                        frac=1.0,
                        random_state=bcs_random_seed,
                    )
                    .groupby(unique_event_id_column_name)[prim_var_name]
                    .idxmax()
                )
            else:
                selected_ids = (
                    df.sample(
                        frac=1.0,
                        random_state=bcs_random_seed,
                    )
                    .groupby(unique_event_id_column_name)[prim_var_name]
                    .idxmin()
                )

        if primary_var.absolute_value:
            del df[prim_var_name]

        return df.loc[selected_ids]

    @staticmethod
    def add_absolute_value_of_column(
        df: pd.DataFrame,
        column_name: str,
    ) -> str:
        assert column_name in df.columns, column_name
        new_abs_val_col_name = f"bcs_abs_{column_name}"
        assert new_abs_val_col_name not in df.columns, new_abs_val_col_name

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, new_abs_val_col_name] = np.abs(df[column_name])
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        return new_abs_val_col_name

    def private_apply_bcs(
        self,
        df: pd.DataFrame,
        primary_var: BCSVariableInfo,
        secondary_var: BCSVariableInfo,
    ) -> pd.DataFrame:
        return self.apply_bcs(
            df=df,
            primary_var=primary_var,
            secondary_var=secondary_var,
            bcs_random_seed=self._bcs_random_seed,
            unique_event_id_column_name=self._u_evt_id_col_name,
        )

    def apply_selection(
        self,
        df: pd.DataFrame,
        selection_info: SelectionInfo,
        evaluate_selection: bool = False,
        data_type: Optional[str] = None,
    ) -> Tuple[pd.DataFrame, Optional[SelectionEvaluationContainer]]:

        if evaluate_selection:
            if data_type is None:
                raise ValueError("Parameter 'data_type' is required if 'evaluate_selection' is set to True!")
            selection_evaluator: Optional[SelectionEvaluator] = SelectionEvaluator(df=df, data_type=data_type)
        else:
            selection_evaluator = None

        selected_df = copy.copy(df)

        # Initial Cuts
        if selection_info.selection.apply_post_mva_selection_cut:
            mva_cut_info: Optional[CutInfo] = selection_info.selection.post_mva_cut_info
            assert mva_cut_info is not None
            selection_cuts: Tuple[CutInfo, ...] = selection_info.selection.cuts + (mva_cut_info,)
        else:
            selection_cuts = selection_info.selection.cuts

        if evaluate_selection:
            assert selection_evaluator is not None
            for cut_info in selection_cuts:
                selected_df = self.apply_cuts(df=selected_df, cut_list=(cut_info,))
                selection_evaluator.update_after_cut(df=selected_df, cut=cut_info.cut_str)
        else:
            selected_df = self.apply_cuts(df=selected_df, cut_list=selection_cuts)

        # Best Candidate Selection (BCS)
        evts_before_bcs: int = len(selected_df[self._u_evt_id_col_name].unique())
        selected_df = self.private_apply_bcs(
            df=selected_df,
            primary_var=selection_info.selection.primary_bcs_var,
            secondary_var=selection_info.selection.secondary_bcs_var,
        )

        # Checking that no whole events have been removed
        assert len(selected_df[self._u_evt_id_col_name].unique()) == evts_before_bcs, (
            len(selected_df[self._u_evt_id_col_name].unique()),
            evts_before_bcs,
        )

        # Checking if BCS selected one unique candidates per event
        if selection_info.selection.primary_bcs_var.bcs_variable is not None:
            # If primary bcs var is None, no BCS was performed, so this check would fail!
            assert len(selected_df[self._u_evt_id_col_name].unique()) == len(selected_df.index), (
                len(selected_df[self._u_evt_id_col_name].unique()),
                len(selected_df.index),
            )

        if evaluate_selection:
            assert selection_evaluator is not None
            bcs = (selection_info.selection.primary_bcs_var, selection_info.selection.secondary_bcs_var)
            selection_evaluator.update_after_best_candidate_selection(df=selected_df, bcs=bcs)

        # Post BCS Cuts
        if selection_info.selection.post_bcs_cuts:
            selected_df = self.apply_post_bcs_selections(
                df=selected_df,
                post_bcs_cuts=selection_info.selection.post_bcs_cuts,
                selection_evaluator=selection_evaluator,
            )

        if evaluate_selection:
            assert selection_evaluator is not None
            selection_evaluator.finalize_selection_evaluation(df=selected_df)
            final_selection_eval_info: Optional[SelectionEvaluationContainer] = selection_evaluator.evaluation_info
        else:
            assert selection_evaluator is None
            final_selection_eval_info = None

        return selected_df, final_selection_eval_info

    def apply_post_bcs_selections(
        self,
        df: pd.DataFrame,
        post_bcs_cuts: Tuple[CutInfo, ...],
        selection_evaluator: Optional[SelectionEvaluator],
    ) -> pd.DataFrame:
        if selection_evaluator is None:
            return self.apply_cuts(df=df, cut_list=post_bcs_cuts)
        else:
            selected_df = copy.copy(df)
            for cut_info in post_bcs_cuts:
                selected_df = self.apply_cuts(df=selected_df, cut_list=(cut_info,))
                selection_evaluator.update_after_cut(df=selected_df, cut=cut_info.cut_str)
            return selected_df

    def get_mva_setup_info(self) -> Optional[MVAInfo]:
        return SelectionSetups.get_selection_mva_info(selection_name=self._selection_name)

    def apply_selection_mva(
        self,
        df: pd.DataFrame,
        mva_selection_weight_file_path: PathType,
        mva_selection_var_list_file_path: PathType,
    ) -> None:
        mva_info = self.get_mva_setup_info()
        if mva_info is None:
            return
        mva_setup = mva_info.mva_setup
        mva_setup.data_preparation(df=df, for_application_only=True)

        sel_classifier = MVASelector.load_classifier(
            weight_file_path=mva_selection_weight_file_path,
            var_file_path=mva_selection_var_list_file_path,
        )
        sel_classifier.apply(data=df, output_column=mva_setup.get_selection_mva_output_column_name())
