"""
Provides tasks to create various analysis plots.

Felix Metzner 2020
"""

import os
import itertools
import b2luigi as luigi

from luigi.util import inherits
from collections.abc import Iterable
from typing import Optional, Dict, Generator, Type, Any

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.analysis_plots.analysis_plot_base import AnalysisPlotBaseClass
from rdstar.offline_analysis.analysis_plots.analysis_plots_manager import AnalysisPlotManager

from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionValidationManagerTask


__all__ = [
    "AnalysisPlotManagerTask",
]


@inherits(
    # Here all tasks required to match the requires-function of any possible AnalysisPlotterTask must be listed!
    ResolutionCorrectionValidationManagerTask,
)
class AnalysisPlotManagerTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        a_p_base_dir_name = AnalysisPlotManager.analysis_plots_base_dir_name
        self.task_working_dir = os.path.join(self.working_dir, a_p_base_dir_name, self.selection_name)

    def get_log_file_dir(self) -> PathType:
        return self.task_working_dir

    def get_task_file_dir(self) -> PathType:
        return self.task_working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for plotter_id_key, plot_class in AnalysisPlotManager.analysis_plot_types.items():
            plot_options = plot_class.get_plotter_options(selection_name=self.selection_name, cs_tag=self.cs_tag)
            value_combinations = list(itertools.product(*plot_options.values()))
            plot_class_arguments_dicts = [dict(zip(plot_options.keys(), vc)) for vc in value_combinations]
            for plot_class_arguments_dict in plot_class_arguments_dicts:
                yield self.clone(
                    AnalysisPlotterTask,
                    analysis_plot_type=plotter_id_key,
                    analysis_plot_class_arguments=plot_class_arguments_dict,
                )


@inherits(AnalysisPlotManagerTask)
class AnalysisPlotterTask(utilities.BaseOfflineTask):
    analysis_plot_type = luigi.Parameter(positional=False)  # type: str
    analysis_plot_class_arguments = luigi.DictParameter(positional=False, hashed=True)  # type: Dict[str, Any]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        a_p_base_dir_name = AnalysisPlotManager.analysis_plots_base_dir_name
        self.base_path = os.path.join(self.working_dir, a_p_base_dir_name, self.selection_name)

        self._plotter = None  # type: Optional[AnalysisPlotBaseClass]
        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def __plotter_class(self) -> Type[AnalysisPlotBaseClass]:
        return AnalysisPlotManager.get_plotter_class(plot_type_key=self.analysis_plot_type)

    @property
    def plotter(self) -> AnalysisPlotBaseClass:
        if self._plotter is None:
            self._plotter = self.__plotter_class(
                input_dictionary=self.input_dict,
                output_dir_path=self.base_path,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                **self.analysis_plot_class_arguments,
            )
        return self._plotter

    def create_analysis_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        if not dry_run:
            os.makedirs(self.plotter.output_dir_path, exist_ok=True)
        return self.plotter.create_plots(dry_run=dry_run)

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.create_analysis_plots()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.create_analysis_plots(dry_run=True)
        return {k: luigi.LocalTarget(p) for k, p in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, Iterable)
        for combo_name, _, relevant_configs in self.config_combinations:
            yield self.clone(
                self.__plotter_class.get_required_task_type(),
                combo_name=combo_name,
                config_params=relevant_configs,
            )
