"""
Analysis Flow Validation
    - Cut Flow Validation
    - BCS Validation

Felix Metzner 2019
"""

import os
import numpy as np
import pandas as pd

from collections import Counter
from typing import Optional, Tuple, List, Sequence, Dict, NamedTuple, Callable

from analysistools.plotting.latex_tables import df_to_split_latex_table

from rdstar.utilities import PathType

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDToLatexLabelMappingType

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.selection.selection_evaluation_information_base import EvalInfoContainerBase
from rdstar.offline_analysis.selection.selection_evaluation_info_container import SelectionEvaluationContainer


__all__ = [
    "create_simple_bcs_table",
    "bcs_info_items",
    "create_bcs_table",
    "create_bcs_table_for_data",
    "create_simple_cut_flow_table",
    "cut_flow_info_items",
    "get_cut_flow_table_col_name",
    "create_cut_flow_table",
    "create_cut_flow_table_for_data",
    "update_table_fraction_output_dict",
]


sig_id_label_mapping: MCSigIDToLatexLabelMappingType = OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label_without_pilnu()
old_sig_id_label_mapping: MCSigIDToLatexLabelMappingType = OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label()


def update_table_fraction_output_dict(
    output_dict: Dict[str, PathType],
    name_prefix: str,
    base_path: PathType,
    create_pdfs: bool,
    table_fractions: Optional[List[int]],
) -> None:
    if table_fractions is None:
        _table_fractions: List[int] = [1]
    else:
        _table_fractions = table_fractions

    for frac in _table_fractions:
        file_name = f"{name_prefix}_table_{frac}_part"
        assert file_name not in output_dict.keys(), file_name

        output_dict.update({f"{file_name}_latex": os.path.join(base_path, f"{file_name}.tex")})
        if create_pdfs:
            output_dict.update({f"{file_name}": os.path.join(base_path, f"{file_name}.pdf")})

        for i in range(frac):
            p_file_name = f"{name_prefix}_table_{i + 1}_of_{frac}"
            assert p_file_name not in output_dict.keys(), p_file_name

            output_dict.update({f"{p_file_name}_latex": os.path.join(base_path, f"{p_file_name}.tex")})
            if create_pdfs:
                output_dict.update({f"{p_file_name}": os.path.join(base_path, f"{p_file_name}.pdf")})


# Simple BCS Validation
class SimpleBCSInfoItem(NamedTuple):
    info_name: str

    @property
    def name(self) -> str:
        return self.info_name


# fmt: off
simple_bcs_info_items: Tuple[SimpleBCSInfoItem, ...] = (
    SimpleBCSInfoItem(info_name="events_per_reco_sig_id_probe_and_matching_decay_mode"),
    SimpleBCSInfoItem(info_name="events_per_reco_sig_id_probe_and_matching_decay_mode_with_down_feed"),
)
# fmt: on

assert len(set(v.name for v in simple_bcs_info_items)) == len(simple_bcs_info_items), [
    n for n, c in Counter([v.name for v in simple_bcs_info_items]).items() if c > 1
]


def create_simple_bcs_table(
    info_container: Optional[SelectionEvaluationContainer],
    selection_tag: str,
    base_path: PathType,
    create_pdfs: bool,
    dry_run: bool,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}
    dir_path: PathType = os.path.join(base_path, "SimpleBCSValidation")
    table_fractions: List[int] = [1]

    for simple_bcs_info in simple_bcs_info_items:
        name_prefix: str = f"{selection_tag}_sbcst_{simple_bcs_info.name}_bcs"
        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=dir_path,
            create_pdfs=create_pdfs,
            table_fractions=table_fractions,
        )

    if not dry_run:
        assert info_container is not None
        os.makedirs(dir_path, exist_ok=True)

        bcs_info_key: str = "cut_and_bcs_info"
        # counts after bcs
        _nom_info_dict = info_container.info_entry_for_full_selection.get_sub_info_by_key(key=bcs_info_key)
        assert _nom_info_dict is not None
        nom_info_dict: EvalInfoContainerBase = _nom_info_dict
        # counts before bcs
        scnd_last_level: int = sorted(info_container.levels)[-2]
        _denom_info_dict = info_container.get_entry_by_level(level=scnd_last_level).get_sub_info_by_key(key=bcs_info_key)
        assert _denom_info_dict is not None
        denom_info_dict: EvalInfoContainerBase = _denom_info_dict

        col_formatters: Dict[str, Callable] = {
            "Before": lambda s: f"{s:.0f}",
            "After": lambda s: f"{s:.0f}",
            "Efficiency": lambda s: f"{s * 100:0.2f}" + r"$\,\si{\percent}$",
        }

        for simple_bcs_info in simple_bcs_info_items:
            name_prefix = f"{selection_tag}_sbcst_{simple_bcs_info.name}_bcs"
            nom = nom_info_dict.get_value(key=simple_bcs_info.info_name)
            denom = denom_info_dict.get_value(key=simple_bcs_info.info_name)

            base_dict: Dict[str, Dict[int, int]] = {
                "Before": denom,
                "After": nom,
            }

            df: pd.DataFrame = pd.DataFrame(base_dict)

            df.loc[:, "Efficiency"] = df["After"] / df["Before"]

            df = df.T
            df = df.reindex(columns=sorted([k for k in sig_id_label_mapping.keys() if k in df.columns]))
            df = df.rename(columns={k: v for k, v in sig_id_label_mapping.items()})
            df = df.T
            df.columns.name = r"\multicolumn{1}{l}{Category}"

            df_to_split_latex_table(
                df=df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=False,
                create_pdf=create_pdfs,
                col_formatters=col_formatters,
                allow_smaller_sub_dicts=True,
            )

    return output_dict


# BCS Validation
class BCSInfoItem(NamedTuple):
    sig: str
    bkg: str
    norm: Optional[str] = None

    @property
    def name(self) -> str:
        return self.sig


# fmt: off
bcs_info_items: Tuple[BCSInfoItem, ...] = (
    BCSInfoItem(sig="reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="u4s_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="u4s_match_reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="b_sig_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="b_sig_match_reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="reco_sig_match_events_per_decay_mode_id_probe", bkg="reco_bkg_match_candidates_per_decay_mode_id_probe", norm="reco_norm_match_events_per_decay_mode_id_probe"),
    BCSInfoItem(sig="u4s_match_reco_sig_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_events_per_decay_mode_id_probe"),
    BCSInfoItem(sig="b_sig_match_reco_sig_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_events_per_decay_mode_id_probe"),
    # With allowing for down feed
    BCSInfoItem(sig="reco_sig_or_norm_match_with_down_feed_events_per_decay_mode_id_probe", bkg="reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="u4s_match_reco_sig_or_norm_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="b_sig_match_reco_sig_or_norm_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm=None),
    BCSInfoItem(sig="reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm="reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
    BCSInfoItem(sig="u4s_match_reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
    BCSInfoItem(sig="b_sig_match_reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_with_down_feed_candidates_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
)
# fmt: on

assert len(set(v.name for v in bcs_info_items)) == len(bcs_info_items), [
    n for n, c in Counter([v.name for v in bcs_info_items]).items() if c > 1
]


def create_bcs_table(
    info_container: SelectionEvaluationContainer,
    bcs_infos: Tuple[BCSInfoItem, ...],
) -> Dict[str, pd.DataFrame]:
    bcs_info_key: str = "cut_and_bcs_info"
    # counts after bcs
    _nom_info_dict = info_container.info_entry_for_full_selection.get_sub_info_by_key(key=bcs_info_key)
    assert _nom_info_dict is not None
    nom_info_dict: EvalInfoContainerBase = _nom_info_dict
    # counts before bcs
    scnd_last_level: int = sorted(info_container.levels)[-2]
    _denom_info_dict = info_container.get_entry_by_level(level=scnd_last_level).get_sub_info_by_key(key=bcs_info_key)
    assert _denom_info_dict is not None
    denom_info_dict: EvalInfoContainerBase = _denom_info_dict

    output: Dict[str, pd.DataFrame] = {}

    for bcs_info in bcs_infos:
        norm_nom, norm_denom = (None, None)
        has_norm: bool = bcs_info.norm is not None

        sig_nom = nom_info_dict.get_value(key=bcs_info.sig)
        sig_denom = denom_info_dict.get_value(key=bcs_info.sig)
        bkg_nom = nom_info_dict.get_value(key=bcs_info.bkg)
        bkg_denom = denom_info_dict.get_value(key=bcs_info.bkg)
        if has_norm:
            assert bcs_info.norm is not None
            norm_nom = nom_info_dict.get_value(key=bcs_info.norm)
            norm_denom = denom_info_dict.get_value(key=bcs_info.norm)

        base_dict: Dict[str, Dict[int, int]] = {
            "Sig before": sig_denom,
            "Sig after": sig_nom,
        }
        if has_norm:
            assert norm_nom is not None and norm_denom is not None
            base_dict.update(
                {
                    "Norm before": norm_denom,
                    "Norm after": norm_nom,
                }
            )
        base_dict.update(
            {
                "Bkg before": bkg_denom,
                "Bkg after": bkg_nom,
            }
        )

        df: pd.DataFrame = pd.DataFrame(base_dict)

        df.loc[:, "Sig Eff"] = df["Sig after"] / df["Sig before"]
        if has_norm:
            df.loc[:, "Norm Eff"] = df["Norm after"] / df["Norm before"]
            df.loc[~np.isfinite(df["Norm Eff"]), "Norm Eff"] = np.nan
        df.loc[:, "Bkg Eff"] = df["Bkg after"] / df["Bkg before"]

        # For norm: set all zeroes to NaNs to be able to remove irrelevant row:
        if has_norm:
            for row in [c for c in df.columns if c.startswith("Norm ")]:
                df.loc[df[row] == 0.0, row] = np.nan

        df = df.T
        df = df.rename(columns={k: rf"{k}_{v}" for k, v in RecoModeInfos.get_reco_id_to_latex_label().items()})

        df.fillna("---", inplace=True)

        output.update({bcs_info.name: df})
    return output


def create_bcs_table_for_data(
    info_container: SelectionEvaluationContainer,
) -> pd.DataFrame:
    data_bcs_info_key: str = "basic_info"
    # counts after bcs
    _nom_info_dict = info_container.info_entry_for_full_selection.get_sub_info_by_key(key=data_bcs_info_key)
    assert _nom_info_dict is not None
    nom_info_dict: EvalInfoContainerBase = _nom_info_dict
    # counts before bcs
    scnd_last_level: int = sorted(info_container.levels)[-2]
    _denom_info_dict = info_container.get_entry_by_level(level=scnd_last_level).get_sub_info_by_key(key=data_bcs_info_key)
    assert _denom_info_dict is not None
    denom_info_dict: EvalInfoContainerBase = _denom_info_dict

    candidates_info_key: str = "candidates"
    cands_per_decay_mode_id_info_key: str = "candidates_per_decay_mode_id"

    cand_nom = nom_info_dict.get_value(key=cands_per_decay_mode_id_info_key)
    cand_denom = denom_info_dict.get_value(key=cands_per_decay_mode_id_info_key)

    assert all([-1 not in di.keys() for di in [cand_nom, cand_denom]])
    cand_nom.update({-1: nom_info_dict.get_value(key=candidates_info_key)})
    cand_denom.update({-1: denom_info_dict.get_value(key=candidates_info_key)})

    base_dict = {
        "Before": cand_denom,
        "After": cand_nom,
    }

    df = pd.DataFrame(base_dict)

    df.loc[:, "Eff"] = df["After"] / df["Before"]

    df = df.T
    df = df.rename(columns={k: rf"{k}_{v}" for k, v in RecoModeInfos.get_reco_id_to_latex_label().items()})
    df = df.rename(columns={-1: "00_All"})

    df.fillna("---", inplace=True)

    return df


# Cut Flow Tables


def get_cut_flow_table_col_name(
    index: int,
    cut_list: Sequence[str],
    selection_name: str,
) -> str:
    assert index <= len(cut_list), (index, len(cut_list), cut_list)
    if index == 0:
        return "No Selection"
    if cut_list[index - 1] == SelectionEvaluationContainer.bcs_current_cut_str:
        return "After BCS"
    return SelectionSetups.get_selection_cut_label_dict(selection_name=selection_name)[cut_list[index - 1]]


# Simple Cut FLow Validation
class SimpleCutInfoItem(NamedTuple):
    info_name: str

    @property
    def name(self) -> str:
        return self.info_name


# fmt: off
simple_cut_flow_info_items: Tuple[SimpleCutInfoItem, ...] = (
    SimpleCutInfoItem(info_name="candidates_per_reco_sig_id_probe_and_matching_decay_mode"),
    SimpleCutInfoItem(info_name="candidates_per_reco_sig_id_probe_and_matching_decay_mode_with_down_feed"),
    SimpleCutInfoItem(info_name="events_per_reco_sig_id_probe_and_matching_decay_mode"),
    SimpleCutInfoItem(info_name="events_per_reco_sig_id_probe_and_matching_decay_mode_with_down_feed"),
)
# fmt: on

assert len(set(v.name for v in simple_cut_flow_info_items)) == len(simple_cut_flow_info_items), [
    n for n, c in Counter([v.name for v in simple_cut_flow_info_items]).items() if c > 1
]


def create_simple_cut_flow_table(
    info_container: Optional[SelectionEvaluationContainer],
    selection_tag: str,
    selection_name: str,
    base_path: PathType,
    create_pdfs: bool,
    dry_run: bool,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}
    dir_path: PathType = os.path.join(base_path, "SimpleCutFlowValidation")
    table_fractions: List[int] = [1, 2]
    # table_fractions = table_fractions[:1]  # Ignoring traction 2, because we do have an uneven number to evaluate atm.

    for simple_cut_flow_info in simple_cut_flow_info_items:
        name_prefix: str = f"{selection_tag}_scft_{simple_cut_flow_info.name}_simple_cut_flow"
        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=dir_path,
            create_pdfs=create_pdfs,
            table_fractions=table_fractions,
        )

    if not dry_run:
        assert info_container is not None
        os.makedirs(dir_path, exist_ok=True)

        n_levels: int = len(info_container)
        cut_list: Tuple[str, ...] = info_container.full_cut_list
        assert len(cut_list) == n_levels - 1, (len(cut_list), n_levels)

        column_formatter: List[Callable] = [
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s :0.2f}",
        ]

        cut_sub_info_key: str = "cut_and_bcs_info"
        info_by_level: Dict[int, EvalInfoContainerBase] = {
            level: info_container.get_entry_by_level(level=level).get_valid_sub_info_by_key(key=cut_sub_info_key)
            for level in info_container.levels
        }

        for simple_cut_flow_info in simple_cut_flow_info_items:
            name_prefix = f"{selection_tag}_scft_{simple_cut_flow_info.name}_simple_cut_flow"

            table_dict: Dict[int, Dict[int, int]] = {
                level: v.get_value(key=simple_cut_flow_info.info_name) for level, v in info_by_level.items()
            }

            df: pd.DataFrame = pd.DataFrame.from_dict(table_dict, orient="index")
            sorted_column_ids: List[int] = sorted(list(df.columns))
            for c in sorted(df.columns):
                df[c, "Counts"] = df[c]
                df[c, "Eff"] = df[c, "Counts"] * 100.0 / df[c, "Counts"].shift(periods=1, fill_value=np.NaN)
                del df[c]

            # Calculate Total Efficiencies
            total_row_name: str = "Total Eff"
            fc = list(df.columns)[1]
            df = df.T
            df.loc[fc::2, total_row_name] = (df.loc[::2, n_levels - 1] * 100.0 / df.loc[::2, 0]).values

            df = df.T
            df.fillna("---", inplace=True)

            df.columns = pd.MultiIndex.from_tuples(df.columns, names=[None, r"\multicolumn{1}{l}{Criteria}"])

            # Column Renaming:
            df = df.rename(columns={k: rf"{k}_{v}" for k, v in sig_id_label_mapping.items()})
            df.columns.set_levels(
                [r"\multicolumn{1}{c}{Counts}", r"\multicolumn{1}{c}{Eff in \si{\percent}}"],
                level=1,
                inplace=True,
            )

            df = df.T
            df.rename(
                columns={
                    cn: get_cut_flow_table_col_name(index=cn, cut_list=cut_list, selection_name=selection_name)
                    if cn != total_row_name
                    else cn
                    for cn in df.columns
                },
                inplace=True,
            )
            df = df.T

            new_sorted_column_names: List[str] = [rf"{cn}_{old_sig_id_label_mapping[cn]}" for cn in sorted_column_ids]
            df = df.reindex(columns=new_sorted_column_names, level=0)

            assert len(df.columns) == 2 * len(df.columns.levels[0]), (
                len(df.columns),
                len(df.columns.levels[0]),
                df.columns,
                simple_cut_flow_info.info_name,
                table_dict,
            )

            df_to_split_latex_table(
                df=df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=True,
                split_col_formatters=column_formatter,
                create_pdf=create_pdfs,
                multicolumn_format="c",
                allow_smaller_sub_dicts=True,
            )

    return output_dict


class CutInfoItem(NamedTuple):
    sig: str
    bkg: str
    norm: Optional[str] = None

    @property
    def name(self) -> str:
        return self.sig


# fmt: off
# Cut Flow Validation
cut_flow_info_items: Tuple[CutInfoItem, ...] = (
    CutInfoItem(sig="reco_sig_or_norm_match_candidates_per_decay_mode_id_probe", bkg="reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="u4s_match_candidates_per_decay_mode_id_probe", bkg="no_u4s_match_candidates_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="b_sig_match_candidates_per_decay_mode_id_probe", bkg="no_b_sig_match_candidates_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="u4s_match_reco_sig_or_norm_match_candidates_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="b_sig_match_reco_sig_or_norm_match_candidates_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="reco_sig_match_candidates_per_decay_mode_id_probe", bkg="reco_bkg_match_candidates_per_decay_mode_id_probe", norm="reco_norm_match_candidates_per_decay_mode_id_probe"),
    CutInfoItem(sig="u4s_match_reco_sig_match_candidates_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_candidates_per_decay_mode_id_probe"),
    CutInfoItem(sig="b_sig_match_reco_sig_match_candidates_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_candidates_per_decay_mode_id_probe"),
    CutInfoItem(sig="reco_sig_match_with_down_feed_candidates_per_decay_mode_id_probe", bkg="reco_bkg_match_candidates_per_decay_mode_id_probe", norm="reco_norm_match_with_down_feed_candidates_per_decay_mode_id_probe"),
    CutInfoItem(sig="u4s_match_reco_sig_match_with_down_feed_candidates_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_with_down_feed_candidates_per_decay_mode_id_probe"),
    CutInfoItem(sig="b_sig_match_reco_sig_match_with_down_feed_candidates_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_candidates_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_with_down_feed_candidates_per_decay_mode_id_probe"),

    CutInfoItem(sig="reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="reco_bkg_match_events_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="u4s_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_events_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="b_sig_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_events_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="u4s_match_reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_events_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="b_sig_match_reco_sig_or_norm_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_events_per_decay_mode_id_probe", norm=None),
    CutInfoItem(sig="reco_sig_match_events_per_decay_mode_id_probe", bkg="reco_bkg_match_events_per_decay_mode_id_probe", norm="reco_norm_match_events_per_decay_mode_id_probe"),
    CutInfoItem(sig="u4s_match_reco_sig_match_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_events_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_events_per_decay_mode_id_probe"),
    CutInfoItem(sig="b_sig_match_reco_sig_match_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_events_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_events_per_decay_mode_id_probe"),
    CutInfoItem(sig="reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="reco_bkg_match_events_per_decay_mode_id_probe", norm="reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
    CutInfoItem(sig="u4s_match_reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_u4s_match_reco_bkg_match_events_per_decay_mode_id_probe", norm="u4s_match_reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
    CutInfoItem(sig="b_sig_match_reco_sig_match_with_down_feed_events_per_decay_mode_id_probe", bkg="no_b_sig_match_reco_bkg_match_events_per_decay_mode_id_probe", norm="b_sig_match_reco_norm_match_with_down_feed_events_per_decay_mode_id_probe"),
)
# fmt: on


assert len(set(v.name for v in cut_flow_info_items)) == len(cut_flow_info_items), [
    n for n, c in Counter([v.name for v in cut_flow_info_items]).items() if c > 1
]


def create_cut_flow_table_df(
    info_container: SelectionEvaluationContainer,
    cut_info_item: CutInfoItem,
    selection_name: str,
    compact_layout: bool,
) -> pd.DataFrame:
    cut_sub_info_key: str = "cut_and_bcs_info"

    total_row_name: str = "Total Eff"
    cut_label_col_name: str = r"\multicolumn{1}{l}{Criteria}"

    counts_sub_col_label: str = r"\multicolumn{1}{c}{Counts}"
    effs_sub_col_label: str = (
        r"\multicolumn{1}{c}{Eff}" if compact_layout else r"\multicolumn{1}{c}{Eff in \si{\percent}}"
    )

    sig_label: str = "Signal"
    norm_label: str = "Norm"
    bkg_label: str = "Bkg"

    has_norm: bool = cut_info_item.norm is not None

    n_levels: int = len(info_container)
    cut_list: Tuple[str, ...] = info_container.full_cut_list
    assert len(cut_list) == n_levels - 1, (len(cut_list), n_levels)

    info_by_level: Dict[int, EvalInfoContainerBase] = {
        level: info_container.get_entry_by_level(level=level).get_valid_sub_info_by_key(key=cut_sub_info_key)
        for level in info_container.levels
    }

    sub_index: int = 2
    # Signal
    table_dicts: List[Dict[Tuple[int, str], Dict[int, int]]] = [
        {(i, "1_sig"): v.get_value(key=cut_info_item.sig) for i, v in info_by_level.items()}
    ]

    # Norm
    if has_norm:
        sub_index = 3
        assert cut_info_item.norm is not None
        table_dicts.append({(i, "2_norm"): v.get_value(key=cut_info_item.norm) for i, v in info_by_level.items()})

    # Background
    table_dicts.append({(i, f"{sub_index}_bkg"): v.get_value(key=cut_info_item.bkg) for i, v in info_by_level.items()})

    # Calculate Efficiencies for each Cut and create Multicolumn Data Frame
    dfs: List[pd.DataFrame] = []
    for table_dict in table_dicts:
        _df = pd.DataFrame.from_dict(table_dict, orient="index")
        for c in _df.columns:
            _df[c, "Counts"] = _df[c]
            _df[c, "Eff"] = _df[c, "Counts"] * 100.0 / _df[c, "Counts"].shift(periods=1, fill_value=np.NaN)
            del _df[c]
        dfs.append(_df)

    df = pd.concat(dfs, axis=0, join="outer", ignore_index=False)

    # Calculate Total Efficiencies
    fc = list(df.columns)[1]
    df = df.T
    df.loc[fc::2, pd.IndexSlice[total_row_name, "1_sig"]] = (
        df.loc[::2, pd.IndexSlice[n_levels - 1, "1_sig"]] * 100.0 / df.loc[::2, pd.IndexSlice[0, "1_sig"]]
    ).values
    if has_norm:
        df.loc[fc::2, pd.IndexSlice[total_row_name, "2_norm"]] = (
            df.loc[::2, pd.IndexSlice[n_levels - 1, "2_norm"]] * 100.0 / df.loc[::2, pd.IndexSlice[0, "2_norm"]]
        ).values
        df.loc[~np.isfinite(df[(total_row_name, "2_norm")]), (total_row_name, "2_norm")] = np.nan
    df.loc[fc::2, pd.IndexSlice[total_row_name, f"{sub_index}_bkg"]] = (
        df.loc[::2, pd.IndexSlice[n_levels - 1, f"{sub_index}_bkg"]]
        * 100.0
        / df.loc[::2, pd.IndexSlice[0, f"{sub_index}_bkg"]]
    ).values

    # For norm: set all zeroes to NaNs to be able to remove irrelevant row:
    if has_norm:
        for row in df.columns.levels[0]:
            df.loc[df[(row, "2_norm")] == 0.0, (row, "2_norm")] = np.nan
    df = df.T

    df.fillna("---", inplace=True)
    df.columns = pd.MultiIndex.from_tuples(df.columns)

    # Renaming:
    df = df.rename(columns={k: rf"{k}_{v}" for k, v in RecoModeInfos.get_reco_id_to_latex_label().items()})
    df.columns.set_levels([counts_sub_col_label, effs_sub_col_label], level=1, inplace=True)

    df.sort_index(axis=0, inplace=True)
    df = df.T
    df.columns.set_levels(
        [
            get_cut_flow_table_col_name(index=i, cut_list=cut_list, selection_name=selection_name)
            if i != total_row_name
            else i
            for i in df.columns.levels[0]
        ],
        level=0,
        inplace=True,
    )
    index_l2_names = [sig_label, bkg_label]
    if has_norm:
        index_l2_names.insert(1, norm_label)
    df.columns.set_levels(index_l2_names, level=1, inplace=True)
    df = df.T

    # df.columns.names = [None, cut_label_col_name]
    df.index.names = [cut_label_col_name, None]

    return df


def create_cut_flow_table(
    info_container: Optional[SelectionEvaluationContainer],
    selection_tag: str,
    selection_name: str,
    base_path: PathType,
    compact_layout: bool,
    create_pdfs: bool,
    dry_run: bool,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}

    c_infos: Tuple[CutInfoItem, ...] = cut_flow_info_items
    table_fractions: List[int] = [2, 3, 4, 6]

    dir_path: PathType = os.path.join(base_path, "CutFlowEvaluation")
    compact_tag: str = "_compact_layout" if compact_layout else ""

    for cut_eval_info in c_infos:
        name_prefix: str = f"{selection_tag}_cft_{cut_eval_info.name}_cut_flow{compact_tag}"

        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=dir_path,
            create_pdfs=create_pdfs,
            table_fractions=table_fractions,
        )

        if not dry_run:
            assert info_container is not None

            column_formatter: List[Callable] = [
                lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s :0.2f}",
            ]

            os.makedirs(dir_path, exist_ok=True)

            df = create_cut_flow_table_df(
                info_container=info_container,
                cut_info_item=cut_eval_info,
                selection_name=selection_name,
                compact_layout=compact_layout,
            )

            df_to_split_latex_table(
                df=df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=True,
                split_col_formatters=column_formatter,
                create_pdf=create_pdfs,
                multicolumn_format="c",
                allow_smaller_sub_dicts=True,
            )

    return output_dict


def create_cut_flow_table_df_for_data(
    info_container: SelectionEvaluationContainer,
    selection_name: str,
    compact_layout: bool,
) -> pd.DataFrame:
    sub_info_key: str = "basic_info"

    events_info_key: str = "events"
    candidates_info_key: str = "candidates"
    cands_per_decay_mode_id_info_key: str = "candidates_per_decay_mode_id"

    total_row_name: str = "Total Eff"
    cut_label_col_name: str = r"\multicolumn{1}{l}{Criteria}"

    counts_sub_col_label: str = r"\multicolumn{1}{c}{Counts}"
    effs_sub_col_label: str = (
        r"\multicolumn{1}{c}{Eff}" if compact_layout else r"\multicolumn{1}{c}{Eff in \si{\percent}}"
    )

    info_by_level: Dict[int, EvalInfoContainerBase] = {
        level: info_container.get_entry_by_level(level=level).get_valid_sub_info_by_key(key=sub_info_key)
        for level in info_container.levels
    }

    n_levels: int = len(info_container)
    cut_list: Tuple[str, ...] = info_container.full_cut_list
    assert len(cut_list) == n_levels - 1, (len(cut_list), n_levels)

    table_dict: Dict[int, Dict[int, int]] = {}
    table_dict.update({i: v.get_value(key=cands_per_decay_mode_id_info_key) for i, v in info_by_level.items()})

    all_infos = [events_info_key, candidates_info_key]
    for level, v in info_by_level.items():
        table_dict[level].update({-(i + 1): v.get_value(key=info_key) for i, info_key in enumerate(all_infos)})

    df = pd.DataFrame.from_dict(table_dict, orient="index")

    # Calculate Efficiencies for each Cut and create Multicolumn Data Frame
    for c in df.columns:
        df[c, "Counts"] = df[c]
        df[c, "Eff"] = df[c, "Counts"] * 100.0 / df[c, "Counts"].shift(1)
        del df[c]

    # Calculate Total Efficiencies
    fc = list(df.columns)[1]
    df = df.T

    df.loc[fc::2, pd.IndexSlice[total_row_name]] = (
        df.loc[::2, pd.IndexSlice[n_levels - 1]] * 100.0 / df.loc[::2, pd.IndexSlice[0]]
    ).values

    df = df.T
    df.fillna("---", inplace=True)
    df.columns = pd.MultiIndex.from_tuples(df.columns, names=[None, cut_label_col_name])

    # Renaming:
    df = df.rename(columns={k: rf"{k}_{v}" for k, v in RecoModeInfos.get_reco_id_to_latex_label().items()})
    df = df.rename(columns={-1: "-2_All Events", -2: "-1_All Candidates"})

    df.columns.set_levels([counts_sub_col_label, effs_sub_col_label], level=1, inplace=True)

    df = df.T
    df.columns = [
        get_cut_flow_table_col_name(index=i, cut_list=cut_list, selection_name=selection_name)
        if i != total_row_name
        else i
        for i in df.columns
    ]

    df = df.T

    return df


def create_cut_flow_table_for_data(
    info_container: Optional[SelectionEvaluationContainer],
    selection_tag: str,
    selection_name: str,
    base_path: PathType,
    compact_layout: bool,
    create_pdfs: bool,
    dry_run: bool,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}

    table_fractions: List[int] = [2, 3, 4, 6]

    compact_tag: str = "_compact_layout" if compact_layout else ""
    name_prefix: str = f"{selection_tag}_cft_data_cut_flow{compact_tag}"
    dir_path: PathType = os.path.join(base_path, "CutFlowEvaluationOnData")

    update_table_fraction_output_dict(
        output_dict=output_dict,
        name_prefix=name_prefix,
        base_path=dir_path,
        create_pdfs=create_pdfs,
        table_fractions=table_fractions,
    )

    if not dry_run:
        assert info_container is not None

        column_formatter: List[Callable] = [
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
            lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s :0.2f}",
        ]

        df = create_cut_flow_table_df_for_data(
            info_container=info_container,
            selection_name=selection_name,
            compact_layout=compact_layout,
        )

        df_to_split_latex_table(
            df=df,
            table_fractions=table_fractions,
            name_prefix=name_prefix,
            output_lookup_dict=output_dict,
            placeholder_element="---",
            remove_column_name_index=True,
            split_col_formatters=column_formatter,
            create_pdf=create_pdfs,
            multicolumn_format="c",
            allow_smaller_sub_dicts=True,
        )

    return output_dict
