"""
Provides information about neutral B meson decays as used in Belle.

Felix Metzner 2020
"""

import numpy as np

from typing import Tuple, List, Set

from rdstar.utilities.logging import log_to_default
from rdstar.branching_fractions.bf_utilities import BelleBBFEntry

logger = log_to_default(name=__name__)

__all__ = [
    "sum_belle_decay_b_zero",
    "sum_belle_decay_anti_b_zero",
    "belle_bf_b0_to_d_l_nu",
    "belle_bf_b0_to_d_st_l_nu",
    "belle_bf_b0_to_d_tau_nu",
    "belle_bf_b0_to_d_st_tau_nu",
    "belle_bf_b0_to_dzero_st_l_nu",
    "belle_bf_b0_to_done_l_nu",
    "belle_bf_b0_to_done_prime_l_nu",
    "belle_bf_b0_to_dtwo_st_l_nu",
    "belle_bf_b0_to_dzero_st_tau_nu",
    "belle_bf_b0_to_done_tau_nu",
    "belle_bf_b0_to_done_prime_tau_nu",
    "belle_bf_b0_to_dtwo_st_tau_nu",
    # Break up of Done cases:
    "belle_bf_b0_to_done_to_dst_pi_l_nu",
    "belle_bf_b0_to_done_to_d_pi_pi_l_nu",
    "belle_bf_b0_to_done_to_dst_pi_tau_nu",
    "belle_bf_b0_to_done_to_d_pi_pi_tau_nu",
    # Sum of old B to Dstst l nu bfs
    "sum_of_belle_b0_to_dss_l_nu_bfs",
    # Removed BFs:
    "belle_bf_b0_to_d_2s_l_nu",
    "belle_bf_b0_to_dst_2s_l_nu",
    "belle_bf_b0_to_dst_pi_zero_l_nu",
    "belle_bf_b0_to_dst_pi_minus_l_nu",
    "belle_bf_b0_to_d_pi_zero_l_nu",
    "belle_bf_b0_to_d_pi_minus_l_nu",
    "sum_of_belle_removed_b0_bfs",
]

# fmt: off
belle_decay_infos_b_zero = (
    BelleBBFEntry(decay="D*-           e+   nu_e", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0533),
    BelleBBFEntry(decay="D-            e+   nu_e", model="PHOTOS  HQET2 1.16;", bf=0.0213),
    BelleBBFEntry(decay="D_1-          e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_0*-         e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0084),
    BelleBBFEntry(decay="D'_1-         e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_2*-         e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0036),
    BelleBBFEntry(decay="D(2S)-        e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)-       e+   nu_e", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*-      pi0  e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="anti-D*0 pi-  e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0006),
    BelleBBFEntry(decay="D-       pi0  e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="anti-D0  pi-  e+   nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0010),
    BelleBBFEntry(decay="D*-           mu+  nu_mu", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0533),
    BelleBBFEntry(decay="D-            mu+  nu_mu", model="PHOTOS  HQET2 1.16;", bf=0.0213),
    BelleBBFEntry(decay="D_1-          mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_0*-         mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0084),
    BelleBBFEntry(decay="D'_1-         mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_2*-         mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0036),
    BelleBBFEntry(decay="D(2S)-        mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)-       mu+  nu_mu", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*-      pi0  mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="anti-D*0 pi-  mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0006),
    BelleBBFEntry(decay="D-       pi0  mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="anti-D0  pi-  mu+  nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0010),
    BelleBBFEntry(decay="D*-           tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0142),
    BelleBBFEntry(decay="D-            tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0071),
    BelleBBFEntry(decay="D_1-          tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="D_0*-         tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0015),
    BelleBBFEntry(decay="D'_1-         tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="D_2*-         tau+ nu_tau", model="PHOTOS ISGW2;", bf=0.0008),
    BelleBBFEntry(decay="eta_c K_S0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000580),
    BelleBBFEntry(decay="eta_c K_L0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000580),
    BelleBBFEntry(decay="K*S eta_c", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000266),
    BelleBBFEntry(decay="K*L eta_c", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000266),
    BelleBBFEntry(decay="K*0T eta_c", model="SVS;", bf=0.001068),
    BelleBBFEntry(decay="eta_c K+ pi-", model="PHOTOS PHSP;", bf=0.000100),
    BelleBBFEntry(decay="eta_c K0 pi0", model="PHSP;", bf=0.000051),
    BelleBBFEntry(decay="eta_c K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c K0 pi0 pi0", model="PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c K+ pi- pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c anti-Lambda0 n0", model="PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c anti-Sigma- p+", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c Xsd", model="PHSP;", bf=0.001193),
    BelleBBFEntry(decay="eta_c(2S) K_S0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000240),
    BelleBBFEntry(decay="eta_c(2S) K_L0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000240),
    BelleBBFEntry(decay="K*S eta_c(2S)", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000106),
    BelleBBFEntry(decay="K*L eta_c(2S)", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000106),
    BelleBBFEntry(decay="K*0T eta_c(2S)", model="SVS;", bf=0.000428),
    BelleBBFEntry(decay="eta_c(2S) K+ pi-", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) K0 pi0 pi0", model="PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) K+ pi- pi0", model="PHOTOS PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) Xsd", model="PHSP;", bf=0.000475),
    BelleBBFEntry(decay="J/psi K_S0", model="SVS_CP beta dm -1 1 0 1 0;", bf=0.000425),
    BelleBBFEntry(decay="J/psi K_L0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000425),
    BelleBBFEntry(decay="J/psi K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000218),
    BelleBBFEntry(decay="J/psi K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000218),
    BelleBBFEntry(decay="J/psi K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000874),
    BelleBBFEntry(decay="J/psi K+ pi-", model="PHOTOS PHSP; #?????", bf=0.000000),
    BelleBBFEntry(decay="J/psi K_10", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001300),
    BelleBBFEntry(decay="J/psi K'_10", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K0 rho0", model="PHSP;", bf=0.000540),
    BelleBBFEntry(decay="J/psi K0 pi+ pi-", model="PHOTOS  PHSP;", bf=0.000460),
    BelleBBFEntry(decay="J/psi K*+ pi-", model="PHOTOS  PHSP;", bf=0.000800),
    BelleBBFEntry(decay="J/psi K*0 pi+ pi-", model="PHOTOS  PHSP;", bf=0.000660),
    BelleBBFEntry(decay="J/psi phi K0", model="PHSP;", bf=0.000094),
    BelleBBFEntry(decay="J/psi pi0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000022),
    BelleBBFEntry(decay="J/psi eta", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000009),
    BelleBBFEntry(decay="J/psi pi+ pi-", model="PHOTOS PHSP;", bf=0.000030),
    BelleBBFEntry(decay="J/psi rho0", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000016),
    BelleBBFEntry(decay="J/psi omega", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi K_0*0", model="SVS;", bf=0.000050),
    BelleBBFEntry(decay="J/psi K_2*0", model="PHSP;", bf=0.000260),
    BelleBBFEntry(decay="J/psi K''*0", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000005),
    BelleBBFEntry(decay="J/psi anti-Lambda0 n0", model="PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi anti-Sigma- p+", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi Xsd", model="PHSP;", bf=0.001109),
    BelleBBFEntry(decay="psi(2S) K_S0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000310),
    BelleBBFEntry(decay="psi(2S) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000310),
    BelleBBFEntry(decay="psi(2S) K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000120),
    BelleBBFEntry(decay="psi(2S) K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000120),
    BelleBBFEntry(decay="psi(2S) K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000480),
    BelleBBFEntry(decay="psi(2S) pi0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000020),
    BelleBBFEntry(decay="psi(2S) Xsd", model="PHSP;", bf=0.001607),
    BelleBBFEntry(decay="chi_c0 K_S0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000250),
    BelleBBFEntry(decay="chi_c0 K_L0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000250),
    BelleBBFEntry(decay="K*S chi_c0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000110),
    BelleBBFEntry(decay="K*L chi_c0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000110),
    BelleBBFEntry(decay="K*0T chi_c0", model="SVS;", bf=0.000440),
    BelleBBFEntry(decay="chi_c0  Xsd", model="PHSP;", bf=0.000967),
    BelleBBFEntry(decay="chi_c1 K_S0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000200),
    BelleBBFEntry(decay="chi_c1 K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000200),
    BelleBBFEntry(decay="chi_c1 K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000068),
    BelleBBFEntry(decay="chi_c1 K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000068),
    BelleBBFEntry(decay="chi_c1 K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000274),
    BelleBBFEntry(decay="chi_c1 pi0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000018),
    BelleBBFEntry(decay="chi_c1 K+ pi-", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c1 K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c1 K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1 K0 pi0 pi0", model="PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1 K+ pi- pi0", model="PHOTOS PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1  Xsd", model="PHSP;", bf=0.000900),
    BelleBBFEntry(decay="chi_c2 K_S0", model="STS_CP beta dm  1 1 0 1 0.0;", bf=0.000038),
    BelleBBFEntry(decay="chi_c2 K_L0", model="STS_CP beta dm -1 1 0 1 0.0;", bf=0.000038),
    BelleBBFEntry(decay="chi_c2 K*S", model="PHSP;", bf=0.000019),
    BelleBBFEntry(decay="chi_c2 K*L", model="PHSP;", bf=0.000019),
    BelleBBFEntry(decay="chi_c2 K*0T", model="PHSP;", bf=0.000076),
    BelleBBFEntry(decay="chi_c2 K+ pi-", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c2 K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c2 K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000268),
    BelleBBFEntry(decay="chi_c2 K0 pi0 pi0", model="PHSP;", bf=0.000268),
    BelleBBFEntry(decay="chi_c2 K+ pi- pi0", model="PHOTOS PHSP;", bf=0.000269),
    BelleBBFEntry(decay="chi_c2  Xsd", model="PHSP;", bf=0.000600),
    BelleBBFEntry(decay="h_c K_S0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000090),
    BelleBBFEntry(decay="h_c K_L0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000090),
    BelleBBFEntry(decay="h_c K*S", model="PHSP;", bf=0.000034),
    BelleBBFEntry(decay="h_c K*L", model="PHSP;", bf=0.000034),
    BelleBBFEntry(decay="h_c K*0T", model="PHSP;", bf=0.000137),
    BelleBBFEntry(decay="h_c  Xsd", model="PHSP;", bf=0.000960),
    BelleBBFEntry(decay="X(3872) K_S0", model="SVS_CP beta dm  -1 1 0 1 0.0;", bf=0.000040),
    BelleBBFEntry(decay="X(3872) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000040),
    BelleBBFEntry(decay="Y(3940) K_S0", model="SVS_CP beta dm  -1 1 0 1 0.0;", bf=0.000035),
    BelleBBFEntry(decay="Y(3940) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000035),
    BelleBBFEntry(decay="D-   D_s+", model="PHOTOS PHSP;", bf=0.00650),
    BelleBBFEntry(decay="D*-  D_s+", model="PHOTOS SVS;", bf=0.00800),
    BelleBBFEntry(decay="D_s*+ D-", model="PHOTOS SVS;", bf=0.00740),
    BelleBBFEntry(decay="D_s*+ D*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01770),
    BelleBBFEntry(decay="D_s+ D_0*-", model="PHOTOS PHSP;", bf=0.00033),
    BelleBBFEntry(decay="D_s*+ D_0*-", model="PHOTOS SVS;", bf=0.00067),
    BelleBBFEntry(decay="D'_1- D_s+", model="PHOTOS SVS;", bf=0.00031),
    BelleBBFEntry(decay="D'_1-  D_s*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00063),
    BelleBBFEntry(decay="D_1-  D_s+", model="PHOTOS SVS;", bf=0.00063),
    BelleBBFEntry(decay="D_1-  D_s*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00126),
    BelleBBFEntry(decay="D_2*-  D_s+", model="PHOTOS STS;", bf=0.00165),
    BelleBBFEntry(decay="D_2*-  D_s*+", model="PHOTOS PHSP;", bf=0.00329),
    BelleBBFEntry(decay="D- D_s0*+", model="PHOTOS PHSP;", bf=0.00097),
    BelleBBFEntry(decay="D*- D_s0*+", model="PHOTOS SVS;", bf=0.00150),
    BelleBBFEntry(decay="D'_s1+ D-", model="PHOTOS SVS;", bf=0.00350),
    BelleBBFEntry(decay="D'_s1+ D*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00930),
    BelleBBFEntry(decay="D_s+  D-  pi0", model="PHOTOS PHSP;", bf=0.00091),
    BelleBBFEntry(decay="D_s+  anti-D0 pi-", model="PHOTOS PHSP;", bf=0.00182),
    BelleBBFEntry(decay="D_s*+ D-  pi0", model="PHOTOS PHSP;", bf=0.00091),
    BelleBBFEntry(decay="D_s*+ anti-D0 pi-", model="PHOTOS PHSP;", bf=0.00182),
    BelleBBFEntry(decay="D_s+  D-  pi-  pi+", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s+  D-  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s+  anti-D0 pi-  pi0", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s*+ D-  pi-  pi+", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s*+ D-  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*+ anti-D0 pi-  pi0", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s-  pi+  K0", model="PHOTOS PHSP;", bf=0.00095),
    BelleBBFEntry(decay="D_s*- pi+  K0", model="PHOTOS PHSP;", bf=0.00095),
    BelleBBFEntry(decay="D_s-  pi+  K0  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*- pi+  K0  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s-  pi+  K+  pi-", model="PHOTOS PHSP;", bf=0.00080),
    BelleBBFEntry(decay="D_s*- pi+  K+  pi-", model="PHOTOS PHSP;", bf=0.00080),
    BelleBBFEntry(decay="D_s-  pi+  K0  pi0 pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*- pi+  K0  pi0 pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s-  pi+  K0  pi+ pi-", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*- pi+  K0  pi+ pi-", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s-  pi+  K+  pi- pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*- pi+  K+  pi- pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D-   D+", model="PHOTOS SSS_CP beta_cdcd  dm  1 1.0 0.0 1.0 0.0;", bf=0.00023),
    BelleBBFEntry(decay="D*+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.00061),
    BelleBBFEntry(decay="D*-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.00061),
    BelleBBFEntry(decay="D*+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.00082),
    BelleBBFEntry(decay="D'_1-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000019),
    BelleBBFEntry(decay="D'_1+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000019),
    BelleBBFEntry(decay="D'_1-  D*+", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D'_1+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1-  D*+", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000079),
    BelleBBFEntry(decay="D_1+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000079),
    BelleBBFEntry(decay="D_2*-  D+", model="PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;", bf=0.000129),
    BelleBBFEntry(decay="D_2*+  D-", model="PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;", bf=0.000129),
    BelleBBFEntry(decay="D_2*-  D*+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_2*+  D*-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_s1+ D-", model="PHOTOS SVS;", bf=0.00043),
    BelleBBFEntry(decay="D_s1+ D*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00083),
    BelleBBFEntry(decay="D_s2*+ D-", model="PHOTOS STS;", bf=0.00038),
    BelleBBFEntry(decay="D_s2*+ D*-", model="PHOTOS PHSP;", bf=0.00038),
    BelleBBFEntry(decay="D_sj(2700)+ D-", model="PHOTOS SVS;", bf=0.00113),
    BelleBBFEntry(decay="D_sj(2700)+ D*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00038),
    BelleBBFEntry(decay="psi(3770) K_S0", model="PHOTOS SVS_CP beta dm  1 1 0 1 0.0;", bf=0.00023),
    BelleBBFEntry(decay="psi(3770) K_L0", model="PHOTOS SVS_CP beta dm -1 1 0 1 0.0;", bf=0.00023),
    BelleBBFEntry(decay="psi(3770) K*S", model="PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.00012),
    BelleBBFEntry(decay="psi(3770) K*L", model="PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.00012),
    BelleBBFEntry(decay="psi(3770) K*0T", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.00047),
    BelleBBFEntry(decay="psi(3770) K+ pi-", model="PHOTOS PHSP;", bf=0.00058),
    BelleBBFEntry(decay="psi(3770) K0 pi0", model="PHSP;", bf=0.00029),
    BelleBBFEntry(decay="psi(3770) K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) K0 pi0 pi0", model="PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) K+ pi- pi0", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(4040) K_S0", model="PHOTOS SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4040) K_L0", model="PHOTOS SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4040) K*S", model="PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4040) K*L", model="PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4040) K*0T", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001400),
    BelleBBFEntry(decay="psi(4040) K+ pi-", model="PHOTOS PHSP;", bf=0.000741),
    BelleBBFEntry(decay="psi(4040) K0 pi0", model="PHSP;", bf=0.000371),
    BelleBBFEntry(decay="psi(4160) K_S0", model="PHOTOS SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4160) K_L0", model="PHOTOS SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4160) K*S", model="PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4160) K*L", model="PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4160) K*0T", model="PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001400),
    BelleBBFEntry(decay="psi(4160) K+ pi-", model="PHOTOS PHSP;", bf=0.000741),
    BelleBBFEntry(decay="psi(4160) K0 pi0", model="PHSP;", bf=0.000371),
    BelleBBFEntry(decay="D+   D-   pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   anti-D0  pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D0   D-  pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   D-   pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*+   anti-D0  pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*0   D-  pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+   D*-   pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+   anti-D*0  pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0   D*-  pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*+   D*-   pi0", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*+   anti-D*0  pi-", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*0   D*-  pi+", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D+   D-   rho0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   anti-D0  rho-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   D-  rho+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*+   D-   rho0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   anti-D0  rho-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*0   D-  rho+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   D*-   rho0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   anti-D*0  rho-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D0   D*-  rho+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   D*-   rho0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*+   anti-D*0  rho-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*0   D*-  rho+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+   D-   pi0  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   D-   pi-  pi+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D+   anti-D0  pi-  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   D-  pi+  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D0   anti-D0  pi-  pi+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*+   D-   pi0  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   D-   pi-  pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   anti-D0  pi-  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*0   D-  pi+  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*0   anti-D0  pi-  pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   D*-   pi0  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   D*-   pi-  pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D+   anti-D*0  pi-  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D0   D*-  pi+  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D0   anti-D*0  pi-  pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*+   D*-   pi0  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*+   D*-   pi-  pi+", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*+   anti-D*0  pi-  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*0   D*-  pi+  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*0   anti-D*0  pi-  pi+", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D-  D0  K+", model="PHOTOS PHSP;", bf=0.001700),
    BelleBBFEntry(decay="D-  D+   K0", model="PHOTOS PHSP;", bf=0.001370),
    BelleBBFEntry(decay="D-  D0  K*+", model="PHOTOS PHSP;", bf=0.000455),
    BelleBBFEntry(decay="D-  D+   K*0", model="PHOTOS PHSP;", bf=0.000455),
    BelleBBFEntry(decay="D*- D0  K+", model="PHOTOS PHSP;", bf=0.003100),
    BelleBBFEntry(decay="D*- D+  K0", model="PHOTOS PHSP;", bf=0.006500),
    BelleBBFEntry(decay="D-  D*0  K+", model="PHOTOS PHSP;", bf=0.004600),
    BelleBBFEntry(decay="D*- D0   K*+", model="PHOTOS PHSP;", bf=0.000910),
    BelleBBFEntry(decay="D*- D+   K*0", model="PHOTOS PHSP;", bf=0.001820),
    BelleBBFEntry(decay="D-  D*0  K*+", model="PHOTOS PHSP;", bf=0.001365),
    BelleBBFEntry(decay="D-  D*+  K*0", model="PHOTOS PHSP;", bf=0.001365),
    BelleBBFEntry(decay="D*- D*0  K+", model="PHOTOS PHSP;", bf=0.011800),
    BelleBBFEntry(decay="D*- D*+   K0", model="PHOTOS PHSP;", bf=0.008800),
    BelleBBFEntry(decay="D*- D*0   K*+", model="PHOTOS PHSP;", bf=0.003185),
    BelleBBFEntry(decay="D*- D*+   K*0", model="PHOTOS PHSP;", bf=0.003185),
    BelleBBFEntry(decay="D_1- D0   K+", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_1- D+   K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D-  D_10  K+", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D-  D_1+  K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_1- D*0  K+", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_1- D*+  K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*- D_10  K+", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*- D_1+  K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_2*- D0   K+", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_2*- D+   K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D-  D_2*0  K+", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D-  D_2*+  K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_2*- D*0  K+", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_2*- D*+  K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*- D_2*0  K+", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*- D_2*+  K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D-  D0   K+  pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="anti-D0  D0   K+  pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-  D0   K0  pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-  D+   K+ pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-  D+   K0 pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="anti-D0  D+   K0  pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*-  D0   K+  pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="anti-D*0  D0   K+  pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*-  D0   K0  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*-  D+   K+ pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*-  D+   K0 pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="anti-D*0  D+   K0  pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D-  D*0   K+  pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="anti-D0  D*0   K+  pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D-  D*0   K0  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D-  D*+   K+ pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D-  D*+   K0 pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="anti-D0  D*+   K0  pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*-  D*0   K+  pi0", model="PHOTOS PHSP;", bf=0.000384),
    BelleBBFEntry(decay="anti-D*0  D*0   K+  pi-", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*-  D*0   K0  pi+", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*-  D*+   K+ pi-", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*-  D*+   K0 pi0", model="PHOTOS PHSP;", bf=0.000384),
    BelleBBFEntry(decay="anti-D*0  D*+   K0  pi-", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D-  D0   K*+  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D0   K*+  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D0   K*0  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+   K*+ pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+   K*0 pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+   K*0  pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D0   K+  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D0   K+  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D0   K0  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+   K+ rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D-  D+   K0 rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="anti-D0  D+   K0  rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  K0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D0  anti-D0  K*0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D*0  K0", model="PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D*0  anti-D0  K0", model="PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D0  anti-D*0  K*0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D*0  anti-D0  K*0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D*0  anti-D*0  K0", model="PHSP;", bf=0.002548),
    BelleBBFEntry(decay="D*0  anti-D*0  K*0", model="PHSP;", bf=0.000819),
    BelleBBFEntry(decay="D_10  anti-D0  K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D_10  K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D_10  anti-D*0  K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*0  anti-D_10  K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D_2*0  anti-D0  K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D_2*0  K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D_2*0  anti-D*0  K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*0  anti-D_2*0  K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D-  pi+", model="PHOTOS PHSP;", bf=0.00268),
    BelleBBFEntry(decay="D-  pi+  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="rho+ D-", model="PHOTOS SVS;", bf=0.0075),
    BelleBBFEntry(decay="D-  pi-  pi+  pi+", model="PHOTOS PHSP;", bf=0.004705),
    BelleBBFEntry(decay="D-  pi0  pi+  pi0", model="PHOTOS PHSP;", bf=0.0009),
    BelleBBFEntry(decay="D-  rho0  pi+", model="PHOTOS PHSP;", bf=0.0014),
    BelleBBFEntry(decay="D-  rho+  pi0", model="PHOTOS PHSP;", bf=0.0009),
    BelleBBFEntry(decay="a_1+ D-", model="PHOTOS SVS;", bf=0.0075),
    BelleBBFEntry(decay="b_1+ D-", model="PHOTOS SVS;", bf=0.0000),
    BelleBBFEntry(decay="D-   pi-  pi+  pi+  pi0", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="D-   omega  pi+", model="PHOTOS PHSP;", bf=0.0028),
    BelleBBFEntry(decay="D-   p+  anti-p-  pi+", model="PHOTOS PHSP;", bf=0.00034),
    BelleBBFEntry(decay="D-   p+  anti-n0", model="PHOTOS PHSP;", bf=0.00150),
    BelleBBFEntry(decay="D*- pi+", model="PHOTOS SVS;", bf=0.00276),
    BelleBBFEntry(decay="D*- pi+  pi0", model="PHOTOS PHSP;", bf=0.00060),
    BelleBBFEntry(decay="rho+ D*-", model="PHOTOS SVV_HELAMP 0.152 1.47 0.936 0 0.317 0.19;", bf=0.00680),
    BelleBBFEntry(decay="D*-  pi-  pi+  pi+", model="PHOTOS PHSP;", bf=0.0011),
    BelleBBFEntry(decay="D*-  pi0  pi+  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D*-  rho0  pi+", model="PHOTOS PHSP;", bf=0.0072),
    BelleBBFEntry(decay="D*-  rho+  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D*-  a_1+", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0130),
    BelleBBFEntry(decay="D*-  b_1+", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0000),
    BelleBBFEntry(decay="D*-  pi-  pi+  pi+  pi0", model="PHOTOS PHSP;", bf=0.0176),
    BelleBBFEntry(decay="D*-  omega  pi+", model="PHOTOS PHSP;", bf=0.0029),
    BelleBBFEntry(decay="D*-   p+  anti-p-  pi+", model="PHOTOS PHSP;", bf=0.00065),
    BelleBBFEntry(decay="D*-   p+  anti-n0", model="PHOTOS PHSP;", bf=0.00150),
    BelleBBFEntry(decay="D_1- pi+", model="PHOTOS SVS;", bf=0.00093),
    BelleBBFEntry(decay="D_2*- pi+", model="PHOTOS STS;", bf=0.00085),
    BelleBBFEntry(decay="D_1- rho+", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="D_2*- rho+", model="PHOTOS PHSP;", bf=0.00085),
    BelleBBFEntry(decay="anti-D_2*0 pi0", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="anti-D_10 pi0", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="anti-D'_10 pi0", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="anti-D_0*0 pi0", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="anti-D_2*0 eta", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="anti-D_10 eta", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="anti-D'_10 eta", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="anti-D_0*0 eta", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="anti-D_2*0 eta'", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="anti-D_10 eta'", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="anti-D'_10 eta'", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="anti-D_0*0 eta'", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="anti-D_2*0 omega", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="anti-D_10 omega", model="PHSP;", bf=0.00009),
    BelleBBFEntry(decay="anti-D'_10 omega", model="PHSP;", bf=0.00007),
    BelleBBFEntry(decay="anti-D_0*0 omega", model="PHSP;", bf=0.00008),
    BelleBBFEntry(decay="anti-D_2*0 rho0", model="PHSP;", bf=0.00006),
    BelleBBFEntry(decay="anti-D_10 rho0", model="PHSP;", bf=0.00011),
    BelleBBFEntry(decay="anti-D'_10 rho0", model="PHSP;", bf=0.00008),
    BelleBBFEntry(decay="anti-D_0*0 rho0", model="PHSP;", bf=0.00010),
    BelleBBFEntry(decay="anti-D0 pi0", model="PHSP;", bf=0.00026),
    BelleBBFEntry(decay="anti-D0  pi-  pi+", model="PHOTOS PHSP;", bf=0.00045),
    BelleBBFEntry(decay="rho0 anti-D0", model="SVS;", bf=0.00029),
    BelleBBFEntry(decay="anti-D0  pi0  pi0", model="PHSP;", bf=0.0005),
    BelleBBFEntry(decay="anti-D0 eta", model="PHSP;", bf=0.00020),
    BelleBBFEntry(decay="anti-D0 eta'", model="PHSP;", bf=0.00013),
    BelleBBFEntry(decay="omega anti-D0", model="SVS;", bf=0.00026),
    BelleBBFEntry(decay="anti-D0  pi-  pi+  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="anti-D0  pi0  pi0  pi0", model="PHSP;", bf=0.0001),
    BelleBBFEntry(decay="anti-D0 p+ anti-p-", model="PHOTOS PHSP;", bf=0.000114),
    BelleBBFEntry(decay="anti-D*0 pi0", model="SVS;", bf=0.00017),
    BelleBBFEntry(decay="anti-D*0 pi-  pi+", model="PHOTOS PHSP;", bf=0.00032),
    BelleBBFEntry(decay="anti-D*0 rho0", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00029),
    BelleBBFEntry(decay="anti-D*0 pi0  pi0", model="PHSP;", bf=0.0005),
    BelleBBFEntry(decay="anti-D*0 eta", model="SVS;", bf=0.00018),
    BelleBBFEntry(decay="anti-D*0 eta'", model="SVS;", bf=0.00012),
    BelleBBFEntry(decay="anti-D*0 omega", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00027),
    BelleBBFEntry(decay="anti-D*0  pi-  pi+  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="anti-D*0  pi0  pi0  pi0", model="PHSP;", bf=0.0001),
    BelleBBFEntry(decay="anti-D*0 p+ anti-p-", model="PHOTOS PHSP;", bf=0.000103),
    BelleBBFEntry(decay="D-  K+", model="PHOTOS PHSP;", bf=0.00020),
    BelleBBFEntry(decay="K*+ D-", model="PHOTOS SVS;", bf=0.00045),
    BelleBBFEntry(decay="D*- K+", model="PHOTOS SVS;", bf=0.00021),
    BelleBBFEntry(decay="D*- K*+", model="PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00033),
    BelleBBFEntry(decay="anti-D0 K0", model="PHSP;", bf=0.000052),
    BelleBBFEntry(decay="K*0 anti-D0", model="SVS;", bf=0.000042),
    BelleBBFEntry(decay="anti-D*0 K0", model="SVS;", bf=0.000036),
    BelleBBFEntry(decay="anti-D*0 K*0", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.000042),
    BelleBBFEntry(decay="u  anti-s  anti-c  d", model="PHOTOS PYTHIA 48;", bf=0.024297),
    BelleBBFEntry(decay="u  anti-d  anti-c  d", model="PHOTOS PYTHIA 48;", bf=0.372909),
    BelleBBFEntry(decay="anti-Lambda_c- p+", model="PHOTOS PHSP;", bf=0.000022),
    BelleBBFEntry(decay="anti-Sigma_c-  p+", model="PHOTOS PHSP;", bf=0.000022),
    BelleBBFEntry(decay="anti-Sigma_c0  n0", model="PHSP;", bf=0.000022),
    BelleBBFEntry(decay="anti-cd_0       ud_0", model="PHOTOS PYTHIA 23;", bf=0.018000),
    BelleBBFEntry(decay="anti-cd_1       ud_1", model="PHOTOS PYTHIA 23;", bf=0.018000),
    BelleBBFEntry(decay="anti-cd_0       ud_0", model="PHOTOS PYTHIA 24;", bf=0.006000),
    BelleBBFEntry(decay="anti-cd_1       ud_1", model="PHOTOS PYTHIA 24;", bf=0.005932),
    BelleBBFEntry(decay="anti-Xi_c-     Lambda_c+", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-Xi_c-     Sigma_c+", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-Xi_c0     Sigma_c0", model="PHSP;", bf=0.001000),
    BelleBBFEntry(decay="anti-cs_0      cd_0", model="PHOTOS PYTHIA 23;", bf=0.009000),
    BelleBBFEntry(decay="anti-Xi_c-     p+", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="anti-Xi_c0     n0", model="PHSP;", bf=0.000001),
    BelleBBFEntry(decay="anti-cs_0      ud_0", model="PHOTOS PYTHIA 23;", bf=0.000900),
    BelleBBFEntry(decay="anti-cs_1      ud_1", model="PHOTOS PYTHIA 23;", bf=0.000900),
    BelleBBFEntry(decay="anti-cs_0      ud_0", model="PHOTOS PYTHIA 24;", bf=0.000300),
    BelleBBFEntry(decay="anti-cs_1      ud_1", model="PHOTOS PYTHIA 24;", bf=0.000300),
    BelleBBFEntry(decay="anti-Lambda_c-    Lambda_c+", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Sigma_c-     Lambda_c+", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Lambda_c-    Sigma_c+", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Sigma_c-     Sigma_c+", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-Sigma_c0     Sigma_c0", model="PHSP;", bf=0.000040),
    BelleBBFEntry(decay="anti-cd_0    cd_0", model="PHOTOS PYTHIA 23;", bf=0.000400),
)  # type: Tuple[BelleBBFEntry, ...]


belle_decay_infos_anti_b_zero = (
    BelleBBFEntry(decay="D*+       e-   anti-nu_e", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0533),
    BelleBBFEntry(decay="D+        e-   anti-nu_e", model="PHOTOS  HQET2 1.16;", bf=0.0213),
    BelleBBFEntry(decay="D_1+      e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_0*+     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0084),
    BelleBBFEntry(decay="D'_1+     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_2*+     e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0036),
    BelleBBFEntry(decay="D(2S)+    e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)+   e-   anti-nu_e", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*+  pi0  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*0  pi+  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0006),
    BelleBBFEntry(decay="D+   pi0  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D0   pi+  e-   anti-nu_e", model="PHOTOS  GOITY_ROBERTS;", bf=0.0010),
    BelleBBFEntry(decay="D*+       mu-  anti-nu_mu", model="PHOTOS  HQET2 1.3 1.18 0.71;", bf=0.0533),
    BelleBBFEntry(decay="D+        mu-  anti-nu_mu", model="PHOTOS  HQET2 1.16;", bf=0.0213),
    BelleBBFEntry(decay="D_1+      mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_0*+     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0084),
    BelleBBFEntry(decay="D'_1+     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0074),
    BelleBBFEntry(decay="D_2*+     mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0036),
    BelleBBFEntry(decay="D(2S)+    mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0002),
    BelleBBFEntry(decay="D*(2S)+   mu-  anti-nu_mu", model="PHOTOS  ISGW2;", bf=0.0005),
    BelleBBFEntry(decay="D*+  pi0  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0003),
    BelleBBFEntry(decay="D*0  pi+  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0006),
    BelleBBFEntry(decay="D+   pi0  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;", bf=0.0005),
    BelleBBFEntry(decay="D0   pi+  mu-  anti-nu_mu", model="PHOTOS  GOITY_ROBERTS;# ====B->DXlnu End", bf=0.0010),
    BelleBBFEntry(decay="D*+       tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0142),
    BelleBBFEntry(decay="D+        tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0071),
    BelleBBFEntry(decay="D_1+      tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="D_0*+     tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0015),
    BelleBBFEntry(decay="D'_1+     tau- anti-nu_tau", model="PHOTOS ISGW2;", bf=0.0012),
    BelleBBFEntry(decay="D_2*+     tau- anti-nu_tau", model="PHOTOS ISGW2;# ====B->DXtaunu End", bf=0.0008),
    BelleBBFEntry(decay="eta_c K_S0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000580),
    BelleBBFEntry(decay="eta_c K_L0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000580),
    BelleBBFEntry(decay="anti-K*S eta_c", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000266),
    BelleBBFEntry(decay="anti-K*L eta_c", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000266),
    BelleBBFEntry(decay="anti-K*0T eta_c", model="SVS;", bf=0.001068),
    BelleBBFEntry(decay="eta_c K- pi+", model="PHOTOS PHSP;", bf=0.000100),
    BelleBBFEntry(decay="eta_c anti-K0 pi0", model="PHSP;", bf=0.000051),
    BelleBBFEntry(decay="eta_c anti-K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c anti-K0 pi0 pi0", model="PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c K- pi+ pi0", model="PHOTOS PHSP;", bf=0.000308),
    BelleBBFEntry(decay="eta_c Lambda0 anti-n0", model="PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c Sigma+ anti-p-", model="PHOTOS PHSP;", bf=0.000003),
    BelleBBFEntry(decay="eta_c anti-Xsd", model="PHSP;", bf=0.001193),
    BelleBBFEntry(decay="eta_c(2S) K_S0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000240),
    BelleBBFEntry(decay="eta_c(2S) K_L0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000240),
    BelleBBFEntry(decay="anti-K*S eta_c(2S)", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000106),
    BelleBBFEntry(decay="anti-K*L eta_c(2S)", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000106),
    BelleBBFEntry(decay="anti-K*0T eta_c(2S)", model="SVS;", bf=0.000428),
    BelleBBFEntry(decay="eta_c(2S) K- pi+", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) anti-K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="eta_c(2S) anti-K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) anti-K0 pi0 pi0", model="PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) K- pi+ pi0", model="PHOTOS PHSP;", bf=0.000140),
    BelleBBFEntry(decay="eta_c(2S) anti-Xsd", model="PHSP;", bf=0.000475),
    BelleBBFEntry(decay="J/psi K_S0", model="SVS_CP beta dm -1 1 0 1 0;", bf=0.000425),
    BelleBBFEntry(decay="J/psi K_L0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000425),
    BelleBBFEntry(decay="J/psi anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000218),
    BelleBBFEntry(decay="J/psi anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000218),
    BelleBBFEntry(decay="J/psi anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000874),
    BelleBBFEntry(decay="J/psi K- pi+", model="PHOTOS PHSP; #?????", bf=0.000000),
    BelleBBFEntry(decay="J/psi anti-K_10", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001300),
    BelleBBFEntry(decay="J/psi anti-K'_10", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi anti-K0 rho0", model="PHSP;", bf=0.000540),
    BelleBBFEntry(decay="J/psi anti-K0 pi+ pi-", model="PHOTOS  PHSP;", bf=0.000460),
    BelleBBFEntry(decay="J/psi K*- pi+", model="PHOTOS  PHSP;", bf=0.000800),
    BelleBBFEntry(decay="J/psi anti-K*0 pi+ pi-", model="PHOTOS  PHSP;", bf=0.000660),
    BelleBBFEntry(decay="J/psi phi anti-K0", model="PHSP;", bf=0.000094),
    BelleBBFEntry(decay="J/psi pi0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000022),
    BelleBBFEntry(decay="J/psi eta", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000009),
    BelleBBFEntry(decay="J/psi pi+ pi-", model="PHOTOS PHSP;", bf=0.000030),
    BelleBBFEntry(decay="J/psi rho0", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000016),
    BelleBBFEntry(decay="J/psi omega", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000000),
    BelleBBFEntry(decay="J/psi anti-K_0*0", model="SVS;", bf=0.000050),
    BelleBBFEntry(decay="J/psi anti-K_2*0", model="PHSP;", bf=0.000260),
    BelleBBFEntry(decay="J/psi anti-K''*0", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000005),
    BelleBBFEntry(decay="J/psi Lambda0 anti-n0", model="PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi Sigma+ anti-p-", model="PHOTOS PHSP;", bf=0.000012),
    BelleBBFEntry(decay="J/psi anti-Xsd", model="PHSP;", bf=0.001109),
    BelleBBFEntry(decay="psi(2S) K_S0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000310),
    BelleBBFEntry(decay="psi(2S) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000310),
    BelleBBFEntry(decay="psi(2S) anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000120),
    BelleBBFEntry(decay="psi(2S) anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000120),
    BelleBBFEntry(decay="psi(2S) anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000480),
    BelleBBFEntry(decay="psi(2S) pi0", model="SVS_CP beta dm  1 1 0 1 0;", bf=0.000020),
    BelleBBFEntry(decay="psi(2S) anti-Xsd", model="PHSP;", bf=0.001607),
    BelleBBFEntry(decay="chi_c0 K_S0", model="SSS_CP beta dm  1 1 0 1 0.0;", bf=0.000250),
    BelleBBFEntry(decay="chi_c0 K_L0", model="SSS_CP beta dm -1 1 0 1 0.0;", bf=0.000250),
    BelleBBFEntry(decay="anti-K*S chi_c0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000110),
    BelleBBFEntry(decay="anti-K*L chi_c0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000110),
    BelleBBFEntry(decay="anti-K*0T chi_c0", model="SVS;", bf=0.000440),
    BelleBBFEntry(decay="chi_c0  anti-Xsd", model="PHSP;", bf=0.000967),
    BelleBBFEntry(decay="chi_c1 K_S0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000200),
    BelleBBFEntry(decay="chi_c1 K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000200),
    BelleBBFEntry(decay="chi_c1 anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000068),
    BelleBBFEntry(decay="chi_c1 anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000068),
    BelleBBFEntry(decay="chi_c1 anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.000274),
    BelleBBFEntry(decay="chi_c1 pi0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000018),
    BelleBBFEntry(decay="chi_c1 K- pi+", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c1 anti-K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c1 anti-K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1 anti-K0 pi0 pi0", model="PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1 K- pi+ pi0", model="PHOTOS PHSP;", bf=0.000500),
    BelleBBFEntry(decay="chi_c1  anti-Xsd", model="PHSP;", bf=0.000900),
    BelleBBFEntry(decay="chi_c2 K_S0", model="STS_CP beta dm  1 1 0 1 0.0;", bf=0.000038),
    BelleBBFEntry(decay="chi_c2 K_L0", model="STS_CP beta dm -1 1 0 1 0.0;", bf=0.000038),
    BelleBBFEntry(decay="chi_c2 anti-K*S", model="PHSP;", bf=0.000019),
    BelleBBFEntry(decay="chi_c2 anti-K*L", model="PHSP;", bf=0.000019),
    BelleBBFEntry(decay="chi_c2 anti-K*0T", model="PHSP;", bf=0.000076),
    BelleBBFEntry(decay="chi_c2 K- pi+", model="PHOTOS PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c2 anti-K0 pi0", model="PHSP;", bf=0.000000),
    BelleBBFEntry(decay="chi_c2 anti-K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.000268),
    BelleBBFEntry(decay="chi_c2 anti-K0 pi0 pi0", model="PHSP;", bf=0.000268),
    BelleBBFEntry(decay="chi_c2 K- pi+ pi0", model="PHOTOS PHSP;", bf=0.000269),
    BelleBBFEntry(decay="chi_c2 anti-Xsd", model="PHSP;", bf=0.000600),
    BelleBBFEntry(decay="h_c K_S0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000090),
    BelleBBFEntry(decay="h_c K_L0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000090),
    BelleBBFEntry(decay="h_c anti-K*S", model="PHSP;", bf=0.000034),
    BelleBBFEntry(decay="h_c anti-K*L", model="PHSP;", bf=0.000034),
    BelleBBFEntry(decay="h_c anti-K*0T", model="PHSP;", bf=0.000137),
    BelleBBFEntry(decay="h_c  anti-Xsd", model="PHSP;", bf=0.000960),
    BelleBBFEntry(decay="X(3872) K_S0", model="SVS_CP beta dm  -1 1 0 1 0.0;", bf=0.000040),
    BelleBBFEntry(decay="X(3872) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000040),
    BelleBBFEntry(decay="Y(3940) K_S0", model="SVS_CP beta dm  -1 1 0 1 0.0;", bf=0.000035),
    BelleBBFEntry(decay="Y(3940) K_L0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000035),
    BelleBBFEntry(decay="D+  D_s-", model="PHOTOS PHSP;", bf=0.00650),
    BelleBBFEntry(decay="D*+  D_s-", model="PHOTOS SVS;", bf=0.00800),
    BelleBBFEntry(decay="D_s*- D+", model="PHOTOS SVS;", bf=0.00740),
    BelleBBFEntry(decay="D*+  D_s*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.01770),
    BelleBBFEntry(decay="D_s- D_0*+", model="PHOTOS PHSP;", bf=0.00033),
    BelleBBFEntry(decay="D_s*- D_0*+", model="PHOTOS SVS;", bf=0.00067),
    BelleBBFEntry(decay="D'_1+  D_s-", model="PHOTOS SVS;", bf=0.00031),
    BelleBBFEntry(decay="D'_1+  D_s*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00063),
    BelleBBFEntry(decay="D_1+  D_s-", model="PHOTOS SVS;", bf=0.00063),
    BelleBBFEntry(decay="D_1+  D_s*-", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00126),
    BelleBBFEntry(decay="D_2*+  D_s-", model="PHOTOS STS;", bf=0.00165),
    BelleBBFEntry(decay="D_2*+  D_s*-", model="PHOTOS PHSP;", bf=0.00329),
    BelleBBFEntry(decay="D+ D_s0*-", model="PHOTOS PHSP;", bf=0.00097),
    BelleBBFEntry(decay="D*+ D_s0*-", model="PHOTOS SVS;", bf=0.00150),
    BelleBBFEntry(decay="D'_s1- D+", model="PHOTOS SVS;", bf=0.00350),
    BelleBBFEntry(decay="D'_s1- D*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00930),
    BelleBBFEntry(decay="D_s-  D+  pi0", model="PHOTOS PHSP;", bf=0.00091),
    BelleBBFEntry(decay="D_s-  D0  pi+", model="PHOTOS PHSP;", bf=0.00182),
    BelleBBFEntry(decay="D_s*- D+  pi0", model="PHOTOS PHSP;", bf=0.00091),
    BelleBBFEntry(decay="D_s*- D0  pi+", model="PHOTOS PHSP;", bf=0.00182),
    BelleBBFEntry(decay="D_s-  D+  pi+  pi-", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s-  D+  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s-  D0  pi+  pi0", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s*- D+  pi+  pi-", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s*- D+  pi0  pi0", model="PHOTOS PHSP;", bf=0.00070),
    BelleBBFEntry(decay="D_s*- D0  pi+  pi0", model="PHOTOS PHSP;", bf=0.00151),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0", model="PHOTOS PHSP;", bf=0.00095),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0", model="PHOTOS PHSP;", bf=0.00095),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0  pi0", model="PHOTOS PHSP;", bf=0.00100),
    BelleBBFEntry(decay="D_s+  pi-  K-  pi+", model="PHOTOS PHSP;", bf=0.00080),
    BelleBBFEntry(decay="D_s*+ pi-  K-  pi+", model="PHOTOS PHSP;", bf=0.00080),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0  pi0 pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0  pi0 pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s+  pi-  anti-K0  pi+ pi-", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*+ pi-  anti-K0  pi+ pi-", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s+  pi-  K-  pi+ pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D_s*+ pi-  K-  pi+ pi0", model="PHOTOS PHSP;", bf=0.00016),
    BelleBBFEntry(decay="D-   D+", model="PHOTOS SSS_CP beta_cdcd  dm  1 1.0 0.0 1.0 0.0;", bf=0.00023),
    BelleBBFEntry(decay="D*-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.00061),
    BelleBBFEntry(decay="D*+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.00061),
    BelleBBFEntry(decay="D*+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.00082),
    BelleBBFEntry(decay="D'_1-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000019),
    BelleBBFEntry(decay="D'_1+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000019),
    BelleBBFEntry(decay="D'_1-  D*+", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D'_1+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1-  D+", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1+  D-", model="PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;", bf=0.000039),
    BelleBBFEntry(decay="D_1-  D*+", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000079),
    BelleBBFEntry(decay="D_1+  D*-", model="PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;", bf=0.000079),
    BelleBBFEntry(decay="D_2*-  D+", model="PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;", bf=0.000129),
    BelleBBFEntry(decay="D_2*+  D-", model="PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;", bf=0.000129),
    BelleBBFEntry(decay="D_2*-  D*+", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_2*+  D*-", model="PHOTOS PHSP;", bf=0.000258),
    BelleBBFEntry(decay="D_s1- D+", model="PHOTOS SVS;", bf=0.00043),
    BelleBBFEntry(decay="D_s1- D*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00083),
    BelleBBFEntry(decay="D_s2*- D+", model="PHOTOS STS;", bf=0.00038),
    BelleBBFEntry(decay="D_s2*- D*+", model="PHOTOS PHSP;", bf=0.00038),
    BelleBBFEntry(decay="D_sj(2700)- D+", model="PHOTOS SVS;", bf=0.00113),
    BelleBBFEntry(decay="D_sj(2700)- D*+", model="PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;", bf=0.00038),
    BelleBBFEntry(decay="psi(3770) K_S0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.00023),
    BelleBBFEntry(decay="psi(3770) K_L0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.00023),
    BelleBBFEntry(decay="psi(3770) anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.00012),
    BelleBBFEntry(decay="psi(3770) anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.00012),
    BelleBBFEntry(decay="psi(3770) anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.00047),
    BelleBBFEntry(decay="psi(3770) K- pi+", model="PHOTOS PHSP;", bf=0.00058),
    BelleBBFEntry(decay="psi(3770) anti-K0 pi0", model="PHSP;", bf=0.00029),
    BelleBBFEntry(decay="psi(3770) anti-K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) anti-K0 pi0 pi0", model="PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(3770) K- pi+ pi0", model="PHOTOS PHSP;", bf=0.00064),
    BelleBBFEntry(decay="psi(4040) K_S0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4040) K_L0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4040) anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4040) anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4040) anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001400),
    BelleBBFEntry(decay="psi(4040) K- pi+", model="PHOTOS PHSP;", bf=0.000741),
    BelleBBFEntry(decay="psi(4040) anti-K0 pi0", model="PHSP;", bf=0.000371),
    BelleBBFEntry(decay="psi(4160) K_S0", model="SVS_CP beta dm  1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4160) K_L0", model="SVS_CP beta dm -1 1 0 1 0.0;", bf=0.000703),
    BelleBBFEntry(decay="psi(4160) anti-K*S", model="SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4160) anti-K*L", model="SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;", bf=0.000352),
    BelleBBFEntry(decay="psi(4160) anti-K*0T", model="SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;", bf=0.001400),
    BelleBBFEntry(decay="psi(4160) K- pi+", model="PHOTOS PHSP;", bf=0.000741),
    BelleBBFEntry(decay="psi(4160) anti-K0 pi0", model="PHSP;", bf=0.000371),
    BelleBBFEntry(decay="D-   D+   pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-   D0   pi+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D0   D+   pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-  D+   pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*-  D0   pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="anti-D*0  D+   pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-  D*+   pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-  D*0   pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="anti-D0  D*+   pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*-  D*+   pi0", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*-  D*0   pi+", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="anti-D*0  D*+   pi-", model="PHOTOS PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D-   D+   rho0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D0   rho+", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D+   rho-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*-  D+   rho0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-  D0   rho+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D*0  D+   rho-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-  D*+   rho0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-  D*0   rho+", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D0  D*+   rho-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-  D*+   rho0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*-  D*0   rho+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="anti-D*0  D*+   rho-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D-   D+   pi0  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D+   pi+  pi-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D-   D0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="anti-D0   D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000009),
    BelleBBFEntry(decay="D*-   D+   pi0  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-   D+   pi+  pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-   D0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D*0   D+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D*0   D0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-   D*+   pi0  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-   D*+   pi+  pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D-   D*0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D0   D*+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="anti-D0   D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000027),
    BelleBBFEntry(decay="D*-   D*+   pi0  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*-   D*+   pi+  pi-", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D*-   D*0   pi+  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="anti-D*0   D*+   pi-  pi0", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="anti-D*0   D*0   pi+  pi-", model="PHOTOS PHSP;", bf=0.000082),
    BelleBBFEntry(decay="D+  anti-D0  K-", model="PHOTOS PHSP;", bf=0.001700),
    BelleBBFEntry(decay="D+  D-   anti-K0", model="PHOTOS PHSP;", bf=0.001370),
    BelleBBFEntry(decay="D+  anti-D0  K*-", model="PHOTOS PHSP;", bf=0.000455),
    BelleBBFEntry(decay="D+  D-   anti-K*0", model="PHOTOS PHSP;", bf=0.000455),
    BelleBBFEntry(decay="D*+ anti-D0  K-", model="PHOTOS PHSP;", bf=0.003100),
    BelleBBFEntry(decay="D*+ D-   anti-K0", model="PHOTOS PHSP;", bf=0.006500),
    BelleBBFEntry(decay="D+  anti-D*0  K-", model="PHOTOS PHSP;", bf=0.004600),
    BelleBBFEntry(decay="D*+ anti-D0  K*-", model="PHOTOS PHSP;", bf=0.000910),
    BelleBBFEntry(decay="D*+ D-   anti-K*0", model="PHOTOS PHSP;", bf=0.001820),
    BelleBBFEntry(decay="D+  anti-D*0  K*-", model="PHOTOS PHSP;", bf=0.001365),
    BelleBBFEntry(decay="D+  D*-  anti-K*0", model="PHOTOS PHSP;", bf=0.001365),
    BelleBBFEntry(decay="D*+ anti-D*0  K-", model="PHOTOS PHSP;", bf=0.011800),
    BelleBBFEntry(decay="D*+ D*-   anti-K0", model="PHOTOS PHSP;", bf=0.008800),
    BelleBBFEntry(decay="D*+ anti-D*0  K*-", model="PHOTOS PHSP;", bf=0.003185),
    BelleBBFEntry(decay="D*+ D*-   anti-K*0", model="PHOTOS PHSP;", bf=0.003185),
    BelleBBFEntry(decay="D_1+ anti-D0  K-", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_1+ D-   anti-K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D+  anti-D_10  K-", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D+  D_1-  anti-K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_1+ anti-D*0  K-", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_1+ D*-   anti-K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*+ anti-D_10  K-", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*+ D_1-   anti-K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_2*+ anti-D0  K-", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_2*+ D-   anti-K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D+  anti-D_2*0  K-", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D+  D_2*-  anti-K0", model="PHOTOS PHSP;", bf=0.000118),
    BelleBBFEntry(decay="D_2*+ anti-D*0  K-", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D_2*+ D*-   anti-K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*+ anti-D_2*0  K-", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D*+ D_2*-   anti-K0", model="PHOTOS PHSP;", bf=0.000355),
    BelleBBFEntry(decay="D+  anti-D0  K-   pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D0  K-   pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+  anti-D0  anti-K0   pi-", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+  D-   K-  pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D+  D-   anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  D-   anti-K0  pi+", model="PHOTOS PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D*+  anti-D0  K-   pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*0  anti-D0  K-   pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*+  anti-D0  anti-K0   pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*+  D-   K-  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*+  D-   anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*0  D-   anti-K0  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D+  anti-D*0  K-   pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D0  anti-D*0  K-   pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D+  anti-D*0  anti-K0   pi-", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D+  D*-   K-  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D+  D*-   anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D0  D*-   anti-K0  pi+", model="PHOTOS PHSP;", bf=0.000182),
    BelleBBFEntry(decay="D*+  anti-D*0  K-   pi0", model="PHOTOS PHSP;", bf=0.000384),
    BelleBBFEntry(decay="D*0  anti-D*0  K-   pi+", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*+  anti-D*0  anti-K0   pi-", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*+  D*-   K-  pi+", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D*+  D*-   anti-K0  pi0", model="PHOTOS PHSP;", bf=0.000384),
    BelleBBFEntry(decay="D*0  D*-   anti-K0  pi+", model="PHOTOS PHSP;", bf=0.000385),
    BelleBBFEntry(decay="D+  anti-D0  K*-   pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  K*-   pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  anti-D0  anti-K*0   pi-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-   K*-  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-   anti-K*0  pi0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-   anti-K*0  pi+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  anti-D0  K-   rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  K-   rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  anti-D0  anti-K0   rho-", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-   K-  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D+  D-   anti-K0  rho0", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  D-   anti-K0  rho+", model="PHOTOS PHSP;", bf=0.000400),
    BelleBBFEntry(decay="D0  anti-D0  anti-K0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D0  anti-D0  anti-K*0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D*0  anti-K*0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D*0  anti-D0  anti-K0", model="PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D0   anti-D*0  anti-K0", model="PHSP;", bf=0.001500),
    BelleBBFEntry(decay="D*0  anti-D0  anti-K*0", model="PHSP;", bf=0.000364),
    BelleBBFEntry(decay="D*0  anti-D*0  anti-K0", model="PHSP;", bf=0.002548),
    BelleBBFEntry(decay="D*0  anti-D*0  anti-K*0", model="PHSP;", bf=0.000819),
    BelleBBFEntry(decay="D_10  anti-D0  anti-K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D_10  anti-K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D_10  anti-D*0  anti-K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*0  anti-D_10  anti-K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D_2*0  anti-D0  anti-K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D0  anti-D_2*0  anti-K0", model="PHSP;", bf=0.000091),
    BelleBBFEntry(decay="D_2*0  anti-D*0  anti-K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D*0  anti-D_2*0  anti-K0", model="PHSP;", bf=0.000273),
    BelleBBFEntry(decay="D+  pi-", model="PHOTOS PHSP;", bf=0.00268),
    BelleBBFEntry(decay="D+  pi-  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="rho- D+", model="PHOTOS SVS;", bf=0.0075),
    BelleBBFEntry(decay="D+  pi+  pi-  pi-", model="PHOTOS PHSP;", bf=0.004705),
    BelleBBFEntry(decay="D+  pi0  pi-  pi0", model="PHOTOS PHSP;", bf=0.0009),
    BelleBBFEntry(decay="D+  rho0  pi-", model="PHOTOS PHSP;", bf=0.0014),
    BelleBBFEntry(decay="D+  rho-  pi0", model="PHOTOS PHSP;", bf=0.0009),
    BelleBBFEntry(decay="a_1- D+", model="PHOTOS SVS;", bf=0.0075),
    BelleBBFEntry(decay="b_1- D+", model="PHOTOS SVS;", bf=0.0000),
    BelleBBFEntry(decay="D+   pi+  pi-  pi-  pi0", model="PHOTOS PHSP;", bf=0.0000),
    BelleBBFEntry(decay="D+   omega  pi-", model="PHOTOS PHSP;", bf=0.0028),
    BelleBBFEntry(decay="D+   p+  anti-p-  pi-", model="PHOTOS PHSP;", bf=0.00034),
    BelleBBFEntry(decay="D+   anti-p-  n0", model="PHOTOS PHSP;", bf=0.00150),
    BelleBBFEntry(decay="D*+ pi-", model="PHOTOS SVS;", bf=0.00276),
    BelleBBFEntry(decay="D*+ pi-  pi0", model="PHOTOS PHSP;", bf=0.00060),
    BelleBBFEntry(decay="rho- D*+", model="PHOTOS SVV_HELAMP 0.152 1.47 0.936 0 0.317 0.19;", bf=0.00680),
    BelleBBFEntry(decay="D*+  pi+  pi-  pi-", model="PHOTOS PHSP;", bf=0.0011),
    BelleBBFEntry(decay="D*+  pi0  pi-  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D*+  rho0  pi-", model="PHOTOS PHSP;", bf=0.0072),
    BelleBBFEntry(decay="D*+  rho-  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D*+  a_1-", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0130),
    BelleBBFEntry(decay="D*+  b_1-", model="PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;", bf=0.0000),
    BelleBBFEntry(decay="D*+  pi+  pi-  pi-  pi0", model="PHOTOS PHSP;", bf=0.0176),
    BelleBBFEntry(decay="D*+  omega  pi-", model="PHOTOS PHSP;", bf=0.0029),
    BelleBBFEntry(decay="D*+   p+  anti-p-  pi-", model="PHOTOS PHSP;", bf=0.00065),
    BelleBBFEntry(decay="D*+   anti-p-  n0", model="PHOTOS PHSP;", bf=0.00150),
    BelleBBFEntry(decay="D_1+ pi-", model="PHOTOS SVS;", bf=0.00093),
    BelleBBFEntry(decay="D_2*+ pi-", model="PHOTOS STS;", bf=0.00085),
    BelleBBFEntry(decay="D_1+ rho-", model="PHOTOS PHSP;", bf=0.00093),
    BelleBBFEntry(decay="D_2*+ rho-", model="PHOTOS PHSP;", bf=0.00085),
    BelleBBFEntry(decay="D_2*0 pi0", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="D_10 pi0", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="D'_10 pi0", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="D_0*0 pi0", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="D_2*0 eta", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="D_10 eta", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="D'_10 eta", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="D_0*0 eta", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="D_2*0 eta'", model="STS;", bf=0.00003),
    BelleBBFEntry(decay="D_10 eta'", model="SVS;", bf=0.00006),
    BelleBBFEntry(decay="D'_10 eta'", model="SVS;", bf=0.00004),
    BelleBBFEntry(decay="D_0*0 eta'", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="D_2*0 omega", model="PHSP;", bf=0.00005),
    BelleBBFEntry(decay="D_10 omega", model="PHSP;", bf=0.00009),
    BelleBBFEntry(decay="D'_10 omega", model="PHSP;", bf=0.00007),
    BelleBBFEntry(decay="D_0*0 omega", model="PHSP;", bf=0.00008),
    BelleBBFEntry(decay="D_2*0 rho0", model="PHSP;", bf=0.00006),
    BelleBBFEntry(decay="D_10 rho0", model="PHSP;", bf=0.00011),
    BelleBBFEntry(decay="D'_10 rho0", model="PHSP;", bf=0.00008),
    BelleBBFEntry(decay="D_0*0 rho0", model="PHSP;", bf=0.00010),
    BelleBBFEntry(decay="D0 pi0", model="PHSP;", bf=0.00026),
    BelleBBFEntry(decay="D0  pi+  pi-", model="PHOTOS PHSP;", bf=0.00045),
    BelleBBFEntry(decay="rho0 D0", model="SVS;", bf=0.00029),
    BelleBBFEntry(decay="D0  pi0  pi0", model="PHSP;", bf=0.0005),
    BelleBBFEntry(decay="D0 eta", model="PHSP;", bf=0.00020),
    BelleBBFEntry(decay="D0 eta'", model="PHSP;", bf=0.00013),
    BelleBBFEntry(decay="omega D0", model="SVS;", bf=0.00026),
    BelleBBFEntry(decay="D0  pi+  pi-  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D0  pi0  pi0  pi0", model="PHSP;", bf=0.0001),
    BelleBBFEntry(decay="D0 p+ anti-p-", model="PHOTOS PHSP;", bf=0.000114),
    BelleBBFEntry(decay="D*0 pi0", model="SVS;", bf=0.00017),
    BelleBBFEntry(decay="D*0 pi+  pi-", model="PHOTOS PHSP;", bf=0.00032),
    BelleBBFEntry(decay="D*0 rho0", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00029),
    BelleBBFEntry(decay="D*0 pi0  pi0", model="PHSP;", bf=0.0005),
    BelleBBFEntry(decay="D*0 eta", model="SVS;", bf=0.00018),
    BelleBBFEntry(decay="D*0 eta'", model="SVS;", bf=0.00012),
    BelleBBFEntry(decay="D*0 omega", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00027),
    BelleBBFEntry(decay="D*0  pi+  pi-  pi0", model="PHOTOS PHSP;", bf=0.0010),
    BelleBBFEntry(decay="D*0  pi0  pi0  pi0", model="PHSP;", bf=0.0001),
    BelleBBFEntry(decay="D*0 p+ anti-p-", model="PHOTOS PHSP;", bf=0.000103),
    BelleBBFEntry(decay="D+  K-", model="PHOTOS PHSP;", bf=0.00020),
    BelleBBFEntry(decay="K*- D+", model="PHOTOS SVS;", bf=0.00045),
    BelleBBFEntry(decay="D*+ K-", model="PHOTOS SVS;", bf=0.00021),
    BelleBBFEntry(decay="D*+ K*-", model="PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.00033),
    BelleBBFEntry(decay="D0 anti-K0", model="PHSP;", bf=0.000052),
    BelleBBFEntry(decay="anti-K*0 D0", model="SVS;", bf=0.000042),
    BelleBBFEntry(decay="D*0 anti-K0", model="SVS;", bf=0.000036),
    BelleBBFEntry(decay="D*0 anti-K*0", model="SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;", bf=0.000042),
    BelleBBFEntry(decay="anti-u  s  c  anti-d", model="PHOTOS PYTHIA 48;", bf=0.024297),
    BelleBBFEntry(decay="anti-u  d  c  anti-d", model="PHOTOS PYTHIA 48;# ====B->DX End", bf=0.372909),
    BelleBBFEntry(decay="Lambda_c+ anti-p-", model="PHOTOS PHSP;", bf=0.000022),
    BelleBBFEntry(decay="Sigma_c+  anti-p-", model="PHOTOS PHSP;", bf=0.000022),
    BelleBBFEntry(decay="Sigma_c0  anti-n0", model="PHSP;", bf=0.000022),
    BelleBBFEntry(decay="cd_0      anti-ud_0", model="PHOTOS PYTHIA 23;", bf=0.018000),
    BelleBBFEntry(decay="cd_1      anti-ud_1", model="PHOTOS PYTHIA 23;", bf=0.018000),
    BelleBBFEntry(decay="cd_0      anti-ud_0", model="PHOTOS PYTHIA 24;", bf=0.006000),
    BelleBBFEntry(decay="cd_1      anti-ud_1", model="PHOTOS PYTHIA 24;", bf=0.005932),
    BelleBBFEntry(decay="Xi_c+     anti-Lambda_c-", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="Xi_c+     anti-Sigma_c-", model="PHOTOS PHSP;", bf=0.001000),
    BelleBBFEntry(decay="Xi_c0     anti-Sigma_c0", model="PHSP;", bf=0.001000),
    BelleBBFEntry(decay="cs_0       anti-cd_0", model="PHOTOS PYTHIA 23;", bf=0.009000),
    BelleBBFEntry(decay="Xi_c+     anti-p-", model="PHOTOS PHSP;", bf=0.000001),
    BelleBBFEntry(decay="Xi_c0     anti-n0", model="PHSP;", bf=0.000001),
    BelleBBFEntry(decay="cs_0       anti-ud_0", model="PHOTOS PYTHIA 23;", bf=0.000900),
    BelleBBFEntry(decay="cs_1       anti-ud_1", model="PHOTOS PYTHIA 23;", bf=0.000900),
    BelleBBFEntry(decay="cs_0       anti-ud_0", model="PHOTOS PYTHIA 24;", bf=0.000300),
    BelleBBFEntry(decay="cs_1       anti-ud_1", model="PHOTOS PYTHIA 24;", bf=0.000300),
    BelleBBFEntry(decay="Lambda_c+    anti-Lambda_c-", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Sigma_c+     anti-Lambda_c-", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Lambda_c+    anti-Sigma_c-", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Sigma_c+     anti-Sigma_c-", model="PHOTOS PHSP;", bf=0.000040),
    BelleBBFEntry(decay="Sigma_c0     anti-Sigma_c0", model="PHSP;", bf=0.000040),
    BelleBBFEntry(decay="cd_0       anti-cd_0", model="PHOTOS PYTHIA 23;# ====B->BaryonX End", bf=0.000400),
)  # type: Tuple[BelleBBFEntry, ...]
# fmt: on

sum_belle_decay_b_zero = sum([bf_e.bf for bf_e in belle_decay_infos_b_zero])  # type: float
sum_belle_decay_anti_b_zero = sum([bf_e.bf for bf_e in belle_decay_infos_anti_b_zero])  # type: float

assert sum_belle_decay_anti_b_zero == sum_belle_decay_b_zero, (sum_belle_decay_anti_b_zero, sum_belle_decay_b_zero)
assert np.isclose(a=sum_belle_decay_b_zero, b=1.0, rtol=1e-5, atol=1e-12), sum_belle_decay_b_zero

# print(f"Sum Decay B0: {sum_belle_decay_b_zero}")
# print(f"Sum Decay anti-B0: {sum_belle_decay_anti_b_zero}")

__belle_bf_b0_to_d_l_nu_keys = (
    "D- e+ nu_e",
    "D- mu+ nu_mu",
    "D+ e- anti-nu_e",
    "D+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_d_st_l_nu_keys = (
    "D*- e+ nu_e",
    "D*- mu+ nu_mu",
    "D*+ e- anti-nu_e",
    "D*+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_d_tau_nu_keys = ("D- tau+ nu_tau", "D+ tau- anti-nu_tau")  # type: Tuple[str, ...]
__belle_bf_b0_to_d_st_tau_nu_keys = ("D*- tau+ nu_tau", "D*+ tau- anti-nu_tau")  # type: Tuple[str, ...]

__comb_list = tuple(
    list(belle_decay_infos_b_zero) + list(belle_decay_infos_anti_b_zero)
)  # type: Tuple[BelleBBFEntry, ...]

__belle_bf_b0_to_d_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_d_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_st_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_d_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_tau_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_d_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_st_tau_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_b0_to_d_l_nu_set) == 1, list(__belle_bf_b0_to_d_l_nu_set)
assert len(__belle_bf_b0_to_d_st_l_nu_set) == 1, list(__belle_bf_b0_to_d_st_l_nu_set)
assert len(__belle_bf_b0_to_d_tau_nu_set) == 1, list(__belle_bf_b0_to_d_tau_nu_set)
assert len(__belle_bf_b0_to_d_st_tau_nu_set) == 1, list(__belle_bf_b0_to_d_st_tau_nu_set)


belle_bf_b0_to_d_l_nu = list(__belle_bf_b0_to_d_l_nu_set)[0]  # type: float
belle_bf_b0_to_d_st_l_nu = list(__belle_bf_b0_to_d_st_l_nu_set)[0]  # type: float
belle_bf_b0_to_d_tau_nu = list(__belle_bf_b0_to_d_tau_nu_set)[0]  # type: float
belle_bf_b0_to_d_st_tau_nu = list(__belle_bf_b0_to_d_st_tau_nu_set)[0]  # type: float


__belle_bf_b0_to_dzero_st_l_nu_keys = (
    "D_0*- e+ nu_e",
    "D_0*- mu+ nu_mu",
    "D_0*+ e- anti-nu_e",
    "D_0*+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_done_l_nu_keys = (
    "D_1- e+ nu_e",
    "D_1- mu+ nu_mu",
    "D_1+ e- anti-nu_e",
    "D_1+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_done_prime_l_nu_keys = (
    "D'_1- e+ nu_e",
    "D'_1- mu+ nu_mu",
    "D'_1+ e- anti-nu_e",
    "D'_1+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_dtwo_st_l_nu_keys = (
    "D_2*- e+ nu_e",
    "D_2*- mu+ nu_mu",
    "D_2*+ e- anti-nu_e",
    "D_2*+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_b0_to_dzero_st_tau_nu_keys = (
    "D_0*- tau+ nu_tau",
    "D_0*+ tau- anti-nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_done_tau_nu_keys = (
    "D_1- tau+ nu_tau",
    "D_1+ tau- anti-nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_done_prime_tau_nu_keys = (
    "D'_1- tau+ nu_tau",
    "D'_1+ tau- anti-nu_tau",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_dtwo_st_tau_nu_keys = (
    "D_2*- tau+ nu_tau",
    "D_2*+ tau- anti-nu_tau",
)  # type: Tuple[str, ...]


__belle_bf_b0_to_dzero_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dzero_st_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_done_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_done_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_done_prime_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_done_prime_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_dtwo_st_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dtwo_st_l_nu_keys]
)  # type: Set[float]

__belle_bf_b0_to_dzero_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dzero_st_tau_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_done_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_done_tau_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_done_prime_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_done_prime_tau_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_dtwo_st_tau_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dtwo_st_tau_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_b0_to_dzero_st_l_nu_set) == 1, list(__belle_bf_b0_to_dzero_st_l_nu_set)
assert len(__belle_bf_b0_to_done_l_nu_set) == 1, list(__belle_bf_b0_to_done_l_nu_set)
assert len(__belle_bf_b0_to_done_prime_l_nu_set) == 1, list(__belle_bf_b0_to_done_prime_l_nu_set)
assert len(__belle_bf_b0_to_dtwo_st_l_nu_set) == 1, list(__belle_bf_b0_to_dtwo_st_l_nu_set)

assert len(__belle_bf_b0_to_dzero_st_tau_nu_set) == 1, list(__belle_bf_b0_to_dzero_st_tau_nu_set)
assert len(__belle_bf_b0_to_done_tau_nu_set) == 1, list(__belle_bf_b0_to_done_tau_nu_set)
assert len(__belle_bf_b0_to_done_prime_tau_nu_set) == 1, list(__belle_bf_b0_to_done_prime_tau_nu_set)
assert len(__belle_bf_b0_to_dtwo_st_tau_nu_set) == 1, list(__belle_bf_b0_to_dtwo_st_tau_nu_set)

belle_bf_b0_to_dzero_st_l_nu = list(__belle_bf_b0_to_dzero_st_l_nu_set)[0]  # type: float
belle_bf_b0_to_done_l_nu = list(__belle_bf_b0_to_done_l_nu_set)[0]  # type: float
belle_bf_b0_to_done_prime_l_nu = list(__belle_bf_b0_to_done_prime_l_nu_set)[0]  # type: float
belle_bf_b0_to_dtwo_st_l_nu = list(__belle_bf_b0_to_dtwo_st_l_nu_set)[0]  # type: float

belle_bf_b0_to_dzero_st_tau_nu = list(__belle_bf_b0_to_dzero_st_tau_nu_set)[0]  # type: float
belle_bf_b0_to_done_tau_nu = list(__belle_bf_b0_to_done_tau_nu_set)[0]  # type: float
belle_bf_b0_to_done_prime_tau_nu = list(__belle_bf_b0_to_done_prime_tau_nu_set)[0]  # type: float
belle_bf_b0_to_dtwo_st_tau_nu = list(__belle_bf_b0_to_dtwo_st_tau_nu_set)[0]  # type: float

sum_of_belle_b0_to_dss_l_nu_bfs = sum(
    [
        2 * belle_bf_b0_to_dzero_st_l_nu,
        2 * belle_bf_b0_to_done_l_nu,
        2 * belle_bf_b0_to_done_prime_l_nu,
        2 * belle_bf_b0_to_dtwo_st_l_nu,
        belle_bf_b0_to_dzero_st_tau_nu,
        belle_bf_b0_to_done_tau_nu,
        belle_bf_b0_to_done_prime_tau_nu,
        belle_bf_b0_to_dtwo_st_tau_nu,
    ]
)  # type: float


belle_decay_infos_done_mixed = (
    BelleBBFEntry(decay="D*0 pi0", model="VVS_PWAVE  0.0 0.0 0.0 0.0 1.0 0.0;", bf=0.1997),
    BelleBBFEntry(decay="D*+ pi-", model="PHOTOS VVS_PWAVE  0.0 0.0 0.0 0.0 1.0 0.0;", bf=0.3994),
    BelleBBFEntry(decay="D0 pi+ pi-", model="PHOTOS PHSP;", bf=0.1719),
    BelleBBFEntry(decay="D0 pi0 pi0", model="PHSP;", bf=0.1145),
    BelleBBFEntry(decay="D+ pi- pi0", model="PHOTOS PHSP;", bf=0.1145),
)  # type: Tuple[BelleBBFEntry, ...]

belle_mixed_done_regular_entries = [
    e for e in belle_decay_infos_done_mixed if e.decay.count(" pi") == 1
]  # type: List[BelleBBFEntry]
belle_mixed_done_gap_entries = [
    e for e in belle_decay_infos_done_mixed if e.decay.count(" pi") == 2
]  # type: List[BelleBBFEntry]
belle_mixed_done_regular_fraction = sum([e.bf for e in belle_mixed_done_regular_entries])  # type: float
belle_mixed_done_gap_fraction = sum([e.bf for e in belle_mixed_done_gap_entries])  # type: float

belle_bf_b0_to_done_to_dst_pi_l_nu = belle_bf_b0_to_done_l_nu * belle_mixed_done_regular_fraction  # type: float
belle_bf_b0_to_done_to_d_pi_pi_l_nu = belle_bf_b0_to_done_l_nu * belle_mixed_done_gap_fraction  # type: float
belle_bf_b0_to_done_to_dst_pi_tau_nu = belle_bf_b0_to_done_tau_nu * belle_mixed_done_regular_fraction  # type: float
belle_bf_b0_to_done_to_d_pi_pi_tau_nu = belle_bf_b0_to_done_tau_nu * belle_mixed_done_gap_fraction  # type: float


__belle_bf_b0_to_d_2s_l_nu_keys = (
    "D(2S)- e+ nu_e",
    "D(2S)- mu+ nu_mu",
    "D(2S)+ e- anti-nu_e",
    "D(2S)+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_dst_2s_l_nu_keys = (
    "D*(2S)- e+ nu_e",
    "D*(2S)- mu+ nu_mu",
    "D*(2S)+ e- anti-nu_e",
    "D*(2S)+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_b0_to_d_2s_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_2s_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_dst_2s_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dst_2s_l_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_b0_to_d_2s_l_nu_set) == 1, list(__belle_bf_b0_to_d_2s_l_nu_set)
assert len(__belle_bf_b0_to_dst_2s_l_nu_set) == 1, list(__belle_bf_b0_to_dst_2s_l_nu_set)

belle_bf_b0_to_d_2s_l_nu = list(__belle_bf_b0_to_d_2s_l_nu_set)[0]  # type: float
belle_bf_b0_to_dst_2s_l_nu = list(__belle_bf_b0_to_dst_2s_l_nu_set)[0]  # type: float


__belle_bf_b0_to_dst_pi_zero_l_nu_keys = (
    "D*- pi0 e+ nu_e",
    "D*- pi0 mu+ nu_mu",
    "D*+ pi0 e- anti-nu_e",
    "D*+ pi0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_dst_pi_minus_l_nu_keys = (
    "anti-D*0 pi- e+ nu_e",
    "anti-D*0 pi- mu+ nu_mu",
    "D*0 pi+ e- anti-nu_e",
    "D*0 pi+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_d_pi_zero_l_nu_keys = (
    "D- pi0 e+ nu_e",
    "D- pi0 mu+ nu_mu",
    "D+ pi0 e- anti-nu_e",
    "D+ pi0 mu- anti-nu_mu",
)  # type: Tuple[str, ...]
__belle_bf_b0_to_d_pi_minus_l_nu_keys = (
    "anti-D0 pi- e+ nu_e",
    "anti-D0 pi- mu+ nu_mu",
    "D0 pi+ e- anti-nu_e",
    "D0 pi+ mu- anti-nu_mu",
)  # type: Tuple[str, ...]

__belle_bf_b0_to_dst_pi_zero_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dst_pi_zero_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_dst_pi_minus_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_dst_pi_minus_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_d_pi_zero_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_pi_zero_l_nu_keys]
)  # type: Set[float]
__belle_bf_b0_to_d_pi_minus_l_nu_set = set(
    [e.bf for e in __comb_list if e.clean_decay in __belle_bf_b0_to_d_pi_minus_l_nu_keys]
)  # type: Set[float]

assert len(__belle_bf_b0_to_dst_pi_zero_l_nu_set) == 1, list(__belle_bf_b0_to_dst_pi_zero_l_nu_set)
assert len(__belle_bf_b0_to_dst_pi_minus_l_nu_set) == 1, list(__belle_bf_b0_to_dst_pi_minus_l_nu_set)
assert len(__belle_bf_b0_to_d_pi_zero_l_nu_set) == 1, list(__belle_bf_b0_to_d_pi_zero_l_nu_set)
assert len(__belle_bf_b0_to_d_pi_minus_l_nu_set) == 1, list(__belle_bf_b0_to_d_pi_minus_l_nu_set)

belle_bf_b0_to_dst_pi_zero_l_nu = list(__belle_bf_b0_to_dst_pi_zero_l_nu_set)[0]  # type: float
belle_bf_b0_to_dst_pi_minus_l_nu = list(__belle_bf_b0_to_dst_pi_minus_l_nu_set)[0]  # type: float
belle_bf_b0_to_d_pi_zero_l_nu = list(__belle_bf_b0_to_d_pi_zero_l_nu_set)[0]  # type: float
belle_bf_b0_to_d_pi_minus_l_nu = list(__belle_bf_b0_to_d_pi_minus_l_nu_set)[0]  # type: float


sum_of_belle_removed_b0_bfs = sum(
    [
        belle_bf_b0_to_d_2s_l_nu,
        belle_bf_b0_to_dst_2s_l_nu,
        belle_bf_b0_to_dst_pi_zero_l_nu,
        belle_bf_b0_to_dst_pi_minus_l_nu,
        belle_bf_b0_to_d_pi_zero_l_nu,
        belle_bf_b0_to_d_pi_minus_l_nu,
    ]
)  # type: float


original_b_zero_decay_lines = [
    "Decay B0",
    "# B->DXlnu",
    "#     anti-B0 -> X_c l nu, sum = 10.45% (using: tau_B+/tau_B0 = 1.09)",
    "#     2005/04/02 H.Kakuno, Update using the following numbers:",
    "#       B->Xlnu     Br=10.9%               (HFAG summer 2004)",
    "#       B0->Dlnu    Br=2.14%,  rho_D=1.15  (HFAG summer 2004)",
    "#       B0->D*lnu   Br=5.33%               (HFAG summer 2004),",
    "#                    rho_A1=1.3            (hepex/0408027),",
    "#                    R1=1.18,  R2=0.71     (PRL76,3893)",
    "#       B->D**lnu   Br=3.04%          (CERN-EP/2001-050)",
    "#       B->D_1 lnu  Br=0.74%          (PDG 2004)",
    "#       B->D_2* lnu Br<0.65%          (PDG 2004)",
    "#       B->D pi lnu Br=1.54%          (PLB475, 407)",
    "#       B->D*pi lnu Br=1.86%          (PLB475, 407)",
    "#       referred also D0 preliminary result from CKM2005 for D**lnu",
    "0.0533   D*-           e+   nu_e    PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0213   D-            e+   nu_e    PHOTOS  HQET2 1.16;",
    "0.0074   D_1-          e+   nu_e    PHOTOS  ISGW2;",
    "0.0084   D_0*-         e+   nu_e    PHOTOS  ISGW2;",
    "0.0074   D'_1-         e+   nu_e    PHOTOS  ISGW2;",
    "0.0036   D_2*-         e+   nu_e    PHOTOS  ISGW2;",
    "0.0002   D(2S)-        e+   nu_e    PHOTOS  ISGW2;",
    "0.0005   D*(2S)-       e+   nu_e    PHOTOS  ISGW2;",
    "0.0003   D*-      pi0  e+   nu_e    PHOTOS  GOITY_ROBERTS;",
    "0.0006   anti-D*0 pi-  e+   nu_e    PHOTOS  GOITY_ROBERTS;",
    "0.0005   D-       pi0  e+   nu_e    PHOTOS  GOITY_ROBERTS;",
    "0.0010   anti-D0  pi-  e+   nu_e    PHOTOS  GOITY_ROBERTS;",
    "0.0533   D*-           mu+  nu_mu   PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0213   D-            mu+  nu_mu   PHOTOS  HQET2 1.16;",
    "0.0074   D_1-          mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0084   D_0*-         mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0074   D'_1-         mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0036   D_2*-         mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0002   D(2S)-        mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0005   D*(2S)-       mu+  nu_mu   PHOTOS  ISGW2;",
    "0.0003   D*-      pi0  mu+  nu_mu   PHOTOS  GOITY_ROBERTS;",
    "0.0006   anti-D*0 pi-  mu+  nu_mu   PHOTOS  GOITY_ROBERTS;",
    "0.0005   D-       pi0  mu+  nu_mu   PHOTOS  GOITY_ROBERTS;",
    "0.0010   anti-D0  pi-  mu+  nu_mu   PHOTOS  GOITY_ROBERTS;",
    "# B->DXtaunu",
    "#                    B0 -> X_c tau nu, sum = 2.4% (using: tau_B+/tau_B0 = 1.09)",
    "#                    H.Kakuno 2005/04/02",
    "0.0142   D*-           tau+ nu_tau          PHOTOS ISGW2;",
    "0.0071   D-            tau+ nu_tau          PHOTOS ISGW2;",
    "0.0012   D_1-          tau+ nu_tau          PHOTOS ISGW2;",
    "0.0015   D_0*-         tau+ nu_tau          PHOTOS ISGW2;",
    "0.0012   D'_1-         tau+ nu_tau          PHOTOS ISGW2;",
    "0.0008   D_2*-         tau+ nu_tau          PHOTOS ISGW2;",
    "# B->ccX",
    "# B -> charmonium updates:",
    "#      Mar. 2001:  Lange",
    "#      Nov. 2002:  Schrenk",
    "#      Oct. 2003:  H. Guler",
    "#      Mar. 2005:  S.E.Lee, Takada and Ishikawa",
    "#      Apr. 2006:  K.Miyabayashi",
    "#       BF[%]           B0      B+",
    "#       eta_c           0.5034  0.5455",
    "#       eta_c(2S)       0.2015  0.2190",
    "#       J/psi           0.7539  0.8118",
    "#       psi(2S)         0.2967  0.3152",
    "#       chi_c0          0.2127  0.2291",
    "#       chi_c1          0.3228  0.3476",
    "#       chi_c2          0.1595  0.1718",
    "#       h_c             0.1345  0.1449",
    "#       X(3872)         0.0080  0.0080",
    "#       Y(3940)         0.0070  0.0071",
    "# -------------------------------------",
    "#       sum             2.6000  2.8000",
    "# B0 --> charmonium X,  2.6000%",
    "# B -> cc= s",
    "# eta_c inclusive",
    "# PDG 2005: < 0.009",
    "# Guess 0.005034",
    "# B0 --> eta_c K0",
    "# PDG 2005: 0.00116 +/- 0.00026",
    "0.000580   eta_c K_S0    SSS_CP beta dm -1 1 0 1 0.0;",
    "0.000580   eta_c K_L0    SSS_CP beta dm  1 1 0 1 0.0;",
    "# B0 --> eta_c K*0(892)",
    "# PDG 2005: 0.0016 +/- 0.0007",
    "0.000266   K*S eta_c          SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000266   K*L eta_c          SVS_CP beta dm  1 1 0 1 0.0;",
    "0.001068   K*0T eta_c    SVS;",
    "# Fill up inclusive rate with low momentum (0.00108)",
    "0.000100   eta_c K+ pi-          PHOTOS PHSP;",
    "0.000051   eta_c K0 pi0      PHSP;",
    "0.000308   eta_c K0 pi+ pi- PHOTOS PHSP;",
    "0.000308   eta_c K0 pi0 pi0  PHSP;",
    "0.000308   eta_c K+ pi- pi0      PHOTOS PHSP;",
    "# baryonic decays",
    "0.000003   eta_c anti-Lambda0 n0  PHSP;",
    "0.000003   eta_c anti-Sigma- p+ PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.001193  eta_c Xsd  PHSP;",
    "# Sum of eta_c: 0.005034",
    "# eta_c(2S) inclusive",
    "# Guess 0.002015  (40% of eta_c)",
    "# anti-B0 --> eta_c(2S) anti-K0",
    "# Guess 0.00048",
    "0.000240   eta_c(2S) K_S0     SSS_CP beta dm -1 1 0 1 0.0;",
    "0.000240   eta_c(2S) K_L0     SSS_CP beta dm  1 1 0 1 0.0;",
    "# anti-B0 --> eta_c(2S) anti-K0*",
    "# Guess 0.00064",
    "0.000106   K*S eta_c(2S)         SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000106   K*L eta_c(2S)         SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000428   K*0T eta_c(2S)   SVS;",
    "# Fill up inclusive rate with low momentum (0.00048)",
    "0.000000   eta_c(2S) K+ pi-          PHOTOS PHSP;",
    "0.000000   eta_c(2S) K0 pi0      PHSP;",
    "0.000140   eta_c(2S) K0 pi+ pi- PHOTOS PHSP;",
    "0.000140   eta_c(2S) K0 pi0 pi0  PHSP;",
    "0.000140   eta_c(2S) K+ pi- pi0      PHOTOS PHSP;",
    "# Fill up by string fragmentation",
    "0.000475  eta_c(2S) Xsd  PHSP;",
    "# J/psi inclusive (direct)",
    "# PDG 2003: 0.007296 +/- 0.0004",
    "# PDG 2005: 0.007539(life B0=1.532ps, B+=1.638ps)",
    "# anti-B0 --> J/psi K0",
    "# PDG 2005: 0.00085 +/- 0.00005",
    "0.000425   J/psi K_S0         SVS_CP beta dm -1 1 0 1 0;",
    "0.000425   J/psi K_L0         SVS_CP beta dm  1 1 0 1 0;",
    "# B0 --> J/psi K+ pi-   PDG 2005: 0.0012  +/- 0.0006",
    "# B0 --> J/psi K*0      PDG 2005: 0.00131 +/- 0.00007",
    "0.000218   J/psi K*S          SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000218   J/psi K*L          SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000874   J/psi K*0T    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000000   J/psi K+ pi-      PHOTOS PHSP; #?????",
    "# anti-B0 --> J/psi K1(1270)",
    "# PDG 2005: 0.0013 +/- 0.0005",
    "0.001300   J/psi K_10    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# anti-B0 --> J/psi K1(1400)",
    "0.000000   J/psi K'_10   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# This is not part of K1(1270)?",
    "# anti-B0 --> J/psi K0 pi+ pi-   PDG 2005: 0.0010  +/- 0.0004  (including rho0, below)",
    "# anti-B0 --> J/psi K0 rho0      PDG 2005: 0.00054 +/- 0.00030",
    "0.000540   J/psi K0 rho0       PHSP;",
    "0.000460   J/psi K0 pi+ pi-  PHOTOS  PHSP;",
    "# anti-B0 --> J/psi K*(892)- pi+",
    "# PDG 2005: 0.0008   +/- 0.0004",
    "0.000800   J/psi K*+ pi-          PHOTOS  PHSP;",
    "# anti-B0 --> J/psi anti-K*(892)0 pi+ pi-",
    "# PDG 2005: 0.00066  +/- 0.00022",
    "0.000660   J/psi K*0 pi+ pi- PHOTOS  PHSP;",
    "# anti-B0 --> J/psi phi anti-K0",
    "0.000094   J/psi phi K0  PHSP;",
    "# anti-B0 --> J/psi pi0",
    "# PDG 2005: 0.000022 +/- 0.000004",
    "0.000022   J/psi pi0          SVS_CP beta dm  1 1 0 1 0;",
    "# anti-B0 --> J/psi eta",
    "# PDG 2005: < 0.000027",
    "# M.C.Chang preliminary value (9.2 +- 1.5)*10^-6 -> 9*10^-6.",
    "0.000009   J/psi eta          SVS_CP beta dm  1 1 0 1 0;",
    "# anti-B0 --> J/psi pi+ pi-",
    "# PDG 2005: 0.000046 +/- 0.000009 (including rho0, below)",
    "0.000030   J/psi pi+ pi-     PHOTOS PHSP;",
    "# anti-B0 --> J/psi rho0",
    "# PDG 2005: 0.000016 +/- 0.000007",
    "0.000016   J/psi rho0         SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "# anti-B0 --> J/psi omega",
    "# PDG 2005: < 0.00027",
    "0.000000   J/psi omega        SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "# anti-B0 --> J/psi K0*(1430) Belle (Nitesh) 0.000050",
    "#                    K0*(1430) --> Kpi 0.93 +/- 0.10",
    "0.000050   J/psi K_0*0    SVS;",
    "# anti-B0 --> J/psi K2*(1430) Belle (Nitesh): 0.000260",
    "#                   K2*(1430) --> Kpi 0.499 +/- 0.012",
    "# STV not implemented",
    "0.000260   J/psi K_2*0    PHSP;",
    "# anti-B0 --> J/psi K*(1680) Belle (Nitesh):  ?",
    "#                   K*(1680) -->  Kpi 0.387 +/- 0.025",
    "0.000005   J/psi K''*0    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Fill up J/psi inclusive rate",
    "# kpipi replaced by K1,",
    "# little evidence for Kpi other than K*, K*(1430)",
    "# -> such multi-body decays is treated by Xsd? K.Miyabayashi",
    "#0.000000   J/psi anti-K0 pi0      PHSP;",
    "#0.000514   J/psi anti-K0 pi0 pi0  PHSP;",
    "#0.000514   J/psi K- pi+ pi0      PHOTOS PHSP;",
    "# baryonic decays",
    "# Assume that B+ -> J/Psi Lambda pbar; PDG2005: 1.2 +0.9-0.6 *10^-5.",
    "0.000012   J/psi anti-Lambda0 n0  PHSP;",
    "0.000012   J/psi anti-Sigma- p+ PHOTOS PHSP;",
    "0.001109   J/psi Xsd PHSP;",
    "# Sum of J/psi: 0.007539  20060407",
    "# Psi(2S) inclusive",
    "# PDG 2005: 0.002967",
    "# B0 --> psi(2S) K0",
    "# PDG 2003: 0.00062 +/- 0.00007",
    "0.000310   psi(2S) K_S0    SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000310   psi(2S) K_L0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# B0 --> psi(2S) K+ pi-",
    "# PDG 2005: < 0.001 (including K*0(892) below)",
    "# anti-B0 --> psi(2S) K*0(892)",
    "# PDG 2005: 0.00080 +/- 0.00013",
    "# Belle-Conf 0349 (LP03): 0.000720 +/- 0.000043 +/- 0.000065",
    "0.000120   psi(2S) K*S    SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000120   psi(2S) K*L    SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000480   psi(2S) K*0T    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# anti-B0 --> psi(2S) pi0",
    "# PDG 2005: No record, N.Kanagawa's master thesis(2004)",
    "0.000020   psi(2S) pi0          SVS_CP beta dm  1 1 0 1 0;",
    "# Fill up inclusive rate with low momentum",
    "#0.000000   psi(2S) K- pi+          PHOTOS PHSP;",
    "#0.000000   psi(2S) anti-K0 pi0      PHSP;",
    "#0.000536   psi(2S) anti-K0 pi+ pi- PHOTOS PHSP; # set to fill up PDG2005",
    "#0.000536   psi(2S) anti-K0 pi0 pi0  PHSP;       #",
    "#0.000535   psi(2S) K- pi+ pi0      PHOTOS PHSP; #",
    "# Use Xsd.",
    "0.001607   psi(2S) Xsd PHSP;",
    "# Sum of Psi(2S): 0.002967",
    "# chi_c0 inclusive",
    "# Guess 0.001750+0.000427",
    "# PDG 2003: < 0.0005",
    "# Guess 0.00056",
    "# Guess 0.00050 after adjusted value.(2004)",
    "# cf), B+ -> chi_c0 K+ is (112+-12+-18)*10^-6 (hep-ex/0509001)",
    "0.000250   chi_c0 K_S0   SSS_CP beta dm  1 1 0 1 0.0;",
    "0.000250   chi_c0 K_L0   SSS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.00066",
    "0.000110   K*S chi_c0          SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000110   K*L chi_c0          SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000440   K*0T chi_c0    SVS;",
    "# Fill up inclusive rate with low momentum",
    "#0.000092   chi_c0 K+ pi-          PHOTOS  PHSP;",
    "#0.000092   chi_c0 K0 pi0       PHSP;",
    "#0.000261   chi_c0 K0 pi+ pi- PHOTOS  PHSP;",
    "#0.000261   chi_c0 K0 pi0 pi0   PHSP;",
    "#0.000261   chi_c0 K+ pi- pi0      PHOTOS  PHSP;",
    "#Switch to string fragmentation.",
    "0.000967   chi_c0  Xsd  PHSP;",
    "# Sum of chi_c0: 0.002127",
    "# chi_c1 inclusive (direct)",
    "# PDG 2005: 0.003228",
    "# PDG 2003: 0.00040 + 0.00012 - 0.00010",
    "0.000200   chi_c1 K_S0    SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000200   chi_c1 K_L0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# PDG 2003: 0.00041 +/- 0.00015",
    "# PDG 2005: same as PDG 2003.",
    "0.000068   chi_c1 K*S   SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000068   chi_c1 K*L   SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000274   chi_c1 K*0T  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# chi_c1 pi0 (1.8+-0.7+-0.2)*10^-5 M.Fujita's master thesis(2005)",
    "0.000018   chi_c1 pi0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   chi_c1 K+ pi-           PHOTOS PHSP;",
    "0.000000   chi_c1 K0 pi0      PHSP;",
    "0.000500   chi_c1 K0 pi+ pi-  PHOTOS PHSP;",
    "0.000500   chi_c1 K0 pi0 pi0  PHSP;",
    "0.000500   chi_c1 K+ pi- pi0       PHOTOS PHSP;",
    "# String frafmentation.",
    "0.000900  chi_c1  Xsd  PHSP;",
    "# Sum of chi_c1: 0.003228",
    "# chi_c2 inclusive (direct)",
    "# PDG 2005: 0.001595",
    "# No real evidence for these decays",
    "# Guess 0.000076",
    "0.000038   chi_c2 K_S0             STS_CP beta dm  1 1 0 1 0.0;",
    "0.000038   chi_c2 K_L0             STS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.000114",
    "# STV not implemented",
    "0.000019   chi_c2 K*S              PHSP;",
    "0.000019   chi_c2 K*L              PHSP;",
    "0.000076   chi_c2 K*0T        PHSP;",
    "# Guess 0.000805",
    "0.000000   chi_c2 K+ pi-          PHOTOS PHSP;",
    "0.000000   chi_c2 K0 pi0      PHSP;",
    "0.000268   chi_c2 K0 pi+ pi- PHOTOS PHSP;",
    "0.000268   chi_c2 K0 pi0 pi0  PHSP;",
    "0.000269   chi_c2 K+ pi- pi0      PHOTOS PHSP;",
    "# String fragmentation.",
    "0.000600  chi_c2  Xsd  PHSP;",
    "# Sum of chi_c2: 0.001595",
    "# h_c Inclusive",
    "# Guess 0.001345",
    "# Guess 0.000180",
    "0.000090   h_c K_S0             SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000090   h_c K_L0             SVS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.000205",
    "# STV not implemented",
    "0.000034   h_c K*S              PHSP;",
    "0.000034   h_c K*L              PHSP;",
    "0.000137   h_c K*0T        PHSP;",
    "# Guess 0.00096",
    "#0.000000   h_c K+ pi-          PHOTOS PHSP;",
    "#0.000000   h_c K0 pi0      PHSP;",
    "#0.000320   h_c K0 pi+ pi- PHOTOS PHSP;",
    "#0.000320   h_c K0 pi0 pi0  PHSP;",
    "#0.000320   h_c K+ pi- pi0      PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.000960    h_c  Xsd  PHSP;",
    "# Sum of h_c: 0.001345",
    "# B0 --> K0 X(3872)",
    "# Register only B -> X(3872) K.",
    "0.000040   X(3872) K_S0  SVS_CP beta dm  -1 1 0 1 0.0;",
    "0.000040   X(3872) K_L0  SVS_CP beta dm  1 1 0 1 0.0;",
    "# omit B -> X(3872) K*, No signal according to Karim's work(BAM200509))",
    "#0.000120   X(3872) anti-K*0  SVV_HELAMP PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# sum of X(3872) = 0.000080",
    "# B0 --> K0 Y(3940) assume that same as charged B.",
    "0.000035   Y(3940) K_S0  SVS_CP beta dm  -1 1 0 1 0.0;",
    "0.000035   Y(3940) K_L0  SVS_CP beta dm  1 1 0 1 0.0;",
    "# B->DsX",
    "# b -> c (sc=)  -->  D Ds X   Sum =  8.3%",
    "#   BF(Ds->phipi)=4.38(Babar2005+PDG) is assumed.",
    "#   Updated on 29/01/2008",
    "#   based on PDG update (Ref: W.-M. Yao et al., J. Phys. G 33, 1 (2006) and 2007 partial update for the 2008)",
    "#   by Nikhil J Joshi (njoshi@bpost)",
    "## Upper Vertex Ds",
    "0.00650   D-   D_s+                     PHOTOS PHSP;",
    "0.00800   D*-  D_s+                     PHOTOS SVS;",
    "0.00740   D_s*+ D-                      PHOTOS SVS;",
    "0.01770   D_s*+ D*-                     PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00033   D_s+ D_0*-                    PHOTOS PHSP;",
    "0.00067   D_s*+ D_0*-                   PHOTOS SVS;",
    "0.00031   D'_1- D_s+                    PHOTOS SVS;",
    "0.00063   D'_1-  D_s*+                  PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00063   D_1-  D_s+                    PHOTOS SVS;",
    "0.00126   D_1-  D_s*+                   PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00165   D_2*-  D_s+                   PHOTOS STS;",
    "0.00329   D_2*-  D_s*+                  PHOTOS PHSP;",
    "# DsJ(2317)=D_s0*-  and  DsJ(2457)=D'_s1-",
    "0.00097   D- D_s0*+                     PHOTOS PHSP;",
    "0.00150   D*- D_s0*+                    PHOTOS SVS;",
    "0.00350   D'_s1+ D-                     PHOTOS SVS;",
    "0.00930   D'_s1+ D*-                    PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00091   D_s+  D-  pi0                 PHOTOS PHSP;",
    "0.00182   D_s+  anti-D0 pi-             PHOTOS PHSP;",
    "0.00091   D_s*+ D-  pi0                 PHOTOS PHSP;",
    "0.00182   D_s*+ anti-D0 pi-             PHOTOS PHSP;",
    "0.00151   D_s+  D-  pi-  pi+            PHOTOS PHSP;",
    "0.00070   D_s+  D-  pi0  pi0            PHOTOS PHSP;",
    "0.00151   D_s+  anti-D0 pi-  pi0        PHOTOS PHSP;",
    "0.00151   D_s*+ D-  pi-  pi+            PHOTOS PHSP;",
    "0.00070   D_s*+ D-  pi0  pi0            PHOTOS PHSP;",
    "0.00151   D_s*+ anti-D0 pi-  pi0        PHOTOS PHSP;",
    "## End of upper vertex",
    "## Lower vertex Ds",
    "0.00095   D_s-  pi+  K0                 PHOTOS PHSP;",
    "0.00095   D_s*- pi+  K0                 PHOTOS PHSP;",
    "0.00100   D_s-  pi+  K0  pi0            PHOTOS PHSP;",
    "0.00100   D_s*- pi+  K0  pi0            PHOTOS PHSP;",
    "0.00080   D_s-  pi+  K+  pi-            PHOTOS PHSP;",
    "0.00080   D_s*- pi+  K+  pi-            PHOTOS PHSP;",
    "0.00016   D_s-  pi+  K0  pi0 pi0        PHOTOS PHSP;",
    "0.00016   D_s*- pi+  K0  pi0 pi0        PHOTOS PHSP;",
    "0.00016   D_s-  pi+  K0  pi+ pi-        PHOTOS PHSP;",
    "0.00016   D_s*- pi+  K0  pi+ pi-        PHOTOS PHSP;",
    "0.00016   D_s-  pi+  K+  pi- pi0        PHOTOS PHSP;",
    "0.00016   D_s*- pi+  K+  pi- pi0        PHOTOS PHSP;",
    "# B0->DDX",
    "# updated by H.Miyake 20021006",
    "# updated by A.Ishikawa 20021011",
    "# updated by H.Miyake 20031019",
    "# updated by J.Brodzicka 20080130 basing on PDG07",
    "# b=-> c= (d=c)",
    "# JB PDG07",
    "0.00023      D-   D+               PHOTOS SSS_CP beta_cdcd  dm  1 1.0 0.0 1.0 0.0;",
    "# JB PDG07",
    "0.00061      D*+  D-               PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.00061      D*-  D+               PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "# JB PDG07",
    "0.00082      D*+  D*-              PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "#uncertain cabibbo suppressed decay is estimated by corresponding allowed decay",
    "# ex) Br(D-D+) = (f(D)/f(D_s))^2 * tan^2 theta_c * Br(D-D_s+),",
    "# where (f(D)/f(D_s))^2=(f(D*)/f(D_s*))^2=1.35 and tan^2 theta_c=0.05",
    "# fudge factor to tune DDX: ~0.909957",
    "0.000019   D'_1-  D+                PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000019   D'_1+  D-                PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000039   D'_1-  D*+               PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000039   D'_1+  D*-               PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000039   D_1-  D+                 PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000039   D_1+  D-                 PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000079   D_1-  D*+                PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000079   D_1+  D*-                PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000129   D_2*-  D+                PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;",
    "0.000129   D_2*+  D-                PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;",
    "0.000258   D_2*-  D*+               PHOTOS PHSP;",
    "0.000258   D_2*+  D*-               PHOTOS PHSP;",
    "# Ds1(2536)+=D_s1+  and  Ds2(2573)+=D_s2*+",
    "# DsJ(2700)+=D_sj(2700)+",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00043   D_s1+ D-                       PHOTOS SVS;",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00083   D_s1+ D*-                      PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00038   D_s2*+ D-                      PHOTOS STS;",
    "0.00038   D_s2*+ D*-                     PHOTOS PHSP;",
    "# from D_sj(2700)D0B  JB hep-ex/0707.3491",
    "0.00113   D_sj(2700)+ D-                 PHOTOS SVS;",
    "0.00038   D_sj(2700)+ D*-                PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00023   psi(3770) K_S0   PHOTOS SVS_CP beta dm  1 1 0 1 0.0;",
    "0.00023   psi(3770) K_L0   PHOTOS SVS_CP beta dm -1 1 0 1 0.0;",
    "0.00012   psi(3770) K*S  PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.00012   psi(3770) K*L  PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.00047   psi(3770) K*0T PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.00058   psi(3770) K+ pi-   PHOTOS PHSP;",
    "0.00029   psi(3770) K0 pi0   PHSP;",
    "0.00064   psi(3770) K0 pi+ pi- PHOTOS PHSP;",
    "0.00064   psi(3770) K0 pi0 pi0  PHSP;",
    "0.00064   psi(3770) K+ pi- pi0      PHOTOS PHSP;",
    "# JB fudge factor to tune DDX: ~0.6392116",
    "0.000703   psi(4040) K_S0  PHOTOS SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000703   psi(4040) K_L0  PHOTOS SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000352   psi(4040) K*S  PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000352   psi(4040) K*L  PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.001400   psi(4040) K*0T PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000741   psi(4040) K+ pi-  PHOTOS PHSP;",
    "0.000371   psi(4040) K0 pi0 PHSP;",
    "0.000703   psi(4160) K_S0  PHOTOS SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000703   psi(4160) K_L0  PHOTOS SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000352   psi(4160) K*S  PHOTOS SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000352   psi(4160) K*L  PHOTOS SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.001400   psi(4160) K*0T PHOTOS SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000741   psi(4160) K+ pi-    PHOTOS PHSP;",
    "0.000371   psi(4160) K0 pi0 PHSP;",
    "#new from qq98",
    "0.000027      D+   D-   pi0              PHOTOS PHSP;",
    "0.000027      D+   anti-D0  pi-          PHOTOS PHSP;",
    "0.000027      D0   D-  pi+               PHOTOS PHSP;",
    "0.000091      D*+   D-   pi0              PHOTOS PHSP;",
    "0.000091      D*+   anti-D0  pi-          PHOTOS PHSP;",
    "0.000091      D*0   D-  pi+               PHOTOS PHSP;",
    "0.000091      D+   D*-   pi0              PHOTOS PHSP;",
    "0.000091      D+   anti-D*0  pi-          PHOTOS PHSP;",
    "0.000091      D0   D*-  pi+               PHOTOS PHSP;",
    "0.000273      D*+   D*-   pi0              PHOTOS PHSP;",
    "0.000273      D*+   anti-D*0  pi-          PHOTOS PHSP;",
    "0.000273      D*0   D*-  pi+               PHOTOS PHSP;",
    "0.000009      D+   D-   rho0              PHOTOS PHSP;",
    "0.000009      D+   anti-D0  rho-          PHOTOS PHSP;",
    "0.000009      D0   D-  rho+               PHOTOS PHSP;",
    "0.000027      D*+   D-   rho0              PHOTOS PHSP;",
    "0.000027      D*+   anti-D0  rho-          PHOTOS PHSP;",
    "0.000027      D*0   D-  rho+               PHOTOS PHSP;",
    "0.000027      D+   D*-   rho0              PHOTOS PHSP;",
    "0.000027      D+   anti-D*0  rho-          PHOTOS PHSP;",
    "0.000027      D0   D*-  rho+               PHOTOS PHSP;",
    "0.000091      D*+   D*-   rho0              PHOTOS PHSP;",
    "0.000091      D*+   anti-D*0  rho-          PHOTOS PHSP;",
    "0.000091      D*0   D*-  rho+               PHOTOS PHSP;",
    "##",
    "0.000009      D+   D-   pi0  pi0         PHOTOS PHSP;",
    "0.000009      D+   D-   pi-  pi+         PHOTOS PHSP;",
    "0.000009      D+   anti-D0  pi-  pi0     PHOTOS PHSP;",
    "0.000009      D0   D-  pi+  pi0     PHOTOS PHSP;",
    "0.000009      D0   anti-D0  pi-  pi+     PHOTOS PHSP;",
    "0.000027      D*+   D-   pi0  pi0         PHOTOS PHSP;",
    "0.000027      D*+   D-   pi-  pi+         PHOTOS PHSP;",
    "0.000027      D*+   anti-D0  pi-  pi0     PHOTOS PHSP;",
    "0.000027      D*0   D-  pi+  pi0     PHOTOS PHSP;",
    "0.000027      D*0   anti-D0  pi-  pi+     PHOTOS PHSP;",
    "0.000027      D+   D*-   pi0  pi0         PHOTOS PHSP;",
    "0.000027      D+   D*-   pi-  pi+         PHOTOS PHSP;",
    "0.000027      D+   anti-D*0  pi-  pi0     PHOTOS PHSP;",
    "0.000027      D0   D*-  pi+  pi0     PHOTOS PHSP;",
    "0.000027      D0   anti-D*0  pi-  pi+     PHOTOS PHSP;",
    "0.000082      D*+   D*-   pi0  pi0         PHOTOS PHSP;",
    "0.000082      D*+   D*-   pi-  pi+         PHOTOS PHSP;",
    "0.000082      D*+   anti-D*0  pi-  pi0     PHOTOS PHSP;",
    "0.000082      D*0   D*-  pi+  pi0     PHOTOS PHSP;",
    "0.000082      D*0   anti-D*0  pi-  pi+     PHOTOS PHSP;",
    "# b -> c (sc=)  -->  D D= K  X   Sum = 0.0%",
    "# take Q2B into account; some values are taken from hep-ex/0207086",
    "# JB PDG07",
    "0.001700   D-  D0  K+                    PHOTOS PHSP;",
    "# (from D0BD0K+) JB PDG07",
    "0.001370   D-  D+   K0                   PHOTOS PHSP;",
    "0.000455   D-  D0  K*+                   PHOTOS PHSP;",
    "0.000455   D-  D+   K*0                  PHOTOS PHSP;",
    "# JB PDG07",
    "0.003100   D*- D0  K+                    PHOTOS PHSP;",
    "#(including D-D*+K0) JB PDG07",
    "0.006500   D*- D+  K0                    PHOTOS PHSP;",
    "# JB PDG07",
    "0.004600   D-  D*0  K+                   PHOTOS PHSP;",
    "0.000910   D*- D0   K*+                  PHOTOS PHSP;",
    "0.001820   D*- D+   K*0                  PHOTOS PHSP;",
    "0.001365   D-  D*0  K*+                  PHOTOS PHSP;",
    "0.001365   D-  D*+  K*0                  PHOTOS PHSP;",
    "# JB PDG07",
    "0.011800   D*- D*0  K+                   PHOTOS PHSP;",
    "# JB PDG07",
    "0.008800   D*- D*+   K0                  PHOTOS PHSP;",
    "0.003185   D*- D*0   K*+                 PHOTOS PHSP;",
    "0.003185   D*- D*+   K*0                 PHOTOS PHSP;",
    "0.000118   D_1- D0   K+                    PHOTOS PHSP;",
    "0.000118   D_1- D+   K0                    PHOTOS PHSP;",
    "0.000118   D-  D_10  K+                    PHOTOS PHSP;",
    "0.000118   D-  D_1+  K0                    PHOTOS PHSP;",
    "0.000355   D_1- D*0  K+                    PHOTOS PHSP;",
    "0.000355   D_1- D*+  K0                    PHOTOS PHSP;",
    "0.000355   D*- D_10  K+                    PHOTOS PHSP;",
    "0.000355   D*- D_1+  K0                    PHOTOS PHSP;",
    "0.000118   D_2*- D0   K+                    PHOTOS PHSP;",
    "0.000118   D_2*- D+   K0                    PHOTOS PHSP;",
    "0.000118   D-  D_2*0  K+                    PHOTOS PHSP;",
    "0.000118   D-  D_2*+  K0                    PHOTOS PHSP;",
    "0.000355   D_2*- D*0  K+                    PHOTOS PHSP;",
    "0.000355   D_2*- D*+  K0                    PHOTOS PHSP;",
    "0.000355   D*- D_2*0  K+                    PHOTOS PHSP;",
    "0.000355   D*- D_2*+  K0                    PHOTOS PHSP;",
    "0.000091   D-  D0   K+  pi0                  PHOTOS PHSP;",
    "0.000091   anti-D0  D0   K+  pi-             PHOTOS PHSP;",
    "0.000091   D-  D0   K0  pi+                  PHOTOS PHSP;",
    "0.000091   D-  D+   K+ pi-                   PHOTOS PHSP;",
    "0.000091   D-  D+   K0 pi0                   PHOTOS PHSP;",
    "0.000091   anti-D0  D+   K0  pi-             PHOTOS PHSP;",
    "0.000182   D*-  D0   K+  pi0                  PHOTOS PHSP;",
    "0.000182   anti-D*0  D0   K+  pi-             PHOTOS PHSP;",
    "0.000182   D*-  D0   K0  pi+                  PHOTOS PHSP;",
    "0.000182   D*-  D+   K+ pi-                   PHOTOS PHSP;",
    "0.000182   D*-  D+   K0 pi0                   PHOTOS PHSP;",
    "0.000182   anti-D*0  D+   K0  pi-             PHOTOS PHSP;",
    "0.000182   D-  D*0   K+  pi0                  PHOTOS PHSP;",
    "0.000182   anti-D0  D*0   K+  pi-             PHOTOS PHSP;",
    "0.000182   D-  D*0   K0  pi+                  PHOTOS PHSP;",
    "0.000182   D-  D*+   K+ pi-                   PHOTOS PHSP;",
    "0.000182   D-  D*+   K0 pi0                   PHOTOS PHSP;",
    "0.000182   anti-D0  D*+   K0  pi-             PHOTOS PHSP;",
    "0.000384   D*-  D*0   K+  pi0                  PHOTOS PHSP;",
    "0.000385   anti-D*0  D*0   K+  pi-             PHOTOS PHSP;",
    "0.000385   D*-  D*0   K0  pi+                  PHOTOS PHSP;",
    "0.000385   D*-  D*+   K+ pi-                   PHOTOS PHSP;",
    "0.000384   D*-  D*+   K0 pi0                   PHOTOS PHSP;",
    "0.000385   anti-D*0  D*+   K0  pi-             PHOTOS PHSP;",
    "0.000400   D-  D0   K*+  pi0                  PHOTOS PHSP;",
    "0.000400   anti-D0  D0   K*+  pi-             PHOTOS PHSP;",
    "0.000400   D-  D0   K*0  pi+                  PHOTOS PHSP;",
    "0.000400   D-  D+   K*+ pi-                   PHOTOS PHSP;",
    "0.000400   D-  D+   K*0 pi0                   PHOTOS PHSP;",
    "0.000400   anti-D0  D+   K*0  pi-             PHOTOS PHSP;",
    "0.000400   D-  D0   K+  rho0                  PHOTOS PHSP;",
    "0.000400   anti-D0  D0   K+  rho-             PHOTOS PHSP;",
    "0.000400   D-  D0   K0  rho+                  PHOTOS PHSP;",
    "0.000400   D-  D+   K+ rho-                   PHOTOS PHSP;",
    "0.000400   D-  D+   K0 rho0                   PHOTOS PHSP;",
    "0.000400   anti-D0  D+   K0  rho-             PHOTOS PHSP;",
    "# color suppressed mode",
    "0.000364   D0  anti-D0  K0               PHSP;",
    "0.000091   D0  anti-D0  K*0              PHSP;",
    "#(from D*-D+K+) JB PDG07",
    "0.001500   D0  anti-D*0  K0              PHSP;",
    "#(from D-D*+K+) JB PDG07",
    "0.001500   D*0  anti-D0  K0              PHSP;",
    "0.000364   D0  anti-D*0  K*0             PHSP;",
    "0.000364   D*0  anti-D0  K*0             PHSP;",
    "0.002548   D*0  anti-D*0  K0             PHSP;",
    "0.000819   D*0  anti-D*0  K*0            PHSP;",
    "0.000091   D_10  anti-D0  K0                PHSP;",
    "0.000091   D0  anti-D_10  K0                PHSP;",
    "0.000273   D_10  anti-D*0  K0               PHSP;",
    "0.000273   D*0  anti-D_10  K0               PHSP;",
    "0.000091   D_2*0  anti-D0  K0                PHSP;",
    "0.000091   D0  anti-D_2*0  K0                PHSP;",
    "0.000273   D_2*0  anti-D*0  K0               PHSP;",
    "0.000273   D*0  anti-D_2*0  K0               PHSP;",
    "# B->DX",
    "#  B -> D(*) X  Exclusive modes",
    "#    updated by T.GERSHON 20031027",
    "#    updated by Horii 20080131",
    "#  B0 -> D- n(pi)",
    "### PDG 2007",
    "0.00268  D-  pi+                  PHOTOS PHSP;",
    "### not in PDG",
    "0.0010   D-  pi+  pi0             PHOTOS PHSP;",
    "### PDG 2007",
    "0.0075   rho+ D-                  PHOTOS SVS;",
    "### PDG 2007",
    "0.004705   D-  pi-  pi+  pi+        PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0009   D-  pi0  pi+  pi0        PHOTOS PHSP;",
    "### PDG 2007 (presume excludes D+ a_1-)",
    "0.0014   D-  rho0  pi+            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0009   D-  rho+  pi0            PHOTOS PHSP;",
    "### PDG 2007",
    "0.0075   a_1+ D-                  PHOTOS SVS;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   b_1+ D-                  PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0000   D-   pi-  pi+  pi+  pi0  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0028   D-   omega  pi+          PHOTOS PHSP;",
    "### not in PDG 2007 !updated!",
    "0.00034   D-   p+  anti-p-  pi+   PHOTOS PHSP;",
    "### not in PDG 2007 - set same as anti-D* case",
    "0.00150   D-   p+  anti-n0         PHOTOS PHSP;",
    "#  B0 -> D*- n(pi)",
    "### PDG 2007",
    "0.00276  D*- pi+                  PHOTOS SVS;",
    "### need to exclude D*rho from PDG value",
    "### see http://belle.kek.jp/secured/directcp/dpi/minutes/2002/0628/shiro.pdf",
    "0.00060  D*- pi+  pi0             PHOTOS PHSP;",
    "### HELAMP parameters taken from ICHEP 98-852  (unchanged)",
    "### BF and HELAMP parameters updated in  PRD 67:112002,2003 (CLEO)",
    "### not yet in PDG - leave the same for now",
    "0.00680  rho+ D*-                 PHOTOS SVV_HELAMP 0.152 1.47 0.936 0 0.317 0.19;",
    "### PDG 2007",
    "0.0011   D*-  pi-  pi+  pi+       PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0010   D*-  pi0  pi+  pi0       PHOTOS PHSP;",
    "### PDG 2007 (presume excludes D+ a_1-)",
    "0.0072   D*-  rho0  pi+           PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0010   D*-  rho+  pi0           PHOTOS PHSP;",
    "### PDG 2007",
    "### helicity amplitudes taken from factorization",
    "0.0130   D*-  a_1+                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   D*-  b_1+                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### PDG 2007",
    "0.0176   D*-  pi-  pi+  pi+  pi0  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0029   D*-  omega  pi+          PHOTOS PHSP;",
    "### PDG 2007",
    "0.00065   D*-   p+  anti-p-  pi+  PHOTOS PHSP;",
    "### PDG 2007   ???",
    "0.00150   D*-   p+  anti-n0       PHOTOS PHSP;",
    "#  B0 -> D**- n(pi)",
    "## these are only UL in PDG2002",
    "## PDG2002 G32 : B0 -> D2(2460)*+ pi+  < 2.2 10^-3",
    "## PDG2002 G33 : B0 -> D2(2460)*+ rho- < 4.9 10^-3",
    "## make BFs the same as those for D**0 pi-",
    "## except exclude broad states on advice from Kuzmin",
    "#0.00083   D_0*- pi+                PHOTOS PHSP;",
    "0.00093   D_1- pi+                 PHOTOS SVS;",
    "#0.00062   D'_1- pi+                PHOTOS SVS;",
    "0.00085   D_2*- pi+                PHOTOS STS;",
    "## do the same for D**- rho+",
    "## not sure these should be modelled by phase space but it will do",
    "#0.00083   D_0*- rho+              PHOTOS PHSP;",
    "0.00093   D_1- rho+                PHOTOS PHSP;",
    "#0.00062   D'_1- rho+               PHOTOS PHSP;",
    "0.00085   D_2*- rho+               PHOTOS PHSP;",
    "#  B0B -> D**0 h0",
    "# These are added from PDG2005.",
    "## PDG2005 G47 : B0B -> D_2*0 pi+ = 0.00053",
    "## PDG2005 G48 : B0B -> D_1 pi+ = 0.00102",
    "## PDG2005 G49 : B0B -> D'_1 pi+ = 0.00075",
    "## PDG2005 G50 : B0B -> D_0*0 pi+ = 0.00091",
    "# and",
    "## PDG2005 G91 : B0B -> D*0 pi0 = 0.00027",
    "## PDG2005 G92 : B0B -> D*0 rho0 = 0.00051 (90% U.L.)",
    "## PDG2005 G93 : B0B -> D*0 eta = 0.00026",
    "## PDG2005 G94 : B0B -> D*0 eta' = 0.00026 (90% U.L.)",
    "0.00003   anti-D_2*0 pi0                STS;",
    "0.00006   anti-D_10 pi0                 SVS;",
    "0.00004   anti-D'_10 pi0                SVS;",
    "0.00005   anti-D_0*0 pi0                PHSP;",
    "0.00003   anti-D_2*0 eta                STS;",
    "0.00006   anti-D_10 eta                 SVS;",
    "0.00004   anti-D'_10 eta                SVS;",
    "0.00005   anti-D_0*0 eta                PHSP;",
    "0.00003   anti-D_2*0 eta'                STS;",
    "0.00006   anti-D_10 eta'                 SVS;",
    "0.00004   anti-D'_10 eta'                SVS;",
    "0.00005   anti-D_0*0 eta'                PHSP;",
    "0.00005   anti-D_2*0 omega               PHSP;",
    "0.00009   anti-D_10 omega                PHSP;",
    "0.00007   anti-D'_10 omega               PHSP;",
    "0.00008   anti-D_0*0 omega              PHSP;",
    "0.00006   anti-D_2*0 rho0               PHSP;",
    "0.00011   anti-D_10 rho0                PHSP;",
    "0.00008   anti-D'_10 rho0               PHSP;",
    "0.00010   anti-D_0*0 rho0              PHSP;",
    "#  B0 -> anti-D0 X - colour suppressed",
    "### PDG 2007",
    "0.00026  anti-D0 pi0                   PHSP;",
    "### PDG 2003 is confusing",
    "### Kuzmin is updating - leave same for now",
    "0.00045  anti-D0  pi-  pi+             PHOTOS PHSP;",
    "### PDG 2007",
    "0.00029  rho0 anti-D0                  SVS;",
    "### not in PDG 2007",
    "0.0005   anti-D0  pi0  pi0             PHSP;",
    "### PDG 2007",
    "0.00020  anti-D0 eta                   PHSP;",
    "### PDG 2007",
    "0.00013  anti-D0 eta'                  PHSP;",
    "### PDG 2007",
    "0.00026  omega anti-D0                 SVS;",
    "### not in PDG 2007",
    "0.0010   anti-D0  pi-  pi+  pi0        PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0001   anti-D0  pi0  pi0  pi0        PHSP;",
    "### not in PDG 2003 - PRL 89, 151802 (2002)",
    "### nb. phase space is not a good model",
    "0.000114 anti-D0 p+ anti-p-            PHOTOS PHSP;",
    "#  B0 -> anti-D*0 X - colour suppressed",
    "### PDG 2007",
    "0.00017  anti-D*0 pi0                  SVS;",
    "### PDG 2003 is confusing",
    "### Kuzmin is updating - leave same for now",
    "0.00032  anti-D*0 pi-  pi+             PHOTOS PHSP;",
    "### not in PDG 2007",
    "### BF set same as anti-D0 case, SVV_HELAMP parameters are the same as in D*0 rho-",
    "0.00029  anti-D*0 rho0                 SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### not in PDG 2007",
    "0.0005   anti-D*0 pi0  pi0             PHSP;",
    "### PDG 2007",
    "0.00018  anti-D*0 eta                  SVS;",
    "### PDG 2007 - set same as anti-D0 case !updated!",
    "0.00012  anti-D*0 eta'                 SVS;",
    "### PDG 2007",
    "0.00027  anti-D*0 omega                SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### not in PDG 2007",
    "0.0010   anti-D*0  pi-  pi+  pi0       PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0001   anti-D*0  pi0  pi0  pi0       PHSP;",
    "### not in PDG 2003 - PRL 89, 151802 (2002)",
    "### nb. phase space is not a good model",
    "0.000103 anti-D*0 p+ anti-p-           PHOTOS PHSP;",
    "#  B0 -> D(*)- K(*)+",
    "### PDG 2007",
    "0.00020  D-  K+                   PHOTOS PHSP;",
    "### PDG 2007",
    "0.00045  K*+ D-                   PHOTOS SVS;",
    "### PDG 2007",
    "0.00021  D*- K+                   PHOTOS SVS;",
    "### PDG 2007 - helicity amplitudes same as D*0 rho-   ???",
    "0.00033  D*- K*+                  PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "# colour suppressed B -> D(*)0 K(*)0",
    "### PDG 2007",
    "0.000052 anti-D0 K0               PHSP;",
    "### PDG 2007",
    "0.000042 K*0 anti-D0              SVS;",
    "### not in PDG 2007 - assume same as anti-D0 case",
    "0.000036 anti-D*0 K0              SVS;",
    "### not in PDG 2007 - assume same as anti-D0 case - HELAMP parameters same as D*- K*+",
    "0.000042 anti-D*0 K*0             SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### 2005 update (same as 2003)",
    "###  - assume same iso-spin violation in b -> c (s ubar) as b -> c (d ubar)",
    "###  - make b -> c (s ubar) sum to 2.56% (2.5% / 0.9764)",
    "###  - sum of b -> c (s ubar) BFs is 0.001346 (2005 -> 0.001303)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.024297  u  anti-s  anti-c  d            PHOTOS PYTHIA 48;",
    "# leftovers",
    "### B0 -> wrong flavour charm",
    "### excluded since not b->c transitions",
    "# mopping up",
    "### 2005 update (same as 2003)",
    "###  - include iso-spin violating effect",
    "###  - make b -> c (d ubar) + b -> c (s ubar) sum to 46.6% (2005 -> 49.8%)",
    "###  - sum of BFs is 11.4038% (2005 -> 11.8358%)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra is softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.372909  u  anti-d  anti-c  d           PHOTOS PYTHIA 48;",
    "#0.379642  u  anti-d  anti-c  d           PHOTOS PYTHIA 48;",
    "#0.363642  u  anti-d  anti-c  d           PHOTOS PYTHIA 48;",
    "# B->BaryonX",
    "# B0 -> charmed baryons, total 9.7%",
    "#  Sigma_c X",
    "0.000022       anti-Lambda_c- p+                 PHOTOS PHSP;",
    "0.000022       anti-Sigma_c-  p+                 PHOTOS PHSP;",
    "0.000022       anti-Sigma_c0  n0                 PHSP;",
    "0.018000       anti-cd_0       ud_0         PHOTOS PYTHIA 23;",
    "0.018000         anti-cd_1       ud_1         PHOTOS PYTHIA 23;",
    "0.006000       anti-cd_0       ud_0         PHOTOS PYTHIA 24;",
    "0.005932         anti-cd_1       ud_1         PHOTOS PYTHIA 24;",
    "#  double charmed baryon Xi_c Sigma_c",
    "#    two body decay",
    "0.001000       anti-Xi_c-     Lambda_c+          PHOTOS PHSP;",
    "0.001000       anti-Xi_c-     Sigma_c+           PHOTOS PHSP;",
    "0.001000       anti-Xi_c0     Sigma_c0           PHSP;",
    "0.009000       anti-cs_0      cd_0         PHOTOS PYTHIA 23;",
    "#  cabibbo suppressed  Xi_c N",
    "#    two body decay",
    "0.000001       anti-Xi_c-     p+                PHOTOS PHSP;",
    "0.000001       anti-Xi_c0     n0                PHSP;",
    "0.000900       anti-cs_0      ud_0        PHOTOS PYTHIA 23;",
    "0.000900       anti-cs_1      ud_1        PHOTOS PYTHIA 23;",
    "0.000300       anti-cs_0      ud_0        PHOTOS PYTHIA 24;",
    "0.000300       anti-cs_1      ud_1        PHOTOS PYTHIA 24;",
    "#  cabibbo suppressed  Sigma_c Sigma_c",
    "#    two body decay",
    "0.000040        anti-Lambda_c-    Lambda_c+     PHOTOS PHSP;",
    "0.000040        anti-Sigma_c-     Lambda_c+     PHOTOS PHSP;",
    "0.000040        anti-Lambda_c-    Sigma_c+      PHOTOS PHSP;",
    "0.000040        anti-Sigma_c-     Sigma_c+      PHOTOS PHSP;",
    "0.000040        anti-Sigma_c0     Sigma_c0      PHSP;",
    "0.000400        anti-cd_0    cd_0         PHOTOS PYTHIA 23;",
    "#",
    "Enddecay",
]  # type: List[str]

original_anti_b_zero_decay_lines = [
    "Decay anti-B0",
    "# B->DXlnu",
    "#     anti-B0 -> X_c l nu, sum = 10.45% (using: tau_B+/tau_B0 = 1.09)",
    "#     2005/04/02 H.Kakuno, Update using the following numbers:",
    "#       B->Xlnu     Br=10.9%               (HFAG summer 2004)",
    "#       B0->Dlnu    Br=2.14%,  rho_D=1.15  (HFAG summer 2004)",
    "#       B0->D*lnu   Br=5.33%               (HFAG summer 2004),",
    "#                    rho_A1=1.3            (hepex/0408027),",
    "#                    R1=1.18,  R2=0.71     (PRL76,3893)",
    "#       B->D**lnu   Br=3.04%          (CERN-EP/2001-050)",
    "#       B->D_1 lnu  Br=0.74%          (PDG 2004)",
    "#       B->D_2* lnu Br<0.65%          (PDG 2004)",
    "#       B->D pi lnu Br=1.54%          (PLB475, 407)",
    "#       B->D*pi lnu Br=1.86%          (PLB475, 407)",
    "#       referred also D0 preliminary result from CKM2005 for D**lnu",
    "0.0533   D*+       e-   anti-nu_e   PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0213   D+        e-   anti-nu_e   PHOTOS  HQET2 1.16;",
    "0.0074   D_1+      e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0084   D_0*+     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0074   D'_1+     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0036   D_2*+     e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0002   D(2S)+    e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0005   D*(2S)+   e-   anti-nu_e   PHOTOS  ISGW2;",
    "0.0003   D*+  pi0  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0006   D*0  pi+  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0005   D+   pi0  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0010   D0   pi+  e-   anti-nu_e   PHOTOS  GOITY_ROBERTS;",
    "0.0533   D*+       mu-  anti-nu_mu  PHOTOS  HQET2 1.3 1.18 0.71;",
    "0.0213   D+        mu-  anti-nu_mu  PHOTOS  HQET2 1.16;",
    "0.0074   D_1+      mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0084   D_0*+     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0074   D'_1+     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0036   D_2*+     mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0002   D(2S)+    mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0005   D*(2S)+   mu-  anti-nu_mu  PHOTOS  ISGW2;",
    "0.0003   D*+  pi0  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0006   D*0  pi+  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0005   D+   pi0  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;",
    "0.0010   D0   pi+  mu-  anti-nu_mu  PHOTOS  GOITY_ROBERTS;# ====B->DXlnu End",
    "# B->DXtaunu",
    "#                    B0 -> X_c tau nu, sum = 2.4% (using: tau_B+/tau_B0 = 1.09)",
    "#                    H.Kakuno 2005/04/02",
    "0.0142   D*+       tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0071   D+        tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0012   D_1+      tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0015   D_0*+     tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0012   D'_1+     tau- anti-nu_tau         PHOTOS ISGW2;",
    "0.0008   D_2*+     tau- anti-nu_tau         PHOTOS ISGW2;# ====B->DXtaunu End",
    "# B->ccX",
    "# B -> charmonium updates:",
    "#      Mar. 2001:  Lange",
    "#      Nov. 2002:  Schrenk",
    "#      Oct. 2003:  H. Guler",
    "#      Mar. 2005:  S.E.Lee, Takada and Ishikawa",
    "#      Apr. 2006:  K.Miyabayashi",
    "#       BF[%]           B0      B+",
    "#       eta_c           0.5034  0.5455",
    "#       eta_c(2S)       0.2015  0.2190",
    "#       J/psi           0.7539  0.8118",
    "#       psi(2S)         0.2967  0.3152",
    "#       chi_c0          0.2127  0.2291",
    "#       chi_c1          0.3228  0.3476",
    "#       chi_c2          0.1595  0.1718",
    "#       h_c             0.1345  0.1449",
    "#       X(3872)         0.0080  0.0080",
    "#       Y(3940)         0.0070  0.0071",
    "# -------------------------------------",
    "#       sum             2.6000  2.8000",
    "# anti-B0 --> charmonium X,  2.6000%",
    "# B -> cc= s",
    "# eta_c inclusive",
    "# PDG 2005: < 0.009",
    "# Guess 0.005034",
    "# anti-B0 --> eta_c K0",
    "# PDG 2005: 0.00116 +/- 0.00026",
    "0.000580   eta_c K_S0    SSS_CP beta dm -1 1 0 1 0.0;",
    "0.000580   eta_c K_L0    SSS_CP beta dm  1 1 0 1 0.0;",
    "# anti-B0 --> eta_c K*0(892)",
    "# PDG 2005: 0.0016 +/- 0.0007",
    "0.000266   anti-K*S eta_c          SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000266   anti-K*L eta_c          SVS_CP beta dm  1 1 0 1 0.0;",
    "0.001068   anti-K*0T eta_c    SVS;",
    "# Fill up inclusive rate with low momentum (0.00108)",
    "0.000100   eta_c K- pi+          PHOTOS PHSP;",
    "0.000051   eta_c anti-K0 pi0      PHSP;",
    "0.000308   eta_c anti-K0 pi+ pi- PHOTOS PHSP;",
    "0.000308   eta_c anti-K0 pi0 pi0  PHSP;",
    "0.000308   eta_c K- pi+ pi0      PHOTOS PHSP;",
    "# baryonic decays",
    "0.000003   eta_c Lambda0 anti-n0  PHSP;",
    "0.000003   eta_c Sigma+ anti-p- PHOTOS PHSP;",
    "# Fill up by string fragmentation.",
    "0.001193  eta_c anti-Xsd  PHSP;",
    "# Sum of eta_c: 0.005034",
    "# eta_c(2S) inclusive",
    "# Guess 0.002015  (40% of eta_c)",
    "# anti-B0 --> eta_c(2S) anti-K0",
    "# Guess 0.00048",
    "0.000240   eta_c(2S) K_S0     SSS_CP beta dm -1 1 0 1 0.0;",
    "0.000240   eta_c(2S) K_L0     SSS_CP beta dm  1 1 0 1 0.0;",
    "# anti-B0 --> eta_c(2S) anti-K0*",
    "# Guess 0.00064",
    "0.000106   anti-K*S eta_c(2S)         SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000106   anti-K*L eta_c(2S)         SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000428   anti-K*0T eta_c(2S)   SVS;",
    "# Fill up inclusive rate with low momentum (0.00048)",
    "0.000000   eta_c(2S) K- pi+          PHOTOS PHSP;",
    "0.000000   eta_c(2S) anti-K0 pi0      PHSP;",
    "0.000140   eta_c(2S) anti-K0 pi+ pi- PHOTOS PHSP;",
    "0.000140   eta_c(2S) anti-K0 pi0 pi0  PHSP;",
    "0.000140   eta_c(2S) K- pi+ pi0      PHOTOS PHSP;",
    "# Fill up by string fragmentation",
    "0.000475  eta_c(2S) anti-Xsd  PHSP;",
    "# Sum of eta_c(2S): 0.002015",
    "# J/psi inclusive (direct)",
    "# PDG 2003: 0.007296 +/- 0.0004",
    "# PDG 2005: 0.007539(life B0=1.532ps, B+=1.638ps)",
    "# anti-B0 --> J/psi K0",
    "# PDG 2005: 0.00085 +/- 0.00005",
    "0.000425   J/psi K_S0         SVS_CP beta dm -1 1 0 1 0;",
    "0.000425   J/psi K_L0         SVS_CP beta dm  1 1 0 1 0;",
    "# anti-B0 --> J/psi K- pi+   PDG 2005: 0.0012  +/- 0.0006",
    "# anti-B0 --> J/psi K*0      PDG 2005: 0.00131 +/- 0.00007",
    "0.000218   J/psi anti-K*S          SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000218   J/psi anti-K*L          SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000874   J/psi anti-K*0T    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000000   J/psi K- pi+      PHOTOS PHSP; #?????",
    "# anti-B0 --> J/psi K1(1270)",
    "# PDG 2005: 0.0013 +/- 0.0005",
    "0.001300   J/psi anti-K_10    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# anti-B0 --> J/psi K1(1400)",
    "0.000000   J/psi anti-K'_10   SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# This is not part of K1(1270)?",
    "# anti-B0 --> J/psi anti-K0 pi+ pi-   PDG 2005: 0.0010  +/- 0.0004  (including rho0, below)",
    "# anti-B0 --> J/psi anti-K0 rho0      PDG 2005: 0.00054 +/- 0.00030",
    "0.000540   J/psi anti-K0 rho0       PHSP;",
    "0.000460   J/psi anti-K0 pi+ pi-  PHOTOS  PHSP;",
    "# anti-B0 --> J/psi K*(892)- pi+",
    "# PDG 2005: 0.0008   +/- 0.0004",
    "0.000800   J/psi K*- pi+          PHOTOS  PHSP;",
    "# anti-B0 --> J/psi anti-K*(892)0 pi+ pi-",
    "# PDG 2005: 0.00066  +/- 0.00022",
    "0.000660   J/psi anti-K*0 pi+ pi- PHOTOS  PHSP;",
    "# anti-B0 --> J/psi phi anti-K0",
    "0.000094   J/psi phi anti-K0  PHSP;",
    "# anti-B0 --> J/psi pi0",
    "# PDG 2005: 0.000022 +/- 0.000004",
    "0.000022   J/psi pi0          SVS_CP beta dm  1 1 0 1 0;",
    "# anti-B0 --> J/psi eta",
    "# PDG 2005: < 0.000027",
    "# M.C.Chang preliminary value (9.2 +- 1.5)*10^-6 -> 9*10^-6.",
    "0.000009   J/psi eta          SVS_CP beta dm  1 1 0 1 0;",
    "# anti-B0 --> J/psi pi+ pi-",
    "# PDG 2005: 0.000046 +/- 0.000009 (including rho0, below)",
    "0.000030   J/psi pi+ pi-     PHOTOS PHSP;",
    "# anti-B0 --> J/psi rho0",
    "# PDG 2005: 0.000016 +/- 0.000007",
    "0.000016   J/psi rho0         SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "# anti-B0 --> J/psi omega",
    "# PDG 2005: < 0.00027",
    "0.000000   J/psi omega        SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "# anti-B0 --> J/psi anti-K0*(1430) Belle (Nitesh) 0.000050",
    "#                    K0*(1430) --> Kpi 0.93 +/- 0.10",
    "0.000050   J/psi anti-K_0*0    SVS;",
    "# anti-B0 --> J/psi anti-K2*(1430) Belle (Nitesh): 0.000260",
    "#                   K2*(1430) --> Kpi 0.499 +/- 0.012",
    "# STV not implemented",
    "0.000260   J/psi anti-K_2*0    PHSP;",
    "# anti-B0 --> J/psi anti-K*(1680) Belle (Nitesh):  ?",
    "#                   K*(1680) -->  Kpi 0.387 +/- 0.025",
    "0.000005   J/psi anti-K''*0    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# Fill up J/psi inclusive rate",
    "# kpipi replaced by K1,",
    "# little evidence for Kpi other than K*, K*(1430)",
    "# -> such multi-body decays is treated by Xsd? K.Miyabayashi",
    "#0.000000   J/psi anti-K0 pi0      PHSP;",
    "#0.000514   J/psi anti-K0 pi0 pi0  PHSP;",
    "#0.000514   J/psi K- pi+ pi0      PHOTOS PHSP;",
    "# baryonic decays",
    "# Assume that B+ -> J/Psi Lambda pbar; PDG2005: 1.2 +0.9-0.6 *10^-5.",
    "0.000012   J/psi Lambda0 anti-n0  PHSP;",
    "0.000012   J/psi Sigma+ anti-p- PHOTOS PHSP;",
    "0.001109   J/psi anti-Xsd PHSP;",
    "# Sum of J/psi: 0.007539  20060407",
    "# Psi(2S) inclusive",
    "# PDG 2005: 0.002967",
    "# anti-B0 --> psi(2S) K0",
    "# PDG 2003: 0.00062 +/- 0.00007",
    "0.000310   psi(2S) K_S0    SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000310   psi(2S) K_L0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# anti-B0 --> psi(2S) K- pi+",
    "# PDG 2005: < 0.001 (including K*0(892) below)",
    "# anti-B0 --> psi(2S) K*0(892)",
    "# PDG 2005: 0.00080 +/- 0.00013",
    "# Belle-Conf 0349 (LP03): 0.000720 +/- 0.000043 +/- 0.000065",
    "0.000120   psi(2S) anti-K*S    SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000120   psi(2S) anti-K*L    SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000480   psi(2S) anti-K*0T    SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# anti-B0 --> psi(2S) pi0",
    "# PDG 2005: No record, N.Kanagawa's master thesis(2004)",
    "0.000020   psi(2S) pi0          SVS_CP beta dm  1 1 0 1 0;",
    "# Fill up inclusive rate with low momentum",
    "#0.000000   psi(2S) K- pi+          PHOTOS PHSP;",
    "#0.000000   psi(2S) anti-K0 pi0      PHSP;",
    "#0.000536   psi(2S) anti-K0 pi+ pi- PHOTOS PHSP; # set to fill up PDG2005",
    "#0.000536   psi(2S) anti-K0 pi0 pi0  PHSP;       #",
    "#0.000535   psi(2S) K- pi+ pi0      PHOTOS PHSP; #",
    "# Use anti-Xsd.",
    "0.001607   psi(2S) anti-Xsd PHSP;",
    "# Sum of Psi(2S): 0.002967",
    "# chi_c0 inclusive",
    "# Guess 0.001750+0.000427",
    "# PDG 2003: < 0.0005",
    "# Guess 0.00056",
    "# Guess 0.00050 after adjusted value.(2004)",
    "# cf), B+ -> chi_c0 K+ is (112+-12+-18)*10^-6 (hep-ex/0509001)",
    "0.000250   chi_c0 K_S0   SSS_CP beta dm  1 1 0 1 0.0;",
    "0.000250   chi_c0 K_L0   SSS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.00066",
    "0.000110   anti-K*S chi_c0          SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000110   anti-K*L chi_c0          SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000440   anti-K*0T chi_c0    SVS;",
    "# Fill up inclusive rate with low momentum",
    "#0.000092   chi_c0 K- pi+          PHOTOS  PHSP;",
    "#0.000092   chi_c0 anti-K0 pi0       PHSP;",
    "#0.000261   chi_c0 anti-K0 pi+ pi- PHOTOS  PHSP;",
    "#0.000261   chi_c0 anti-K0 pi0 pi0   PHSP;",
    "#0.000261   chi_c0 K- pi+ pi0      PHOTOS  PHSP;",
    "#Switch to string fragmentation.",
    "0.000967   chi_c0  anti-Xsd  PHSP;",
    "# Sum of chi_c0: 0.002127",
    "# chi_c1 inclusive (direct)",
    "# PDG 2005: 0.003228",
    "# PDG 2003: 0.00040 + 0.00012 - 0.00010",
    "0.000200   chi_c1 K_S0    SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000200   chi_c1 K_L0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# PDG 2003: 0.00041 +/- 0.00015",
    "# PDG 2005: same as PDG 2003.",
    "0.000068   chi_c1 anti-K*S   SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000068   chi_c1 anti-K*L   SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000274   chi_c1 anti-K*0T  SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# chi_c1 pi0 (1.8+-0.7+-0.2)*10^-5 M.Fujita's master thesis(2005)",
    "0.000018   chi_c1 pi0    SVS_CP beta dm  1 1 0 1 0.0;",
    "# Fill up inclusive rate with low momentum",
    "0.000000   chi_c1 K- pi+           PHOTOS PHSP;",
    "0.000000   chi_c1 anti-K0 pi0      PHSP;",
    "0.000500   chi_c1 anti-K0 pi+ pi-  PHOTOS PHSP;",
    "0.000500   chi_c1 anti-K0 pi0 pi0  PHSP;",
    "0.000500   chi_c1 K- pi+ pi0       PHOTOS PHSP;",
    "# String frafmentation.",
    "0.000900  chi_c1  anti-Xsd  PHSP;",
    "# Sum of chi_c1: 0.003228",
    "# chi_c2 inclusive (direct)",
    "# PDG 2005: 0.001595",
    "# No real evidence for these decays",
    "# Guess 0.000076",
    "0.000038   chi_c2 K_S0             STS_CP beta dm  1 1 0 1 0.0;",
    "0.000038   chi_c2 K_L0             STS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.000114",
    "# STV not implemented",
    "0.000019   chi_c2 anti-K*S              PHSP;",
    "0.000019   chi_c2 anti-K*L              PHSP;",
    "0.000076   chi_c2 anti-K*0T        PHSP;",
    "# Guess 0.000805",
    "0.000000   chi_c2 K- pi+          PHOTOS PHSP;",
    "0.000000   chi_c2 anti-K0 pi0      PHSP;",
    "0.000268   chi_c2 anti-K0 pi+ pi- PHOTOS PHSP;",
    "0.000268   chi_c2 anti-K0 pi0 pi0  PHSP;",
    "0.000269   chi_c2 K- pi+ pi0      PHOTOS PHSP;",
    "# String fragmentation.",
    "0.000600  chi_c2 anti-Xsd  PHSP;",
    "# Sum of chi_c2: 0.001595",
    "# h_c Inclusive",
    "# Guess 0.001345",
    "# Guess 0.000180",
    "0.000090   h_c K_S0             SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000090   h_c K_L0             SVS_CP beta dm -1 1 0 1 0.0;",
    "# Guess 0.000205",
    "# STV not implemented",
    "0.000034   h_c anti-K*S              PHSP;",
    "0.000034   h_c anti-K*L              PHSP;",
    "0.000137   h_c anti-K*0T        PHSP;",
    "# Guess 0.00096",
    "#0.000000   h_c K- pi+          PHOTOS PHSP;",
    "#0.000000   h_c anti-K0 pi0      PHSP;",
    "#0.000320   h_c anti-K0 pi+ pi- PHOTOS PHSP;",
    "#0.000320   h_c anti-K0 pi0 pi0  PHSP;",
    "#0.000320   h_c K- pi+ pi0      PHOTOS PHSP;",
    "# Switch to string fragmentation.",
    "0.000960    h_c  anti-Xsd  PHSP;",
    "# Sum of h_c: 0.001345",
    "# anti-B0 --> anti-K0 X(3872)",
    "# Register only B -> X(3872) K.",
    "0.000040   X(3872) K_S0  SVS_CP beta dm  -1 1 0 1 0.0;",
    "0.000040   X(3872) K_L0  SVS_CP beta dm  1 1 0 1 0.0;",
    "# omit B -> X(3872) K*, No signal according to Karim's work(BAM200509))",
    "#0.000120   X(3872) anti-K*0  SVV_HELAMP PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "# sum of X(3872) = 0.000080",
    "# anti-B0 --> anti-K0 Y(3940) assume that same as charged B.",
    "0.000035   Y(3940) K_S0  SVS_CP beta dm  -1 1 0 1 0.0;",
    "0.000035   Y(3940) K_L0  SVS_CP beta dm  1 1 0 1 0.0;",
    "# B->DsX",
    "# b -> c (sc=)  -->  D Ds X   Sum = 8.3%",
    "#   BF(Ds->phipi)=4.38(Babar2005+PDG) is assumed.",
    "#   So the BF listed here is 0.82 * BF listed in PDG.",
    "# Update (PDG2002), Y.Mikami 02/12/17",
    "# added DsJ(2317) and DsJ(2457). A.Ishikawa 20031106",
    "# updated by Ishikawa 20050329",
    "#   Updated on 29/01/2008",
    "#   based on PDG update (Ref: W.-M. Yao et al., J. Phys. G 33, 1 (2006) and 2007 partial update for the 2008)",
    "#   by Nikhil J Joshi (njoshi@bpost)",
    "## Upper Vertex Ds",
    "0.00650   D+  D_s-                      PHOTOS PHSP;",
    "0.00800   D*+  D_s-                     PHOTOS SVS;",
    "0.00740   D_s*- D+                      PHOTOS SVS;",
    "0.01770  D*+  D_s*-                     PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00033   D_s- D_0*+                    PHOTOS PHSP;",
    "0.00067   D_s*- D_0*+                   PHOTOS SVS;",
    "0.00031   D'_1+  D_s-                   PHOTOS SVS;",
    "0.00063   D'_1+  D_s*-                  PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00063   D_1+  D_s-                    PHOTOS SVS;",
    "0.00126   D_1+  D_s*-                   PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00165   D_2*+  D_s-                   PHOTOS STS;",
    "0.00329   D_2*+  D_s*-                  PHOTOS PHSP;",
    "# DsJ(2317)=D_s0*-  and  DsJ(2457)=D'_s1-",
    "0.00097   D+ D_s0*-                     PHOTOS PHSP;",
    "0.00150   D*+ D_s0*-                    PHOTOS SVS;",
    "0.00350   D'_s1- D+                     PHOTOS SVS;",
    "0.00930   D'_s1- D*+                    PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00091   D_s-  D+  pi0                 PHOTOS PHSP;",
    "0.00182   D_s-  D0  pi+                 PHOTOS PHSP;",
    "0.00091   D_s*- D+  pi0                 PHOTOS PHSP;",
    "0.00182   D_s*- D0  pi+                 PHOTOS PHSP;",
    "0.00151   D_s-  D+  pi+  pi-            PHOTOS PHSP;",
    "0.00070   D_s-  D+  pi0  pi0            PHOTOS PHSP;",
    "0.00151   D_s-  D0  pi+  pi0            PHOTOS PHSP;",
    "0.00151   D_s*- D+  pi+  pi-            PHOTOS PHSP;",
    "0.00070   D_s*- D+  pi0  pi0            PHOTOS PHSP;",
    "0.00151   D_s*- D0  pi+  pi0            PHOTOS PHSP;",
    "## End of upper vertex",
    "## Lower vertex Ds",
    "0.00095   D_s+  pi-  anti-K0                    PHOTOS PHSP;",
    "0.00095   D_s*+ pi-  anti-K0                    PHOTOS PHSP;",
    "0.00100   D_s+  pi-  anti-K0  pi0               PHOTOS PHSP;",
    "0.00100   D_s*+ pi-  anti-K0  pi0               PHOTOS PHSP;",
    "0.00080   D_s+  pi-  K-  pi+                    PHOTOS PHSP;",
    "0.00080   D_s*+ pi-  K-  pi+                    PHOTOS PHSP;",
    "0.00016   D_s+  pi-  anti-K0  pi0 pi0           PHOTOS PHSP;",
    "0.00016   D_s*+ pi-  anti-K0  pi0 pi0           PHOTOS PHSP;",
    "0.00016   D_s+  pi-  anti-K0  pi+ pi-           PHOTOS PHSP;",
    "0.00016   D_s*+ pi-  anti-K0  pi+ pi-           PHOTOS PHSP;",
    "0.00016   D_s+  pi-  K-  pi+ pi0                PHOTOS PHSP;",
    "0.00016   D_s*+ pi-  K-  pi+ pi0                PHOTOS PHSP;",
    "## ====B->DsX End",
    "## End of lower vertex",
    "# B0B->DDX",
    "# updated by H.Miyake 20021006",
    "# updated by H.Miyake 20031019",
    "# updated by J.Brodzicka 20080130 basing on PDG07",
    "# b-> c (dc=)",
    "# JB PDG07",
    "0.00023      D-   D+               PHOTOS SSS_CP beta_cdcd  dm  1 1.0 0.0 1.0 0.0;",
    "# JB PDG07",
    "0.00061      D*-  D+               PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.00061      D*+  D-               PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "# JB PDG07",
    "0.00082      D*+  D*-              PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "#uncertain cabibbo suppressed decay is estimated by corresponding allowed decay",
    "# ex) Br(D-D+) = (f(D)/f(D_s))^2 * tan^2 theta_c * Br(D-D_s+),",
    "# where (f(D)/f(D_s))^2=(f(D*)/f(D_s*))^2=1.35 and tan^2 theta_c=0.05",
    "# fudge factor to tune DDX: ~0.909957",
    "0.000019   D'_1-  D+                PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000019   D'_1+  D-                PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000039   D'_1-  D*+               PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000039   D'_1+  D*-               PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000039   D_1-  D+                 PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000039   D_1+  D-                 PHOTOS SVS_CP beta_cdcd  dm -1 1.0 0.0 1.0 0.0;",
    "0.000079   D_1-  D*+                PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000079   D_1+  D*-                PHOTOS SVV_CP beta_cdcd  dm -1 0.62 0.0 1.05 0.0 0.25 0.0;",
    "0.000129   D_2*-  D+                PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;",
    "0.000129   D_2*+  D-                PHOTOS STS_CP beta_cdcd  dm -1 1 0 1 0.0;",
    "0.000258   D_2*-  D*+               PHOTOS PHSP;",
    "0.000258   D_2*+  D*-               PHOTOS PHSP;",
    "# Ds1(2536)-=D_s1-  and  Ds2(2573)-=D_s2*-",
    "# DsJ(2700)-=D_sj(2700)-",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00043   D_s1- D+                       PHOTOS SVS;",
    "# JB PRD(RC) 77, 011102(2008)",
    "0.00083   D_s1- D*+                      PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00038   D_s2*- D+                      PHOTOS STS;",
    "0.00038   D_s2*- D*+                     PHOTOS PHSP;",
    "# from D_sj(2700) D0B JB hep-ex/0707.3491",
    "0.00113   D_sj(2700)- D+                 PHOTOS SVS;",
    "0.00038   D_sj(2700)- D*+                PHOTOS SVV_HELAMP 0.48 0.0 0.734 0.0 0.48 0.0;",
    "0.00023   psi(3770) K_S0   SVS_CP beta dm  1 1 0 1 0.0;",
    "0.00023   psi(3770) K_L0   SVS_CP beta dm -1 1 0 1 0.0;",
    "0.00012   psi(3770) anti-K*S  SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.00012   psi(3770) anti-K*L  SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.00047   psi(3770) anti-K*0T SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.00058   psi(3770) K- pi+   PHOTOS PHSP;",
    "0.00029   psi(3770) anti-K0 pi0   PHSP;",
    "0.00064   psi(3770) anti-K0 pi+ pi- PHOTOS PHSP;",
    "0.00064   psi(3770) anti-K0 pi0 pi0  PHSP;",
    "0.00064   psi(3770) K- pi+ pi0      PHOTOS PHSP;",
    "# JB fudge factor to tune DDX: ~0.6392116",
    "0.000703   psi(4040) K_S0  SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000703   psi(4040) K_L0  SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000352   psi(4040) anti-K*S  SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000352   psi(4040) anti-K*L  SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.001400   psi(4040) anti-K*0T SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000741   psi(4040) K- pi+  PHOTOS PHSP;",
    "0.000371   psi(4040) anti-K0 pi0 PHSP;",
    "0.000703   psi(4160) K_S0  SVS_CP beta dm  1 1 0 1 0.0;",
    "0.000703   psi(4160) K_L0  SVS_CP beta dm -1 1 0 1 0.0;",
    "0.000352   psi(4160) anti-K*S  SVV_CP beta dm  1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.000352   psi(4160) anti-K*L  SVV_CP beta dm -1 Aplus phAplus Azero phAzero Aminus phAminus;",
    "0.001400   psi(4160) anti-K*0T SVV_HELAMP    PKHplus PKphHplus PKHzero PKphHzero PKHminus PKphHminus;",
    "0.000741   psi(4160) K- pi+    PHOTOS PHSP;",
    "0.000371   psi(4160) anti-K0 pi0 PHSP;",
    "# new from qq98",
    "0.000027      D-   D+   pi0         PHOTOS PHSP;",
    "0.000027      D-   D0   pi+         PHOTOS PHSP;",
    "0.000027      anti-D0   D+   pi-         PHOTOS PHSP;",
    "0.000091      D*-  D+   pi0         PHOTOS PHSP;",
    "0.000091      D*-  D0   pi+         PHOTOS PHSP;",
    "0.000091      anti-D*0  D+   pi-         PHOTOS PHSP;",
    "0.000091      D-  D*+   pi0         PHOTOS PHSP;",
    "0.000091      D-  D*0   pi+         PHOTOS PHSP;",
    "0.000091      anti-D0  D*+   pi-         PHOTOS PHSP;",
    "0.000273      D*-  D*+   pi0         PHOTOS PHSP;",
    "0.000273      D*-  D*0   pi+         PHOTOS PHSP;",
    "0.000273      anti-D*0  D*+   pi-         PHOTOS PHSP;",
    "0.000009      D-   D+   rho0         PHOTOS PHSP;",
    "0.000009      D-   D0   rho+         PHOTOS PHSP;",
    "0.000009      anti-D0   D+   rho-         PHOTOS PHSP;",
    "0.000027      D*-  D+   rho0         PHOTOS PHSP;",
    "0.000027      D*-  D0   rho+         PHOTOS PHSP;",
    "0.000027      anti-D*0  D+   rho-         PHOTOS PHSP;",
    "0.000027      D-  D*+   rho0         PHOTOS PHSP;",
    "0.000027      D-  D*0   rho+         PHOTOS PHSP;",
    "0.000027      anti-D0  D*+   rho-         PHOTOS PHSP;",
    "0.000091      D*-  D*+   rho0         PHOTOS PHSP;",
    "0.000091      D*-  D*0   rho+         PHOTOS PHSP;",
    "0.000091      anti-D*0  D*+   rho-         PHOTOS PHSP;",
    "##",
    "0.000009      D-   D+   pi0  pi0    PHOTOS PHSP;",
    "0.000009      D-   D+   pi+  pi-    PHOTOS PHSP;",
    "0.000009      D-   D0   pi+  pi0    PHOTOS PHSP;",
    "0.000009      anti-D0   D+   pi-  pi0    PHOTOS PHSP;",
    "0.000009      anti-D0   D0   pi+  pi-    PHOTOS PHSP;",
    "0.000027      D*-   D+   pi0  pi0    PHOTOS PHSP;",
    "0.000027      D*-   D+   pi+  pi-    PHOTOS PHSP;",
    "0.000027      D*-   D0   pi+  pi0    PHOTOS PHSP;",
    "0.000027      anti-D*0   D+   pi-  pi0    PHOTOS PHSP;",
    "0.000027      anti-D*0   D0   pi+  pi-    PHOTOS PHSP;",
    "0.000027      D-   D*+   pi0  pi0    PHOTOS PHSP;",
    "0.000027      D-   D*+   pi+  pi-    PHOTOS PHSP;",
    "0.000027      D-   D*0   pi+  pi0    PHOTOS PHSP;",
    "0.000027      anti-D0   D*+   pi-  pi0    PHOTOS PHSP;",
    "0.000027      anti-D0   D*0   pi+  pi-    PHOTOS PHSP;",
    "0.000082      D*-   D*+   pi0  pi0    PHOTOS PHSP;",
    "0.000082      D*-   D*+   pi+  pi-    PHOTOS PHSP;",
    "0.000082      D*-   D*0   pi+  pi0    PHOTOS PHSP;",
    "0.000082      anti-D*0   D*+   pi-  pi0    PHOTOS PHSP;",
    "0.000082      anti-D*0   D*0   pi+  pi-    PHOTOS PHSP;",
    "# b -> c (sc=)  -->  D D= K  X   Sum = 0.0%",
    "#   average of B0 and B+ by taking account for the internal and external W emission",
    "# take Q2B into account; some values are taken from hep-ex/0207086",
    "# JB PDG07",
    "0.001700   D+  anti-D0  K-                    PHOTOS PHSP;",
    "# (from D0D0BK-) JB PDG07",
    "0.001370   D+  D-   anti-K0                   PHOTOS PHSP;",
    "0.000455   D+  anti-D0  K*-                   PHOTOS PHSP;",
    "0.000455   D+  D-   anti-K*0                  PHOTOS PHSP;",
    "# JB PDG07",
    "0.003100   D*+ anti-D0  K-                    PHOTOS PHSP;",
    "# (including D+D*-K0) JB PDG07",
    "0.006500   D*+ D-   anti-K0                   PHOTOS PHSP;",
    "# JB PDG07",
    "0.004600   D+  anti-D*0  K-                   PHOTOS PHSP;",
    "0.000910   D*+ anti-D0  K*-                   PHOTOS PHSP;",
    "0.001820   D*+ D-   anti-K*0                  PHOTOS PHSP;",
    "0.001365   D+  anti-D*0  K*-                  PHOTOS PHSP;",
    "0.001365   D+  D*-  anti-K*0                  PHOTOS PHSP;",
    "# JB PDG07",
    "0.011800   D*+ anti-D*0  K-                   PHOTOS PHSP;",
    "# JB PDG07",
    "0.008800   D*+ D*-   anti-K0                  PHOTOS PHSP;",
    "0.003185   D*+ anti-D*0  K*-                  PHOTOS PHSP;",
    "0.003185   D*+ D*-   anti-K*0                 PHOTOS PHSP;",
    "0.000118   D_1+ anti-D0  K-                    PHOTOS PHSP;",
    "0.000118   D_1+ D-   anti-K0                   PHOTOS PHSP;",
    "0.000118   D+  anti-D_10  K-                   PHOTOS PHSP;",
    "0.000118   D+  D_1-  anti-K0                   PHOTOS PHSP;",
    "0.000355   D_1+ anti-D*0  K-                   PHOTOS PHSP;",
    "0.000355   D_1+ D*-   anti-K0                  PHOTOS PHSP;",
    "0.000355   D*+ anti-D_10  K-                   PHOTOS PHSP;",
    "0.000355   D*+ D_1-   anti-K0                  PHOTOS PHSP;",
    "0.000118   D_2*+ anti-D0  K-                    PHOTOS PHSP;",
    "0.000118   D_2*+ D-   anti-K0                   PHOTOS PHSP;",
    "0.000118   D+  anti-D_2*0  K-                   PHOTOS PHSP;",
    "0.000118   D+  D_2*-  anti-K0                   PHOTOS PHSP;",
    "0.000355   D_2*+ anti-D*0  K-                   PHOTOS PHSP;",
    "0.000355   D_2*+ D*-   anti-K0                  PHOTOS PHSP;",
    "0.000355   D*+ anti-D_2*0  K-                   PHOTOS PHSP;",
    "0.000355   D*+ D_2*-   anti-K0                  PHOTOS PHSP;",
    "0.000091   D+  anti-D0  K-   pi0                 PHOTOS PHSP;",
    "0.000091   D0  anti-D0  K-   pi+                 PHOTOS PHSP;",
    "0.000091   D+  anti-D0  anti-K0   pi-            PHOTOS PHSP;",
    "0.000091   D+  D-   K-  pi+                      PHOTOS PHSP;",
    "0.000091   D+  D-   anti-K0  pi0                 PHOTOS PHSP;",
    "0.000091   D0  D-   anti-K0  pi+                 PHOTOS PHSP;",
    "0.000182   D*+  anti-D0  K-   pi0                 PHOTOS PHSP;",
    "0.000182   D*0  anti-D0  K-   pi+                 PHOTOS PHSP;",
    "0.000182   D*+  anti-D0  anti-K0   pi-            PHOTOS PHSP;",
    "0.000182   D*+  D-   K-  pi+                      PHOTOS PHSP;",
    "0.000182   D*+  D-   anti-K0  pi0                 PHOTOS PHSP;",
    "0.000182   D*0  D-   anti-K0  pi+                 PHOTOS PHSP;",
    "0.000182   D+  anti-D*0  K-   pi0                 PHOTOS PHSP;",
    "0.000182   D0  anti-D*0  K-   pi+                 PHOTOS PHSP;",
    "0.000182   D+  anti-D*0  anti-K0   pi-            PHOTOS PHSP;",
    "0.000182   D+  D*-   K-  pi+                      PHOTOS PHSP;",
    "0.000182   D+  D*-   anti-K0  pi0                 PHOTOS PHSP;",
    "0.000182   D0  D*-   anti-K0  pi+                 PHOTOS PHSP;",
    "0.000384   D*+  anti-D*0  K-   pi0                 PHOTOS PHSP;",
    "0.000385   D*0  anti-D*0  K-   pi+                 PHOTOS PHSP;",
    "0.000385   D*+  anti-D*0  anti-K0   pi-            PHOTOS PHSP;",
    "0.000385   D*+  D*-   K-  pi+                      PHOTOS PHSP;",
    "0.000384   D*+  D*-   anti-K0  pi0                 PHOTOS PHSP;",
    "0.000385   D*0  D*-   anti-K0  pi+                 PHOTOS PHSP;",
    "0.000400   D+  anti-D0  K*-   pi0                 PHOTOS PHSP;",
    "0.000400   D0  anti-D0  K*-   pi+                 PHOTOS PHSP;",
    "0.000400   D+  anti-D0  anti-K*0   pi-            PHOTOS PHSP;",
    "0.000400   D+  D-   K*-  pi+                      PHOTOS PHSP;",
    "0.000400   D+  D-   anti-K*0  pi0                 PHOTOS PHSP;",
    "0.000400   D0  D-   anti-K*0  pi+                 PHOTOS PHSP;",
    "0.000400   D+  anti-D0  K-   rho0                 PHOTOS PHSP;",
    "0.000400   D0  anti-D0  K-   rho+                 PHOTOS PHSP;",
    "0.000400   D+  anti-D0  anti-K0   rho-            PHOTOS PHSP;",
    "0.000400   D+  D-   K-  rho+                      PHOTOS PHSP;",
    "0.000400   D+  D-   anti-K0  rho0                 PHOTOS PHSP;",
    "0.000400   D0  D-   anti-K0  rho+                 PHOTOS PHSP;",
    "# color suppressed mode",
    "0.000364   D0  anti-D0  anti-K0               PHSP;",
    "0.000091   D0  anti-D0  anti-K*0              PHSP;",
    "#(from D*+D-K-) JB PDG07",
    "0.000364   D0  anti-D*0  anti-K*0             PHSP;",
    "0.001500   D*0  anti-D0  anti-K0              PHSP;",
    "#(from D+D*-K-) JB PDG07",
    "0.001500   D0   anti-D*0  anti-K0              PHSP;",
    "0.000364   D*0  anti-D0  anti-K*0             PHSP;",
    "0.002548   D*0  anti-D*0  anti-K0             PHSP;",
    "0.000819   D*0  anti-D*0  anti-K*0            PHSP;",
    "0.000091   D_10  anti-D0  anti-K0               PHSP;",
    "0.000091   D0  anti-D_10  anti-K0               PHSP;",
    "0.000273   D_10  anti-D*0  anti-K0              PHSP;",
    "0.000273   D*0  anti-D_10  anti-K0              PHSP;",
    "0.000091   D_2*0  anti-D0  anti-K0               PHSP;",
    "0.000091   D0  anti-D_2*0  anti-K0               PHSP;",
    "0.000273   D_2*0  anti-D*0  anti-K0              PHSP;",
    "0.000273   D*0  anti-D_2*0  anti-K0              PHSP;",
    "# ====B->DDX End",
    "# B->DX",
    "#  B0B -> D(*) X  Exclusive modes",
    "#    updated by T.GERSHON 20031027",
    "#    updated by Kikuchi 20060328",
    "#    updated by Horii 20080131",
    "#  B0B -> D+ n(pi)",
    "### PDG 2007",
    "0.00268   D+  pi-                  PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0010    D+  pi-  pi0             PHOTOS PHSP;",
    "### PDG 2007",
    "0.0075    rho- D+                  PHOTOS SVS;",
    "### PDG 2007 (nonresonant)",
    "0.004705    D+  pi+  pi-  pi-        PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0009    D+  pi0  pi-  pi0        PHOTOS PHSP;",
    "### PDG 2007 (presume excludes D+ a_1-)",
    "0.0014    D+  rho0  pi-            PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0009    D+  rho-  pi0            PHOTOS PHSP;",
    "### PDG 2007",
    "0.0075    a_1- D+                  PHOTOS SVS;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000    b_1- D+                  PHOTOS SVS;",
    "### not in PDG 2007",
    "0.0000    D+   pi+  pi-  pi-  pi0  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0028    D+   omega  pi-          PHOTOS PHSP;",
    "### PDG 2007",
    "0.00034   D+   p+  anti-p-  pi-    PHOTOS PHSP;",
    "### not in PDG 2007 - set same as D* case",
    "0.00150   D+   anti-p-  n0         PHOTOS PHSP;",
    "#  B0B -> D*+ n(pi)",
    "### PDG 2003",
    "0.00276  D*+ pi-                  PHOTOS SVS;",
    "### need to exclude D*rho from PDG value",
    "### see http://belle.kek.jp/secured/directcp/dpi/minutes/2002/0628/shiro.pdf",
    "0.00060  D*+ pi-  pi0             PHOTOS PHSP;",
    "### PDG 2007",
    "0.00680  rho- D*+                 PHOTOS SVV_HELAMP 0.152 1.47 0.936 0 0.317 0.19;",
    "### PDG 2007",
    "0.0011   D*+  pi+  pi-  pi-       PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0010   D*+  pi0  pi-  pi0       PHOTOS PHSP;",
    "### PDG 2007 (presume excludes D+ a_1-)",
    "0.0072   D*+  rho0  pi-           PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0010   D*+  rho-  pi0           PHOTOS PHSP;",
    "### PDG 2007",
    "### helicity amplitudes taken from factorization",
    "0.0130   D*+  a_1-                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### not in PDG 2007 (who put it in DECAY.DEC?)",
    "0.0000   D*+  b_1-                PHOTOS SVV_HELAMP 0.336 0.0 0.88 0.0 0.336 0.0;",
    "### PDG 2007",
    "0.0176   D*+  pi+  pi-  pi-  pi0  PHOTOS PHSP;",
    "### PDG 2007",
    "0.0029   D*+  omega  pi-          PHOTOS PHSP;",
    "### PDG 2007",
    "0.00065   D*+   p+  anti-p-  pi-  PHOTOS PHSP;",
    "### PDG 2003",
    "0.00150   D*+   anti-p-  n0       PHOTOS PHSP;",
    "#  B0B -> D**+ n(pi)",
    "## these are only UL in PDG2002",
    "## PDG2002 G32 : B0B -> D2(2460)*+ pi-  < 2.2 10^-3",
    "## PDG2002 G33 : B0B -> D2(2460)*+ rho- < 4.9 10^-3",
    "## make BFs the same as those for D**0 pi-",
    "## except exclude broad states on advice from Kuzmin",
    "#0.00083   D_0*+ pi-                PHOTOS PHSP;",
    "0.00093   D_1+ pi-                 PHOTOS SVS;",
    "#0.00062   D'_1+ pi-                PHOTOS SVS;",
    "0.00085   D_2*+ pi-                PHOTOS STS;",
    "## do the same for D**+ rho-",
    "## not sure these should be modelled by phase space but it will do",
    "### 2003: leave the same for now",
    "#0.00083   D_0*+ rho-              PHOTOS PHSP;",
    "0.00093   D_1+ rho-                PHOTOS PHSP;",
    "#0.00062   D'_1+ rho-               PHOTOS PHSP;",
    "0.00085   D_2*+ rho-               PHOTOS PHSP;",
    "#  B0B -> D**0 h0",
    "# These are added from PDG2005.",
    "## PDG2005 G47 : B0B -> D_2*0 pi+ = 0.00053",
    "## PDG2005 G48 : B0B -> D_1 pi+ = 0.00102",
    "## PDG2005 G49 : B0B -> D'_1 pi+ = 0.00075",
    "## PDG2005 G50 : B0B -> D_0*0 pi+ = 0.00091",
    "# and",
    "## PDG2005 G91 : B0B -> D*0 pi0 = 0.00027",
    "## PDG2005 G92 : B0B -> D*0 rho0 = 0.00051 (90% U.L.)",
    "## PDG2005 G93 : B0B -> D*0 eta = 0.00026",
    "## PDG2005 G94 : B0B -> D*0 eta' = 0.00026 (90% U.L.)",
    "0.00003   D_2*0 pi0                STS;",
    "0.00006   D_10 pi0                 SVS;",
    "0.00004   D'_10 pi0                SVS;",
    "0.00005   D_0*0 pi0                PHSP;",
    "0.00003   D_2*0 eta                STS;",
    "0.00006   D_10 eta                 SVS;",
    "0.00004   D'_10 eta                SVS;",
    "0.00005   D_0*0 eta                PHSP;",
    "0.00003   D_2*0 eta'                STS;",
    "0.00006   D_10 eta'                 SVS;",
    "0.00004   D'_10 eta'                SVS;",
    "0.00005   D_0*0 eta'                PHSP;",
    "0.00005   D_2*0 omega               PHSP;",
    "0.00009   D_10 omega                PHSP;",
    "0.00007   D'_10 omega               PHSP;",
    "0.00008   D_0*0 omega              PHSP;",
    "0.00006   D_2*0 rho0               PHSP;",
    "0.00011   D_10 rho0                PHSP;",
    "0.00008   D'_10 rho0               PHSP;",
    "0.00010   D_0*0 rho0              PHSP;",
    "#  B0B -> D0 X - colour suppressed",
    "### PDG 2007",
    "0.00026  D0 pi0                   PHSP;",
    "### PDG 2003 is confusing",
    "### Kuzmin is updating - leave same for now",
    "0.00045  D0  pi+  pi-             PHOTOS PHSP;",
    "### PDG 2007",
    "0.00029  rho0 D0                  SVS;",
    "### not in PDG 2007",
    "0.0005   D0  pi0  pi0             PHSP;",
    "### PDG 2007",
    "0.00020  D0 eta                   PHSP;",
    "### PDG 2007",
    "0.00013  D0 eta'                  PHSP;",
    "### PDG 2007",
    "0.00026  omega D0                 SVS;",
    "### not in PDG 2007",
    "0.0010   D0  pi+  pi-  pi0        PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0001   D0  pi0  pi0  pi0        PHSP;",
    "### not in PDG 2003 - PRL 89, 151802 (2002)",
    "### nb. phase space is not a good model",
    "### PDG 2007 updated(0.000118->0.000114)",
    "0.000114 D0 p+ anti-p-            PHOTOS PHSP;",
    "#  B0B -> D*0 X - colour suppressed",
    "### PDG 2007",
    "0.00017  D*0 pi0                  SVS;",
    "### PDG 2003 is confusing",
    "### Kuzmin is updating - leave same for now",
    "0.00032  D*0 pi+  pi-             PHOTOS PHSP;",
    "### not in PDG 2007",
    "### BF set same as D0 case, SVV_HELAMP parameters are the same as in D*0 rho-",
    "0.00029  D*0 rho0                 SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### not in PDG 2007",
    "0.0005   D*0 pi0  pi0             PHSP;",
    "### PDG 2007",
    "0.00018  D*0 eta                  SVS;",
    "### PDG 2007",
    "0.00012  D*0 eta'                 SVS;",
    "### PDG 2007",
    "0.00027  D*0 omega                SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### not in PDG 2007",
    "0.0010   D*0  pi+  pi-  pi0       PHOTOS PHSP;",
    "### not in PDG 2007",
    "0.0001   D*0  pi0  pi0  pi0       PHSP;",
    "### not in PDG 2003 - PRL 89, 151802 (2002)",
    "### nb. phase space is not a good model",
    "0.000103 D*0 p+ anti-p-           PHOTOS PHSP;",
    "#  B0B -> D(*)+ K(*)-",
    "### PDG 2007",
    "0.00020  D+  K-                   PHOTOS PHSP;",
    "### PDG 2007",
    "0.00045  K*- D+                   PHOTOS SVS;",
    "### PDG 2007",
    "0.00021  D*+ K-                   PHOTOS SVS;",
    "### PDG 2007 - helicity amplitudes same as D*0 rho-",
    "0.00033  D*+ K*-                  PHOTOS SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "# colour suppressed B -> D(*)0 K(*)0",
    "### PDG 2007",
    "0.000052 D0 anti-K0               PHSP;",
    "### PDG 2007",
    "0.000042 anti-K*0 D0              SVS;",
    "### not in PDG 2007",
    "0.000036 D*0 anti-K0              SVS;",
    "### not in PDG 2007 - assume same as D0 case - HELAMP parameters same as D*+ K*-",
    "0.000042 D*0 anti-K*0             SVV_HELAMP 0.228 0.0 0.932 0.0 0.283 0;",
    "### 2005 update (same as 2003)",
    "###  - assume same iso-spin violation in b -> c (s ubar) as b -> c (d ubar)",
    "###  - make b -> c (s ubar) sum to 2.56% (2.5% / 0.9764)",
    "###  - sum of b -> c (s ubar) BFs is 0.001346 (2005 -> 0.001303)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra are softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.024297  anti-u  s  c  anti-d            PHOTOS PYTHIA 48;",
    "# leftovers",
    "## B0B -> wrong flavour charm",
    "## excluded since not b->c transitions",
    "# mopping up",
    "### 2005 update (same as 2003)",
    "###  - include iso-spin violating effect",
    "###  - make b -> c (d ubar) + b -> c (s ubar) sum to 46.6% (2005 -> 49.8%)",
    "###  - sum of BFs is 11.4038% (2005 -> 11.8358%)",
    "###  - use PYTHIA 48 as per previous comments",
    "# ( Do not use PYTHIA 2? since D spectra are softer. A.Ishikawa 20021210 )",
    "###                              T. Gershon 2003/10/27 & 2005/03/29",
    "0.372909  anti-u  d  c  anti-d           PHOTOS PYTHIA 48;# ====B->DX End",
    "#0.379642  anti-u  d  c  anti-d           PHOTOS PYTHIA 48;",
    "#0.363642  anti-u  d  c  anti-d           PHOTOS PYTHIA 48;",
    "# B->BaryonX",
    "# B0 -> charmed baryons, total 6.3%",
    "#  Sigma_c N",
    "#    two body decay",
    "0.000022       Lambda_c+ anti-p-                 PHOTOS PHSP;",
    "0.000022       Sigma_c+  anti-p-                 PHOTOS PHSP;",
    "0.000022       Sigma_c0  anti-n0                 PHSP;",
    "0.018000       cd_0      anti-ud_0          PHOTOS PYTHIA 23;",
    "0.018000       cd_1      anti-ud_1          PHOTOS PYTHIA 23;",
    "0.006000       cd_0      anti-ud_0          PHOTOS PYTHIA 24;",
    "0.005932       cd_1      anti-ud_1          PHOTOS PYTHIA 24;",
    "#  double charmed baryon Xi_c Sigma_c",
    "#    two body decay",
    "0.001000       Xi_c+     anti-Lambda_c-          PHOTOS PHSP;",
    "0.001000       Xi_c+     anti-Sigma_c-           PHOTOS PHSP;",
    "0.001000       Xi_c0     anti-Sigma_c0           PHSP;",
    "0.009000       cs_0       anti-cd_0         PHOTOS PYTHIA 23;",
    "#  cabibbo suppressed  Xi_c N",
    "#    two body decay",
    "0.000001       Xi_c+     anti-p-                PHOTOS PHSP;",
    "0.000001       Xi_c0     anti-n0                PHSP;",
    "0.000900       cs_0       anti-ud_0        PHOTOS PYTHIA 23;",
    "0.000900       cs_1       anti-ud_1        PHOTOS PYTHIA 23;",
    "0.000300       cs_0       anti-ud_0        PHOTOS PYTHIA 24;",
    "0.000300       cs_1       anti-ud_1        PHOTOS PYTHIA 24;",
    "#  cabibbo suppressed  Sigma_c Sigma_c",
    "#    two body decay",
    "0.000040        Lambda_c+    anti-Lambda_c-     PHOTOS PHSP;",
    "0.000040        Sigma_c+     anti-Lambda_c-     PHOTOS PHSP;",
    "0.000040        Lambda_c+    anti-Sigma_c-      PHOTOS PHSP;",
    "0.000040        Sigma_c+     anti-Sigma_c-      PHOTOS PHSP;",
    "0.000040        Sigma_c0     anti-Sigma_c0      PHSP;",
    "0.000400         cd_0       anti-cd_0         PHOTOS PYTHIA 23;# ====B->BaryonX End",
    "Enddecay",
]  # type: List[str]
